
#include <stdlib.h>
#include <string.h>
#include <stream.h>

#include "Agent.h"
#include "Common.h"
#include "Log.h"
#include "WorldModel.h"
#include "Configuration.h"

/*
 * Aktueller Stand
 *
 *   $RCSfile: Agent.cc,v $
 *   $Revision: 1.41 $
 *   $Date: 1998/08/14 13:19:53 $
 *   $Author: myritz $ 
 *
 */


int main(int argc, char* argv[]) 
{
  try {  
    Agent* theAgent = new Agent (argc, argv);
    theAgent->run();
  }
  catch (Exception e) {
    ERROR( e.toString() );
    exit( -1 );
  }
  return 0;    
}

Agent::Agent(int argCount, char* argv []) 
{
  globalLog = new Log(1);
  globalConf = new Configuration();

  teamName = "AT_Humboldt";
  playerNumber = UNDEF_INT;
  isGoalie = NO;
  serverName = "localhost";
  serverConfigurationFile = "server.conf";
  agentConfigurationFile = "agent.conf";
  extendedConfigurationFiles = Array<String>();
  isHelpWanted = NO;

  readParameter(argCount, argv);
  if( isHelpWanted ) {
    showHelp( String(argv[0]) );
    exit(0);
  }
  globalConf->readConfigurationFile( serverConfigurationFile );
  globalConf->readConfigurationFile( agentConfigurationFile );
  if( extendedConfigurationFiles.count() != 0 ) {
    for( int i = 0; i < extendedConfigurationFiles.count(); i++ ) {
      globalConf->readConfigurationFile( extendedConfigurationFiles[i] );
    }
  }
  globalLog->init( globalConf );

  globalConf->logCurrentParameterValues();
  INFO( Log::CONFIGURATION, 2, globalLog->toString() );

  serverPort = globalConf->portNumber;
  readParameter(argCount, argv);

  startComponents();    
}

void Agent::run() {
  globalTimer->run();
  while( YES ) {
    globalTimer->mainLoop();
    pause();
  }
}


void Agent::readParameter( int argCount, char* argv [] ) {
  int argIndex;
  String parameter;
  
  for(argIndex = 1; argIndex < argCount; argIndex++) {
    parameter = String( argv[argIndex] );
    INFO( Log::CONFIGURATION, 5, "Parameter: " + parameter + "\n" );
    if( parameter == "-goalie" ) {  // einstellige Parameter
      isGoalie = YES;
    }
    else if( parameter == "-help" || parameter == "-h" ) {
      isHelpWanted = YES;
      return;
    }
    else {  // zweistellige Parameter
      argIndex++;
      if( !(argIndex < argCount) ) {
	throwException(InitFailedException( "Missing parameter after option " +
	    String( argv[argIndex-1] ) + ". Aborting initialization."));
      }
      
      if( parameter == "-team" ) {
	teamName = String( argv[argIndex] );
      }
      else if( parameter == "-number" ) {
	playerNumber = atoi( argv[argIndex] );
      }
      else if( parameter == "-host" ) {
	serverName = String( argv[argIndex] );
      }
      else if( parameter == "-port" ) {
	serverPort = atoi( argv[argIndex] );
      }
      else if( parameter == "-sfile" ) {
	serverConfigurationFile = String( argv[argIndex] );
      }
      else if( parameter == "-afile" ) {
	agentConfigurationFile = String( argv[argIndex] );
      }
      else if( parameter == "-efile" ) {
	extendedConfigurationFiles.addObject( String( argv[argIndex] ) );
      }
      else {
	argIndex--;
        WARNING( Log::CONFIGURATION, "Unknown option " +
                 String( argv[argIndex] ) + " is ignored." );
      } 
    }
  }  // Ende for-Schleife
}

void Agent::initRandomNumberGenerator()
{
  static long probabilityState[32] = {
    3,
    0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342,
    0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd,
    0x8c2e680f, 0xeb3d799f, 0xb11ee0b7, 0x2d436b86,
    0xda672e2a, 0x1588ca88, 0xe369735d, 0x904f35f7,
    0xd7158fd6, 0x6fa6f051, 0x616e6b96, 0xac94efdc,
    0xde3b81e0, 0xdf0a6fb5, 0xf103bc02, 0x48f340fb,
    0x36413f93, 0xc622c298, 0xf5a42ab8, 0x8a88d77b,
    0xf5ad9d0e, 0x8999220b, 0x27fb47b9
  };
  unsigned int seed = 1;
  int n = 128;
  
  initstate(seed, (char *)probabilityState, n);
  setstate( (const char *)probabilityState );
  // Cast wegen altmodischer Interfacedeklaration von initstate und setstate.
}


void Agent::startComponents() {
  bool isFirstHalf = (playerNumber == UNDEF_INT);
  initRandomNumberGenerator();
  
  // Sensoren und Effektoren
  theCommunicator = new Communicator( serverName, serverPort );
  theSensors = new Sensors( theCommunicator );
  theEffectors = new Effectors( theCommunicator );
  
  // globales Weltmodell, Planen und Ueberlegen
  globalWorldModel = new WorldModel( theSensors );
  globalSkills = new Skills();
  theDeliberator = new Deliberator( theEffectors );
  globalTimer = new Timer( theSensors, theEffectors, theDeliberator );

  // Feld betreten
  if (isFirstHalf) {
    theEffectors->enterField( teamName, isGoalie );
  }
  else {
    theEffectors->reEnterField( teamName, playerNumber );
  }    
}

void Agent::showHelp( String programName ) {
  cout << endl << "*******  AT Humboldt 98  *******" << endl << endl;
  cout << "Humboldt University Berlin" << endl;
  cout << "Department of Computer Science" << endl;
  cout << "AI Domain" << endl;
  cout << endl << endl;
  cout << "Usage: " << programName << " [options]" << endl;
  cout << "Options:" << endl;

  cout << "  -h, -help          " << "\t" << 
    "this help info" << endl;
  cout << "  -host HOSTNAME     " << "\t" << 
    "the host where the server runs (localhost)" << endl;
  cout << "  -team TEAMNAME     " << "\t" << 
    "the name of the team (AT_Humboldt)" << endl;
  cout << "  -number NUMBER     " << "\t" << 
    "the number of the player (only needed for the second half)" << endl;
  cout << "  -goalie            " << "\t" << 
    "if the player should act as an goalie" << endl;
  cout << "  -port PORT_NUMBER  " << "\t" << 
    "the port at which the server waits for the agents (6000)" << endl;
  cout << "  -sfile SERVER_FILE " << "\t" << 
    "the configuration file of the soccerserver (server.conf)" << endl;
  cout << "  -afile AGENT_FILE  " << "\t" << 
    "the configuration file of the agent (agent.conf)" << endl;
  cout << "  -efile AGENT_FILE  " << "\t" << 
    "the extended configuration file of the agent (extended.conf) " <<
    "- multiple use allowed" << endl;
}










