/*
 * Aktueller Stand
 *
 *   $RCSfile: Configuration.cc,v $
 *   $Revision: 1.66 $
 *   $Date: 1998/08/14 13:20:04 $
 *   $Author: myritz $ 
 *
 */
 
#include <stdio.h>
#include <ctype.h>
#include "Configuration.h"
#include "Exception.h"
 
Configuration *globalConf;

Configuration::Configuration() {

/* Die ganze Datenstruktur mit default-Werten */
  static options serverOptionsTemp[] = {
    // Server-Parameter
    {"goal_width",          (void *)&goalWidth,           DOUBLE,  7.32},
    {"player_size",         (void *)&playerSize,          DOUBLE,  1.0},
    {"player_decay",        (void *)&playerDecay,         DOUBLE,  0.5},
    {"player_rand",         (void *)&playerRand,          DOUBLE,  0.1},
    {"player_weight",       (void *)&playerWeight,        DOUBLE,  60.0},
    {"player_speed_max",    (void *)&playerSpeedMax,      DOUBLE,  32.0},
    {"stamina_max",         (void *)&playerStaminaMax,    DOUBLE,  2500.0},
    {"stamina_inc_max",     (void *)&playerStaminaInc,    DOUBLE,  50.0},
    {"recover_dec_thr",     (void *)&playerRecoverDecThr, DOUBLE, 0.3},
    {"recover_min",         (void *)&playerRecoverMin,    DOUBLE,  0.1},
    {"recover_dec",         (void *)&playerRecoverDec,    DOUBLE,  0.05},
    {"effort_dec_thr",      (void *)&playerEffortDecThr,  DOUBLE,  0.4},
    {"effort_min",          (void *)&playerEffortMin,     DOUBLE,  0.1},
    {"effort_dec",          (void *)&playerEffortDec,     DOUBLE,  0.05},
    {"effort_inc_thr",      (void *)&playerEffortIncThr,  DOUBLE,  0.9},
    {"effort_inc",          (void *)&playerEffortInc,     DOUBLE,  0.05},
    {"ball_size",           (void *)&ballSize,            DOUBLE,  0.15},
    {"ball_decay",          (void *)&ballDecay,           DOUBLE,  0.8},
    {"ball_rand",           (void *)&ballRand,            DOUBLE,  0.2},
    {"ball_weight",         (void *)&ballWeight,          DOUBLE,  0.2},
    {"ball_speed_max",      (void *)&ballSpeedMax,        DOUBLE,  32.0},
    {"dash_power_rate",     (void *)&dashPowerRate,       DOUBLE,  0.1},
    {"kick_power_rate",     (void *)&kickPowerRate,       DOUBLE,  0.1},
    {"kickable_margin",     (void *)&kickMarginServer,    DOUBLE,  1.0},
    {"control_radius",      (void *)&controlRadius,       DOUBLE,  2.0},
    {"maxpower",            (void *)&maxPower,            DOUBLE,  100.0},
    {"minpower",            (void *)&minPower,            DOUBLE,  -30.0},
    {"maxmoment",           (void *)&maxMoment,           DOUBLE,  180.0},
    {"minmoment",           (void *)&minMoment,           DOUBLE,  -180.0},
    {"visible_angle",       (void *)&visibleAngleInDegree,DOUBLE,  90.0},
    {"visible_distance",    (void *)&visibleDistance,     DOUBLE,  3.0},
    {"wind_dir",            (void *)&windDirection,       DOUBLE,  0.0},
    {"wind_force",          (void *)&windForce,           DOUBLE,  10.0},
    {"wind_rand",           (void *)&windRand,            DOUBLE,  0.3},
    {"catchable_area_l",    (void *)&catchAreaLength,     DOUBLE,  2.0},
    {"catchable_area_w",    (void *)&catchAreaWidth,      DOUBLE,  1.0},
    {"catch_probability",   (void *)&catchProbability,    DOUBLE,  1.0},
    {"ckick_margin",        (void *)&cornerKickMargin,    DOUBLE,  1.0},
    {"wind_none",           (void *)&isWindFactor,        OPTIONAL,    0},
    {"wind_random",         (void *)&windRandom,          OPTIONAL,    0},
    {"port",                (void *)&portNumber,          INT,     6000},
    {"coach_port",          (void *)&coachPortNumber,     INT,     6001},
    {"half_time",           (void *)&halfTimeInSecs,      INT,     600},
    {"simulator_step",      (void *)&simulatorStep,       INT,     100},
    {"send_step",           (void *)&sendStep,            INT,     150},
    {"recv_step",           (void *)&receiveStep,         INT,     20},
    {"say_msg_size",        (void *)&sayMessageSize,      INT,     256},
    {"hear_max",            (void *)&hearCapacityMax,     INT,     2},
    {"hear_inc",            (void *)&hearCapacityInc,     INT,     1},
    {"hear_decay",          (void *)&hearCapacityDecay,   INT,     1},
    {"catch_ban_cycle",     (void *)&catchCycle,          INT,     5},
    {"use_offside",         (void *)&useOffSide,          BOOL,   1},
    {"forbid_kick_off_offside", (void *)&forbidKickOffOffside, BOOL, 1},
    {"audio_cut_distance",      (void *)&audioCutOffDistance, DOUBLE,  50.0},
    {"quantize_step",       (void *)&distQuantStep,   DOUBLE,  0.1},
    {"quantize_step_l",     (void *)&landmarksQuantStep,  DOUBLE,  0.01},
    {"log_file",            (void *)&logFile,             STRING,  0},
    {"record",              (void *)&recFile,             STRING,  0},
    {"record_log",          (void *)&recLog,              OPTIONAL,    1},
    {"record_version",      (void *)&recordVersion,       INT,    2},
    {"replay",              (void *)&replay,              STRING,  0},
    {"coach",               (void *)&coachMode,           OPTIONAL,    0},
    {"coach_w_referee",     (void *)&coachWithRefereeMode,OPTIONAL,    0},
    {"old_coach_hear",      (void *)&oldCoachHear,        OPTIONAL,    0},
    {"verbose",             (void *)&verbose,             OPTIONAL,    0},
    {"inertia_moment",      (void *)&inertiaMoment,       DOUBLE,  5.0},
    {"offside_active_area_size", (void *)&offsideArea,    DOUBLE,  9.15},
    
    // Agent-Parameter
    {"optimalKickDist",     (void *)&optimalKickDist,    DOUBLE,  0.5},
    {"maxKickDist",     (void *)&maxKickDist,    DOUBLE,  1.0},
    {"historyCount",        (void *)&historyCount,       INT,     7},
    {"maxLengthOfPlan",     (void *)&maxLengthOfPlan,    INT,     7},
    {"toleratedStabilityCost", (void *)&toleratedStabilityCost, INT, 0},
    {"minConstraintUtility", (void *)&minConstraintUtility, INT,  7},
    {"minDesireUtility",    (void *)&minDesireUtility,   INT,     7},
    {"enforcedMinDistToBall", (void *)&enforcedMinDistToBall, DOUBLE, 9.15},
    {"maxStepsOfPlayerMovement", (void *)&maxStepsOfPlayerMovement, INT, 7},
    {"maxStepsOfBallMovement", (void *)&maxStepsOfBallMovement, INT, 25},
    {"maxEffectiveBallSpeed", (void *)&maxEffectiveBallSpeed, DOUBLE, 2.4},
    {"averageStepsOfBallMovement", (void *)&averageStepsOfBallMovement,INT,15},
    {"roles", (void *)&roles, ARRAY_OF_INT, 11},
    {"kickoffPositionsOffensive", (void *)&kickoffPositionsOffensive, ARRAY_OF_VECTOR, 11},
    {"kickoffPositionsDefensive", (void *)&kickoffPositionsDefensive, ARRAY_OF_VECTOR, 11},
    {"homePositions", (void *)&homePositions, ARRAY_OF_VECTOR, 12},
    {"maxGoalKickUtility",    (void *)&maxGoalKickUtility,   INT, 900},
    {"goalKickUtilityWeights", (void *)&goalKickUtilityWeights, ARRAY_OF_INT, 4},
    {"maxForwardPassUtility",    (void *)&maxForwardPassUtility,   INT, 600},
    {"minForwardPassUtility",    (void *)&minForwardPassUtility,   INT, 50},
    {"maxDirectPassUtility",    (void *)&maxDirectPassUtility,   INT,   700},
    {"maxDoublePassUtility",    (void *)&maxDoublePassUtility,   INT,   600},
    {"maxDribbleUtility",    (void *)&maxDribbleUtility,   INT,     600},
    {"maxControlBallUtility",    (void *)&maxControlBallUtility,   INT, 1000},
    {"maxAvoidOffsideUtility", (void *)&maxAvoidOffsideUtility, INT, 500},
    {"avoidOffsideStaminaLimitBonus", (void *)&avoidOffsideStaminaLimitBonus, DOUBLE, 200},
    {"maxConserveStaminaUtility", (void *)&maxConserveStaminaUtility, INT, 1000},
    {"conserveStaminaActivation", (void *)&conserveStaminaActivation, INT, 1500},
    {"conserveStaminaUtilityIncrease", (void *)&conserveStaminaUtilityIncrease, VECTOR, UNDEF_INT },
    {"dribbleSectorLength",    (void *)&dribbleSectorLength, DOUBLE , 5.0},
    {"dribbleSectorWidth",    (void *)&dribbleSectorWidth, DOUBLE, 10.0},
    {"minCollectDataUtility", (void *)&minCollectDataUtility, INT, 100},    
    {"normalCollectDataUtility", (void *)&normalCollectDataUtility, INT, 200},
    {"maxCollectDataUtility", (void *)&maxCollectDataUtility, INT, 300},
    {"normalGoToBaseUtility", (void *)&normalGoToBaseUtility, INT, 150},
    {"minGoToBaseUtility", (void *)&minGoToBaseUtility, INT, 50},
    {"defenderMoveForwardBallDistActivation", (void *)&defenderMoveForwardBallDistActivation, DOUBLE, 30},
    {"defenderMoveForwardStopLine", (void *)&defenderMoveForwardStopLine, DOUBLE, -10},
    {"midfieldMoveForwardBallDistActivation", (void *)&midfieldMoveForwardBallDistActivation, DOUBLE, 15},
    {"midfieldMoveForwardStopLine", (void *)&midfieldMoveForwardStopLine, DOUBLE, 10},
    {"gotoBasePositionAccuracy", (void *)&gotoBasePositionAccuracy, DOUBLE, 1.0},
    {"getIntoSpaceAccuracy", (void *)&getIntoSpaceAccuracy, DOUBLE, 1.0},
    {"interceptBallAccuracy", (void *)&interceptBallAccuracy, DOUBLE, 1.0},
    {"normalInterceptBallUtility", (void *)&normalInterceptBallUtility, INT, 800},
    {"normalGetIntoSpaceUtility", (void *)&normalGetIntoSpaceUtility, INT, 700},
    {"leaveBasePositionRange", (void *)&leaveBasePositionRange, DOUBLE, 12},
    {"getIntoSpaceDist", (void *)&getIntoSpaceDist, INT, 20},
    {"getIntoSpaceMinDist", (void *)&getIntoSpaceMinDist, DOUBLE, 10},
    {"getIntoSpaceSimSteps", (void *)&getIntoSpaceSimSteps, INT, 7},
    {"getIntoSpaceAvoidBasePositionDist", (void *)&getIntoSpaceAvoidBasePositionDist, DOUBLE, 5},
    {"avoidCoverActivationDist", (void *)&avoidCoverActivationDist, DOUBLE, 3},
    {"avoidCoverBestPositionDist", (void *)&avoidCoverBestPositionDist, DOUBLE, 2},
    {"maxForceContinuationTime", (void *)&maxForceContinuationTime, INT, 25},
    {"distToBallForEvaluation", (void *)&distToBallForEvaluation,DOUBLE,40.0},
    {"maxStepsToCalculate", (void *)&maxStepsToCalculate, INT, 40 },
    {"debugModes",              (void *)&debugModes,  ARRAY_OF_INT,     13},
    {"tournamentMode",              (void *)&tournamentMode,  BOOL,     0},
    {"useCommunication",              (void *)&useCommunication,  BOOL,     0},
    {"communicationTimeout", (void *)&communicationTimeout, INT, 10},
    {"dribbleStaminaLimit", (void *)&dribbleStaminaLimit, DOUBLE, 800},
    {"interceptBallStaminaLimit", (void *)&interceptBallStaminaLimit, DOUBLE, 500},
    {"gotoBaseStaminaLimit", (void *)&gotoBaseStaminaLimit, DOUBLE, 900},
    {"getIntoSpaceStaminaLimit", (void *)&getIntoSpaceStaminaLimit, DOUBLE, 600},
    {"interceptionTolerance", (void *)&interceptionTolerance, INT, 2 },
    {"goalieDistFromGoal", (void *)&goalieDistFromGoal, DOUBLE, 7.0 },
    {"passLeadForTeammates", (void *)&passLeadForTeammates, INT, 5 },
    {"emergencyPassDist", (void *)&emergencyPassDist, INT, 5 },
    {"reactionTimeSteps", (void *)&reactionTimeSteps, INT, 2 },
    {"dribbleInOffsideBonus", (void *)&dribbleInOffsideBonus, INT, 100 },
    {"maxGoalKickDist", (void *)&maxGoalKickDist, DOUBLE, 19.0 },

  //Online Learning Parameter    
    { "xResolution",               (void *)&xResolution,             INT,  5 },
    { "yResolution",               (void *)&yResolution,             INT,  4 },
    { "incPlayer",                 (void *)&incPlayer,               INT, 10 },
    { "incBall",                   (void *)&incBall,                 INT, 20 },
    { "decPlayer",                 (void *)&decPlayer,               INT,  5 },
    { "decBall",                   (void *)&decBall,                 INT,  0 },
    { "threshold",                 (void *)&threshold,               INT, 30 },
    { "minEdgeLength",             (void *)&minEdgeLength,           INT,  6 },
    { "maxEdgeLength",             (void *)&maxEdgeLength,           INT, 35 },
    { "manGuardingBorder",         (void *)&manGuardingBorder,       INT, 55 },
    { "distFromOpponent",      (void *)&distFromOpponent,    INT,  3 },
    { "weightDistFromHome",(void *)&weightDistFromHome, DOUBLE, 9.0 },
    { "weightDistFromCurrent", (void *)&weightDistFromCurrent,DOUBLE,2.0 },
    { "weightBallAppearance",     (void *)&weightBallAppearance,DOUBLE, 0.5 },
    { "weightEdgeColour",         (void *)&weightEdgeColour,   DOUBLE, 1.0  },
    { "maxUtilityDist",       (void *)&maxUtilityDist, INT, 40 },
    { "minUtilityDist",       (void *)&minUtilityDist, INT, 12 },
    { "ballLearnDist",        (void *)&ballLearnDist,  INT, 12 },
    { "maxDistFromHome",      (void *)&maxDistFromHome, INT, 20 },
    { "learnNoGoalieDist",    (void *)&learnNoGoalieDist, INT, 3 }, 
    { "defenderFreeKickDist", (void *)&defenderFreeKickDist, DOUBLE, 13 }, 
    { "defenderFreeKickMoveUp", (void *)&defenderFreeKickMoveUp, DOUBLE, 5 }, 
    { "dangerousRangeFromGoal", (void *)&dangerousRangeFromGoal, DOUBLE, 40 },
    { "outerKickCircleTolerance", (void *)&outerKickCircleTolerance, DOUBLE, 0.15 },
    { "innerKickCircleTolerance", (void *)&innerKickCircleTolerance, DOUBLE, 0.2 },
    { "intermediateKickTolerance", (void *)&intermediateKickTolerance, DOUBLE, 0.2 },
    { "ourMessageIdentifier", (void *)&ourMessageIdentifier, INT, 666 },
    { "escapeFieldOfViewDist", (void *)&escapeFieldOfViewDist, DOUBLE, 5 },
    { "maximalMergeDist", (void *)&maximalMergeDist, DOUBLE, 10 },
    { "distToPlayerForSay", (void *)&distToPlayerForSay, DOUBLE, 30 },
    { "sayAngleDifference", (void *)&sayAngleDifference, DOUBLE, 20 },

    // Ende
    {"\0",                          NULL,                0,       0}
  };

  serverOptions = serverOptionsTemp;

  int i;
  /* Einstellen der default-Werte */
  for(i=0; serverOptions[i].pointer != NULL; i++) {
    switch( serverOptions[i].type ) {
      case DOUBLE:
        *(double *)serverOptions[i].pointer = serverOptions[i].value;
        break;
      case INT:
        *(int *)serverOptions[i].pointer = (int)serverOptions[i].value;
        break;
      case BOOL:
      case OPTIONAL:
        *(bool *)serverOptions[i].pointer = (bool)serverOptions[i].value;
        break;
      case STRING:
        *(char *)serverOptions[i].pointer = '\0';
        break;
      case VECTOR:
        *(Vector *)serverOptions[i].pointer = Vector();
        break;
      case ARRAY_OF_VECTOR:
        *(Array<Vector> *)serverOptions[i].pointer = 
            Array<Vector>( (int)serverOptions[i].value );
        break;
      case ARRAY_OF_INT:
        *(Array<int> *)serverOptions[i].pointer = 
            Array<int>( (int)serverOptions[i].value );
        break;
      default:
        throwException( Exception( "Unknown Parametertype" ) );
    }
  } // Ende for-Schleife

  calculateDependentParameters();
}

char *Configuration::readLine( FILE *filePointer ) 
{
  static char buffer[MAX_MESSAGE];
  bool isDataAvailable = NO;
  int length = 0;
  do {
    if ( fgets( buffer + length, MAX_MESSAGE - length, filePointer ) == NULL) {
      if ( length > 0 ) {
        break;
      } 
      else {
        return NULL;
      }
    }
    /* Kommentare und Leerzeilen ignorieren */
    if (buffer[length] == '#' || buffer[length] == '\n') {
      continue;
    }
    
    length = strlen( buffer );
    if ( length >= 2 && strcmp( buffer+length-2, "\\\n") == 0 ) {
      length -= 2;
      continue;
    }
    
    if ( length > 0 ) {
      isDataAvailable = YES;
    }
  } while ( isDataAvailable == NO );
  
  // Ersten ':' in ' ' umwandeln
  char *t = buffer;
  while(*t != '\0') {
    if (*t == ':') {
      *t = ' ' ;
      break;
    }
    t++;
  } 
   
  return buffer;
}


int Configuration::scanArrayOfVector( char *buffer, Array<Vector> *ArrayPtr )
{
  int count;
  char parameterName[256];
  char parameterBuffer[256];
  int n;
  n = sscanf(buffer, "%s %d %[^\n]", parameterName, &count, parameterBuffer);
  if ( n == 3 && count >=0 ) {
    ArrayPtr->removeAll();
    int index = 0;
    double x, y;
    char *parameterPtr = parameterBuffer;
    while ( isspace( parameterPtr[0] ) ) {
      parameterPtr++;
    }
    while ( ( n = sscanf( parameterPtr, "(%lf,%lf)%[^\n]", 
        &x, &y, parameterPtr ) ) >= 2 ) {
      if ( n == 2 ) {
        // Das letzte Element gelesen. ParameterPtr von Hand setzen
        parameterPtr[0] = '\0';
      }
      if ( index >= count ) {
        WARNING( Log::CONFIGURATION, "Too many elements given for array " +
                 String(parameterName) );
        break;
      }
      ArrayPtr->addObject( Vector( x, y) );
      index++;

      while ( isspace( parameterPtr[0] ) ) {
        parameterPtr++;
      }
    }              
    if ( n == 1 ) {
      WARNING( Log::CONFIGURATION, "Incomplete element given for array " +
               String(parameterName) );
      return n;
    }
    if ( index < count ) {
      WARNING( Log::CONFIGURATION, "Not enough elements given for array " + 
               String(parameterName) );
    } 
    else {
      n = 2;
    }
  } 
  else {
    n = 1;
  }
  return n;
}

int Configuration::scanArrayOfInt( char *buffer, Array<int> *ArrayPtr )
{

  int count;
  char parameterName[256];
  char parameterBuffer[256];
  int n;
  n = sscanf(buffer, "%s %d %[^\n]", parameterName, &count, parameterBuffer);
  if ( n == 3 && count >=0 ) {
    ArrayPtr->removeAll();
    int index = 0;
    double value;
    char *parameterPtr = parameterBuffer;
    while ( isspace( parameterPtr[0] ) ) {
      parameterPtr++;
    }
    while ( ( n = sscanf( parameterPtr, "%lf %[^\n]", 
        &value, parameterPtr ) ) >= 1 ) {
      if ( n == 1 ) {
        // Das letzte Element gelesen. ParameterPtr von Hand setzen
        parameterPtr[0] = '\0';
      }
      if ( index >= count ) {
        WARNING( Log::CONFIGURATION, "Too many elements given for array " +
                 String(parameterName) );
        break;
      }
      ArrayPtr->addObject( value );
      index++;

      while ( isspace( parameterPtr[0] ) ) {
        parameterPtr++;
      }
    }              
    if ( index < count ) {
      WARNING( Log::CONFIGURATION, "Not enough elements given for array " + 
               String(parameterName) );
    } 
    else {
      n = 2;
    }
  } 
  else {
    n = 1;
  }
  return n;
}

void Configuration::readConfigurationFile(String configurationFile) {
  int i;
  FILE *filePointer;
  char *buffer;
  
  /* Oeffnen des Files */
  if ((filePointer = fopen(configurationFile.cString(),"r")) == NULL) {
    WARNING( Log::CONFIGURATION, "Can't open config file " +
             configurationFile );
    return;
  }

  /* Einlesen der Parameter */

  while( ( buffer = readLine( filePointer ) ) != NULL) {
    int n;
    char parameterName[256];
    char onOff[16] ;

    n = sscanf(buffer,"%s", parameterName) ;
    if (n < 1) {
      WARNING( Log::CONFIGURATION, "Illegal line: (" + String(buffer) + ")" );
      continue ;
    }

    for(i = 0; serverOptions[i].pointer != NULL; i++) {
      if( strcmp(parameterName, serverOptions[i].optionName) ) {
        continue;
      }

      // Parameter gefunden
      switch( serverOptions[i].type ) {
        case DOUBLE:
          n = sscanf(buffer, "%s %lf", parameterName, 
              (double *)serverOptions[i].pointer) ;
          break ;
        case INT:
          n = sscanf(buffer, "%s %d", parameterName,
              (int *)serverOptions[i].pointer) ;
          break ;
        case OPTIONAL:
          // Wenn es da steht, dann ist es yes
          // sonst der Default-Wert.
          n = 2;
          *((int *)serverOptions[i].pointer) = YES;
          break;
        case BOOL:
          n = sscanf(buffer, "%s %s", parameterName, onOff);
          if ( ( strcasecmp( onOff, "yes" ) == 0 ) || 
               ( strcasecmp( onOff, "on" ) == 0 ) ) {
            *((int *)serverOptions[i].pointer) = YES;
          }
          else {
            *((int *)serverOptions[i].pointer) = NO;
          }  
          break ;
        case STRING:
          n = sscanf(buffer, "%s %s", parameterName,
              (char *)serverOptions[i].pointer) ;
          break ;
        case VECTOR:
          double x, y;
          n = sscanf(buffer, "%s (%lf, %lf)", parameterName, &x, &y);
          if ( n == 3 ) {
            *((Vector *)serverOptions[i].pointer) = Vector( x, y);
          } 
          else {
            n = 1;
          }
          break ;
        case ARRAY_OF_VECTOR:
          n = scanArrayOfVector( buffer, 
              (Array<Vector> *)serverOptions[i].pointer );
          break;
        case ARRAY_OF_INT:
          n = scanArrayOfInt( buffer, 
              (Array<int> *)serverOptions[i].pointer );
          break;
      }

      if (n < 2) {
        WARNING( Log::CONFIGURATION, "Illegal line (" + 
                 String(parameterName) + ")" );
      }
      break ;
    }
    
    if (serverOptions[i].pointer == NULL) {
      WARNING( Log::CONFIGURATION, "Illegal line (" +
               String(parameterName) + ")" );
    }
  }

  fclose(filePointer);

  calculateDependentParameters();

  return;
}

void Configuration::calculateDependentParameters() 
{
  kickMargin = kickMarginServer * maxKickDist;

  halfTime = halfTimeInSecs * 1000 / simulatorStep;
  
  visibleAngle = radFromDegree(visibleAngleInDegree);
  
  sendStepNorm = sendStep/4;
  
  kickRange = kickMargin + playerSize + ballSize;
  
  maxShotLength = ballSpeedMax * 
      ( ( pow( ballDecay, maxStepsOfBallMovement ) - 1 ) / ( ballDecay - 1 ) );
  
  averageShotLength = maxEffectiveBallSpeed * 
      ( ( pow( ballDecay, averageStepsOfBallMovement ) - 1 ) /
        ( ballDecay - 1 ) );

  distPerStaminaPoint = dashPowerRate *
      ( ( pow( playerDecay, maxStepsOfPlayerMovement ) - 1 ) / 
        ( playerDecay - 1 ) );
  
  sumOfGoalKickUtilityWeights = 0;
  for ( int i = 0; i < goalKickUtilityWeights.count(); i++ ) {
    sumOfGoalKickUtilityWeights += goalKickUtilityWeights[i];
  }
  conserveStaminaGradient = ( conserveStaminaActivation - 
                              conserveStaminaUtilityIncrease.x() ) / 
    conserveStaminaUtilityIncrease.y();
}

void Configuration::logCurrentParameterValues() 
{
  if ( IS_ACTIVE( Log::CONFIGURATION, 6 ) ) {
    for(int i = 0; serverOptions[i].pointer != NULL; i++) {
      globalLog->info( String( "\nParameter: " ) +
                       serverOptions[i].optionName + "\t Value: " );
      switch(serverOptions[i].type) {
        case DOUBLE:
          globalLog->info( String(*((double *)serverOptions[i].pointer)) );
          break;
        case INT:
          globalLog->info( String(*((int *)serverOptions[i].pointer)) );
          break;
        case BOOL:
          globalLog->info( 
              ( (*((int *)serverOptions[i].pointer)) ? "YES" : "NO" ) );
          break;
        case STRING:
          globalLog->info( String((char *)serverOptions[i].pointer) );
          break;
        case VECTOR:
          globalLog->info( 
              (*((Vector *)serverOptions[i].pointer)).toString() );
          break;
        case ARRAY_OF_VECTOR:
          cout << (*((Array<Vector> *)serverOptions[i].pointer));
          break;
        case ARRAY_OF_INT:
          cout <<  (*((Array<int> *)serverOptions[i].pointer));
          break;
      }
    }
    
    // Die automatischen Parameter muessen hier extra angegeben werden.
    globalLog->info( "\nDep. parameter kickMargin: " + String( kickMargin ) );
    globalLog->info( "\nDep. parameter halfTime : " + halfTime );
    globalLog->info( "\nDep. parameter visibleAngle : " + 
                     String( visibleAngle ) );
    globalLog->info( "\nDep. parameter kickRange : " + 
                     String( kickRange ) );
    globalLog->info( "\nDep. parameter maxShotLength : " + 
                     String( maxShotLength ) );
    globalLog->info( "\nDep. parameter averageShotLength : " +
                     String( averageShotLength ) );
    globalLog->info( "\nDep. parameter distPerStaminaPoint : " + 
                     String( distPerStaminaPoint ) );
    globalLog->info( "\nDep. parameter halfTime : " + String( halfTime ) );
    globalLog->info( "\nDep. parameter sumOfGoalKickUtilityWeights : " +
                     String( sumOfGoalKickUtilityWeights ) );
    globalLog->info( "\nDep. parameter conserveStaminaGradient: " +
                     String( conserveStaminaGradient ) );

    globalLog->info( "\n" );
  }
}



