
/*
 * Aktueller Stand
 *
 *   $RCSfile: DirectPassOption.cc,v $
 *   $Revision: 1.10 $
 *   $Date: 1998/08/14 13:20:16 $
 *   $Author: myritz $ 
 *
 */
 
#include "DirectPassOption.h"
#include "Constraints.h"
#include "WorldModel.h"
#include "Skills.h"
#include "Vector.h"
#include "Common.h"
#include "Reliability.h"
#include "InvalidIdException.h"


DirectPassOption::DirectPassOption() : KickOption() 
{
  theId = DIRECT_PASS_OPTION_ID;
}  


int DirectPassOption::baseUtility() const 
{

  INFO( Log::ACTIVE_OPTIONS, 2, "ENTER DirectPassOption::baseUtility\n" );

  int bestUtility = KickOption::baseUtility();
  if ( bestUtility == MAX_UTILITY ) {
    // Es liegt eine Kick-Situation vor. Jetzt genaue Werte ausrechnen.

    INFO( Log::ACTIVE_OPTIONS, 2, "ENTER Ball in kickRange\n" ); 

    // Szenario vorbereiten
    Situation scenario( globalWorldModel->situation() );
    ModelledBall ball = scenario.ball();
    ModelledPlayer player = scenario.playerWithId(globalWorldModel->id());
    ModelledPlayer bestPlayer = scenario.playerWithId(globalWorldModel->id()); 
    ModelledPlayer tempPlayer = scenario.playerWithId(globalWorldModel->id());

    Vector speed;
    bool toBeSimulated, isBestPlayerFound;
    double bestPlayerDifferenceAngle, tempPlayerDifferenceAngle, 
    bestPlayerdistToPlayer, tempPlayerdistToPlayer, 
    maxRange, minRange,idealRange, minDifference, toleranceAngle, 
    preferredAngle;
    
    // Mich selbst aus scenario streichen
    scenario.removePlayerWithId ( globalWorldModel->id() );
    // Torwart raus. Das heisst wir nehmen an, 11 ist der Torwart.
    try {
      scenario.removePlayerWithId( 430 );
    }
    catch ( Situation::NoPlayerWithThisIdException ) {
      // Wenn er nicht drin ist, muss er auch nicht rausgeschmissen
      // weden
    }

    // Winkelabweichung unseres bestPlayers von Vorzugsrichtung 
    // : initial mit 60 Grad
    if ( globalWorldModel->gameState() == WorldModel::KICK_OFF && 
         globalWorldModel->initiative() == OUR_TEAM_ID ) {
      bestPlayerDifferenceAngle = radFromDegree( 120 );
    } 
    else {
      bestPlayerDifferenceAngle = radFromDegree( 60 );
    }
    // Winkelabweichung unseres tempPlayers von Vorzugsrichtung
    tempPlayerDifferenceAngle = 0;
    // minimale Winkelabweichung - zur Prioritaetsberechnung ( 5 Grad )
    minDifference = radFromDegree( 5 );
    // maximaler Abstand zum Teammate
    maxRange = globalConf->maxShotLength; 
    // minimaler Abstand zum Teammate
    minRange = 6; 
    // idealer Abstand zum Teammate
    idealRange = globalConf->averageShotLength;
    // Abstandsabweichung unseres bestPlayers
    bestPlayerdistToPlayer = maxRange;
    // Abstandsabweichung unseres tempPlayers
    tempPlayerdistToPlayer = 0;
    // Toleranz um bestPlayerDifferenceAngle
    toleranceAngle = radFromDegree( 10 ); 
    // Winkel der Vorzugsrichtung
    preferredAngle = fabs( ( globalWorldModel->preferredKickDirection
	( player.pos() ) ).angle() );
    
    // Spieler die im Abseits stehen streichen
    INFO( Log::ACTIVE_OPTIONS, 5, "Abseitsueberpruefung ! \n");
 
    for ( int i = ( scenario.playerCount() -1 ); i >= 0; i--) {  
      if ( scenario.playerWithIndex(i).posReliability().get() > 
          Reliability::RANDOM ) {
            scenario.removePlayerWithIndex(i);
            continue;
         }
      else {
        if ( scenario.playerWithIndex(i).team() == OUR_TEAM_ID ) {
          if ( scenario.isOffsideWithIndex(i) ) {
            scenario.removePlayerWithIndex(i);
          } 
        }
      }   
    }            /* for */ 

    INFO( Log::ACTIVE_OPTIONS, 5, "Abseitsueberpruefung beendet. \n");

    // Spieler , die ausserhalb des Einflussbereiches stehen aus
    // Situation streichen

    INFO( Log::ACTIVE_OPTIONS, 5, "Spieler aus Situation streichen ! \n");
 
    for ( int i = ( scenario.playerCount() -1); i >= 0; i--) {  
      tempPlayer = scenario.playerWithIndex(i);
      toTarget = tempPlayer.pos() - player.pos();
      tempPlayerDifferenceAngle = fabs ( 
          preferredAngle - fabs ( toTarget.angle() ) );

      INFO( Log::ACTIVE_OPTIONS, 8, "tempPlayerDifferenceAngle: " + String (
          tempPlayerDifferenceAngle) + " \n" );

      if ( ( toTarget.length() > maxRange ) || ( tempPlayerDifferenceAngle > 
             bestPlayerDifferenceAngle + toleranceAngle ) ) {
        scenario.removePlayerWithIndex(i); 
      }                              
    } /* for */

    INFO( Log::ACTIVE_OPTIONS, 5, "Situation bereinigt. \n");
 
    /* Alle Teammates auf preferredKickDirection +- X untersuchen, 
       den besten ermitteln und anspielen */

    for ( int i = ( scenario.playerCount() -1 ); i >= 0; i--) {  
      if ( scenario.playerWithIndex(i).team() != OUR_TEAM_ID ) {
        continue;
      }
      tempPlayer = scenario.playerWithIndex(i);
      toTarget = tempPlayer.pos() - player.pos();
      tempPlayerDifferenceAngle = fabs ( preferredAngle - 
          fabs ( toTarget.angle() ) );
      
      INFO( Log::ACTIVE_OPTIONS, 8, "Spielerid: " + String( tempPlayer.id() ) + 
          "\ntempPlayerDifferenceAngle: " + 
          String( tempPlayerDifferenceAngle ) + 
          "\nbestPlayerDifferenceAngle: " + 
          String( bestPlayerDifferenceAngle ) +
          "\n" );
      
      tempPlayerdistToPlayer = toTarget.length();
      toBeSimulated = NO;

      // Vergleich mit bestPlayer - nur wer besser ist, wird simuliert
      if ( ( tempPlayerDifferenceAngle <= bestPlayerDifferenceAngle ) &&
           ( tempPlayerdistToPlayer >= minRange ) ) {
        toBeSimulated = YES;
        if ( ( ( bestPlayerDifferenceAngle - tempPlayerDifferenceAngle ) 
            <= minDifference ) &&
            ( ( tempPlayerdistToPlayer - idealRange ) > 
            ( bestPlayerdistToPlayer - idealRange ) ) ) {
          INFO( Log::ACTIVE_OPTIONS, 11, "Wir simulieren nicht\n");
          toBeSimulated = NO;
        }
        else {
          INFO( Log::ACTIVE_OPTIONS, 11, "Wir simulieren\n");
        }
      }

      /* Simulation */
      if ( toBeSimulated == YES ) { 
        INFO( Log::ACTIVE_OPTIONS, 5, "Simulation beginnt. \n");
        toTarget.normalize();
        ball = globalWorldModel->ball();
        speed = globalConf->ballSpeedMax * toTarget;
        ball.setSpeed( speed );
        ball.simulateSteps( 1 );
        scenario.setBall( ball );
        try {  
          const ModelledPlayer& playerWhoInterceptFirst = 
              scenario.nearestPlayer();
          if ( playerWhoInterceptFirst.id() == 
              scenario.playerWithIndex(i).id() && 
              scenario.stepsOfNearestPlayer( THEIR_TEAM_ID ) > 
              scenario.stepsOfNearestPlayer( OUR_TEAM_ID ) + 
              globalConf->passLeadForTeammates ) {
            bestPlayer = scenario.playerWithIndex(i);
            bestPlayerDifferenceAngle = tempPlayerDifferenceAngle;
            bestPlayerdistToPlayer = tempPlayerdistToPlayer;
          }
        } 
        catch ( InvalidIdException ) {
          break;
        }
      }                
    } /* for */
      
    INFO( Log::ACTIVE_OPTIONS, 5, "Simulation beendet. \n");
    if ( isBestPlayerFound = ( bestPlayer.id() != player.id() ) ) {
      toTarget = bestPlayer.pos()-player.pos();
      bestUtility = globalConf->maxDirectPassUtility;
    }
    // kein bester Anspielpartner gefunden - bevorzugte Richtung spielen
    else { 
      toTarget = globalWorldModel->preferredKickDirection( player.pos() );
      bestUtility = NEUTRAL_UTILITY;
    }  
  }
  else {
    // nichts machen. Die default-Werte sind ausreichend.
  }
  INFO( Log::ACTIVE_OPTIONS, 5, "Utility " + String( bestUtility ) +
      " fuer Schuss in Richtung " + toTarget.toString() + "\n" );

  return bestUtility;
}


StorableObject* DirectPassOption::copy() const 
{
  return new DirectPassOption( *this );
}

String DirectPassOption::toString() const  
{
  return String( "( DirectPassOption utility: " ) +
    String( expectedUtility() ) + String( " ) " + 
    toTarget.toString() ) ;
}



 



