
/*
 * Aktueller Stand
 *
 *   $RCSfile: DribbleOption.cc,v $
 *   $Revision: 1.16 $
 *   $Date: 1998/08/14 13:20:24 $
 *   $Author: myritz $ 
 *
 */
 
#include "DribbleOption.h"
#include "Constraints.h"
#include "WorldModel.h"
#include "Skills.h"

DribbleOption::DribbleOption() : KickOption() 
{
  theId = DRIBBLE_OPTION_ID;
}  



int DribbleOption::baseUtility() const 
{
  INFO( Log::ACTIVE_OPTIONS, 2, 
      "ENTER DribbleOption::baseUtility\n" );
  int bestUtility = KickOption::baseUtility();
  toTarget = Vector(); // target muss neu bestimmt werden
  if ( bestUtility == MAX_UTILITY ) {
    
    // Es liegt eine Kick-Situation vor. Jetzt genaue Werte ausrechnen.
    bestUtility = (int)( globalConf->maxDribbleUtility );
    const Situation& situation = globalWorldModel->situation();
    const ModelledPlayer& player = globalWorldModel->player();
    
    // wir dribbeln nicht im eigenen Strafraum
    if ( player.pos().x() <=globalWorldModel->field().posOfPoint( 
        UPPER_LEFT_PENALTY_FLAG_ID ).x() &&
        player.pos().y() <= globalWorldModel->field().posOfPoint(
        UPPER_LEFT_PENALTY_FLAG_ID ).y() &&
	      player.pos().y() >=
	      globalWorldModel->field().posOfPoint( 
        LOWER_LEFT_PENALTY_FLAG_ID).y() ) {
      bestUtility = NEUTRAL_UTILITY;
    }
    else {        
      for ( int i = situation.playerCount() - 1; i >= 0; i-- ) {
        const ModelledPlayer& currentPlayer = situation.playerWithIndex( i );

        if ( ( currentPlayer.id() != globalWorldModel->id() &&
            situation.ball().isInKickRangeOfPlayer( currentPlayer ) == YES ) || 
            ( currentPlayer.team() != OUR_TEAM_ID &&
            player.distTo( currentPlayer.pos() ) < 
            globalConf->emergencyPassDist ) ) {
          // wir koennen nicht dribbeln, weil wir nicht die alleinige
          // Ballkontrolle haben, oder weil uns ein anderer Spieler
          // gefaehrlich nahe kommt.
          bestUtility = NEUTRAL_UTILITY;
          setIsOptionPossible( NO );
        }
      }
    }
    // Sind wir im Abseits, bekommt das Dribbeln einen Bonus
    if ( globalWorldModel->situation().offsideLineX() < 
            player.pos().x() ) {
      bestUtility = bestUtility + globalConf->dribbleInOffsideBonus;
    }
  }
  INFO( Log::ACTIVE_OPTIONS, 5, "Utility " + String( bestUtility ) + "\n" );
  return bestUtility;
}


bool DribbleOption::isPossible() const 
{
  INFO( Log::ACTIVE_OPTIONS, 3, "ENTER DribbleOption::isPossible\n" );
  if ( toTarget.isValid() == NO ) {
    Vector myPos( globalWorldModel->player().pos() );
    globalSkills->resetSituation();
    globalSkills->initEmptySectorSearch(
        globalConf->dribbleSectorLength * 
            globalWorldModel->preferredKickDirection( myPos ) + myPos,
        radFromDegree( globalConf->dribbleSectorWidth ),
        globalConf->dribbleSectorLength,
        PI/2,
        radFromDegree( 10.0 ) );
    toTarget = globalSkills->findEmptySector( globalWorldModel->situation() );
    INFO( Log::ACTIVE_OPTIONS, 7, "dribble target: " + toTarget.toString() +
        "\n" );
  }
  else {
    WARNING( Log::ACTIVE_OPTIONS, "dribble target was already defined\n" );
  }
    
  INFO( Log::ACTIVE_OPTIONS, 3, "LEAVE DribbleOption::isPossible\n" );
  return ( toTarget.isValid() && isOptionPossible() );
}


Plan DribbleOption::generatePlan() const 
{
  INFO( Log::ACTIVE_OPTIONS, 3, "ENTER DribbleOption::generatePlan\n" );
  // isPossible waehlt das Target (ueber chooseTarget) und
  // gibt zurueck, ob die Auswahl erfolgreich war
  if ( isPossible() == NO ) {
    WARNING( Log::ACTIVE_OPTIONS, "Generating Plan for impossible Option.\n" );
    INFO( Log::ACTIVE_OPTIONS, 3, "LEAVE DribbleOption::generatePlan\n" );
    return Plan();
  }
  double direction;
  
  INFO( Log::ACTIVE_OPTIONS, 9, "Dribble to relative position: " +
      toTarget.toString() + "\n" );
  direction = toTarget.angle();
  globalSkills->resetSituation();
  Plan plan;
  plan.addPlan( globalSkills->setViewWidth( VIEW_WIDTH_NARROW ) );
  
  // wir dribbeln nur einfach in Region 6 und 7, sowie vor und im eigenen
  // Strafraum ( wegen der Sicherheit )
  int region = globalWorldModel->region( globalWorldModel->player().pos() );
  if ( region == 6 || region == 7 || 
      globalWorldModel->player().pos().x() < -20.0 ) {
    plan.addPlan( globalSkills->trivialDribble( 
        globalConf->dribbleStaminaLimit, direction, NO_SIDE_ID ) );
  }
  else {
    plan.addPlan( globalSkills->dribble( globalConf->dribbleStaminaLimit, 
        direction, chooseDribbleSide( direction ) ) );
  }
  return plan;
}

int DribbleOption::chooseDribbleSide( double direction ) const
{
  INFO( Log::ACTIVE_OPTIONS, 4, "ENTER DribbleOption::chooseDribbleSide\n" );
  Vector toCriticalPos;
  const Vector& ownPos = globalWorldModel->player().pos();
  int side;
  //   const Array<int>& sortedOpponentIndices = 
  //       globalWorldModel->situation().getTheirTeamSorted();
  //   const Situation& situation = globalWorldModel->situation();
  //   bool isCriticalPosFound = NO;
  //   for ( int i = 0; i < sortedOpponentIndices.count(); i++ ) {
  //     toCriticalPos = situation.playerWithIndex( 
  //         sortedOpponentIndices[i] ).pos() - ownPos;
  //     if ( fabs( toCriticalPos.angle() - direction ) > PI/2 ) {
  //       continue;
  //     }
  //     if ( toCriticalPos.length() > globalConf->dribbleSectorLength ) {
  //       continue;
  //     }
  //     isCriticalPosFound = YES;
  //     break;
  //   } 
  //   if ( isCriticalPosFound ) {
  //     INFO( Log::ACTIVE_OPTIONS, 8, "Gegner gefunden: " +
  //         toCriticalPos.toString() + "\n" );
  //     // Ball auf die dem gegnerischen Spieler abgelegene Seite nehmen
  //     if ( toCriticalPos.angle() > direction ) {
  //       side = RIGHT_SIDE_ID;
  //     }
  //     else {
  //       side = LEFT_SIDE_ID;
  //     }
  //   }
  //   else {
    // auf der Seite dribbeln, auf der der Ball schon liegt
    toCriticalPos = globalWorldModel->ball().pos() - ownPos;
    if ( toCriticalPos.angle() > direction ) {
      side = LEFT_SIDE_ID;
    }
    else {
      side = RIGHT_SIDE_ID;
    }
  // }
  INFO( Log::ACTIVE_OPTIONS, 8, String( "Seite: " ) + 
      ( ( side == LEFT_SIDE_ID ) ? "links\n" : "rechts\n" ) );
  INFO( Log::ACTIVE_OPTIONS, 4, "LEAVE DribbleOption::chooseDribbleSide\n" );
  return side;
}


int DribbleOption::constraintCost( const Constraint *constraint ) const 
{
  int cost;
  switch( constraint->id() ) {
    case CONSERVE_STAMINA_CONSTRAINT_ID:
      cost = MAX_COST;
      break;
    default:
      cost = KickOption::constraintCost( constraint );
  }
  INFO( Log::ACTIVE_OPTIONS, 11, "DribbleOption::constraintCost = " + 
      String( cost ) + "\n" );
  return cost;
}


StorableObject* DribbleOption::copy() const 
{
  return new DribbleOption( *this );
}

String DribbleOption::toString() const  
{
  return String( "( DribbleOption utility: " ) +
    String( theExpectedUtility ) + String( " ) " );
}



 



