
#include <math.h>

#include "Field.h"
#include "Configuration.h"
#include "InvalidIdException.h"

/**
 * Aktueller Stand
 *
 *   $RCSfile: Field.cc,v $
 *   $Revision: 1.16 $
 *   $Date: 1998/08/14 13:20:31 $
 *   $Author: myritz $ 
 *
 */
 
 
const double Field::theFieldLength         = 105;
const double Field::theFieldWidth          = 68;
const double Field::thePenaltyAreaLength   = 16.5;
const double Field::thePenaltyAreaWidth    = 40.32;
double Field::theGoalWidth                 = 7.32;
const double Field::theKickoffCircleRadius = 9.15;

Vector Field::theFlagPositions[] = {
  // UPPER_LEFT_FLAG_ID
  Vector( -Field::theFieldLength / 2, Field::theFieldWidth / 2 ), 
  // UPPER_MIDDLE_FLAG_ID
  Vector( 0, Field::theFieldWidth / 2 ),
  // UPPER_RIGHT_FLAG_ID
  Vector( Field::theFieldLength / 2, Field::theFieldWidth / 2 ),
  // LOWER_LEFT_FLAG_ID
  Vector( -Field::theFieldLength / 2, -Field::theFieldWidth / 2 ),
  // LOWER_MIDDLE_FLAG_ID
  Vector( 0, -Field::theFieldWidth / 2 ),
  // LOWER_RIGHT_FLAG_ID
  Vector( Field::theFieldLength / 2, -Field::theFieldWidth / 2 ),
  // UPPER_LEFT_PENALTY_FLAG_ID
  Vector( -Field::theFieldLength / 2 + Field::thePenaltyAreaLength, 
	  Field::thePenaltyAreaWidth / 2 ),
  // MIDDLE_LEFT_PENALTY_FLAG_ID
  Vector( -Field::theFieldLength / 2 + Field::thePenaltyAreaLength, 0 ),
  // LOWER_LEFT_PENALTY_FLAG_ID
  Vector( -Field::theFieldLength / 2 + Field::thePenaltyAreaLength, 
	  -Field::thePenaltyAreaWidth / 2 ),
  // UPPER_RIGHT_PENALTY_FLAG_ID
  Vector( Field::theFieldLength / 2 - Field::thePenaltyAreaLength, 
	  Field::thePenaltyAreaWidth / 2 ),
  // MIDDLE_RIGHT_PENALTY_FLAG_ID
  Vector( Field::theFieldLength / 2 - Field::thePenaltyAreaLength, 0 ),
  // LOWER_RIGHT_PENALTY_FLAG_ID
  Vector( Field::theFieldLength / 2 - Field::thePenaltyAreaLength, 
	  -Field::thePenaltyAreaWidth / 2 ),
  // LEFT_GOAL_CENTER_FLAG_ID
  Vector( -Field::theFieldLength / 2, 0),
  // RIGHT_GOAL_CENTER_FLAG_ID
  Vector( Field::theFieldLength / 2, 0),

  // CENTER_FLAG_ID
  Vector( 0, 0),
  // LEFT_GOAL_TOP_FLAG_ID
  Vector( -Field::theFieldLength / 2, Field::theGoalWidth / 2 ),
  // LEFT_GOAL_BOTTOM_FLAG_ID
  Vector( -Field::theFieldLength / 2, -Field::theGoalWidth / 2 ),
  // RIGHT_GOAL_TOP_FLAG_ID
  Vector( Field::theFieldLength / 2, Field::theGoalWidth / 2 ),
  // RIGHT_GOAL_BOTTOM_FLAG_ID
  Vector( Field::theFieldLength / 2, -Field::theGoalWidth / 2 ),

  // OUTSIDE_TOP_LEFT_10_FLAG_ID
  Vector( -10, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_LEFT_20_FLAG_ID
  Vector( -20, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_LEFT_30_FLAG_ID
  Vector( -30, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_LEFT_40_FLAG_ID
  Vector( -40, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_LEFT_50_FLAG_ID
  Vector( -50, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_RIGHT_10_FLAG_ID
  Vector( 10, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_RIGHT_20_FLAG_ID
  Vector( 20, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_RIGHT_30_FLAG_ID
  Vector( 30, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_RIGHT_40_FLAG_ID
  Vector( 40, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_RIGHT_50_FLAG_ID
  Vector( 50, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_BOTTOM_LEFT_10_FLAG_ID
  Vector( -10, -Field::theFieldWidth / 2 - 5 ),
  // OUTSIDE_BOTTOM_LEFT_20_FLAG_ID
  Vector( -20, -Field::theFieldWidth / 2 - 5 ),
  // OUTSIDE_BOTTOM_LEFT_30_FLAG_ID
  Vector( -30, -Field::theFieldWidth / 2 - 5 ),
  // OUTSIDE_BOTTOM_LEFT_40_FLAG_ID
  Vector( -40, -Field::theFieldWidth / 2 - 5 ),
  // OUTSIDE_BOTTOM_LEFT_50_FLAG_ID
  Vector( -50, -Field::theFieldWidth / 2 - 5 ),
  // OUTSIDE_BOTTOM_RIGHT_10_FLAG_ID
  Vector( 10, -Field::theFieldWidth / 2 - 5 ),
  // OUTSIDE_BOTTOM_RIGHT_20_FLAG_ID
  Vector( 20, -Field::theFieldWidth / 2 - 5 ),
  // OUTSIDE_BOTTOM_RIGHT_30_FLAG_ID
  Vector( 30, -Field::theFieldWidth / 2 - 5 ),
  // OUTSIDE_BOTTOM_RIGHT_40_FLAG_ID
  Vector( 40, -Field::theFieldWidth / 2 - 5 ),
  // OUTSIDE_BOTTOM_RIGHT_50_FLAG_ID
  Vector( 50, -Field::theFieldWidth / 2 - 5 ),

  // OUTSIDE_LEFT_TOP_10_FLAG_ID
  Vector( -Field::theFieldLength / 2 - 5, 10 ),
  // OUTSIDE_LEFT_TOP_20_FLAG_ID
  Vector( -Field::theFieldLength / 2 - 5, 20 ),
  // OUTSIDE_LEFT_TOP_30_FLAG_ID
  Vector( -Field::theFieldLength / 2 - 5, 30 ),
  // OUTSIDE_LEFT_BOTTOM_10_FLAG_ID
  Vector( -Field::theFieldLength / 2 - 5, -10 ),
  // OUTSIDE_LEFT_BOTTOM_20_FLAG_ID
  Vector( -Field::theFieldLength / 2 - 5, -20 ),
  // OUTSIDE_LEFT_BOTTOM_30_FLAG_ID
  Vector( -Field::theFieldLength / 2 - 5, -30 ),
  // OUTSIDE_RIGHT_TOP_10_FLAG_ID
  Vector( Field::theFieldLength / 2 + 5, 10 ),
  // OUTSIDE_RIGHT_TOP_20_FLAG_ID
  Vector( Field::theFieldLength / 2 + 5, 20 ),
  // OUTSIDE_RIGHT_TOP_30_FLAG_ID
  Vector( Field::theFieldLength / 2 + 5, 30 ),
  // OUTSIDE_RIGHT_BOTTOM_10_FLAG_ID
  Vector( Field::theFieldLength / 2 + 5, -10 ),
  // OUTSIDE_RIGHT_BOTTOM_20_FLAG_ID
  Vector( Field::theFieldLength / 2 + 5, -20 ),
  // OUTSIDE_RIGHT_BOTTOM_30_FLAG_ID
  Vector( Field::theFieldLength / 2 + 5, -30 ),
  // OUTSIDE_TOP_CENTER_FLAG_ID
  Vector( 0, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_BOTTOM_CENTER_FLAG_ID
  Vector( 0, -Field::theFieldWidth / 2 - 5 ),
  // OUTSIDE_LEFT_CENTER_FLAG_ID
  Vector( -Field::theFieldLength / 2 - 5, 0 ),
  // OUTSIDE_RIGHT_CENTER_FLAG_ID
  Vector( Field::theFieldLength / 2 + 5, 0 )
};

const Vector Field::theGoalPositions[] = {
  Vector( Field::theFieldLength / 2, 0 ),
  Vector( -Field::theFieldLength / 2, 0 ),
};

Field::Field() {
  theGoalWidth = globalConf->goalWidth;
  // Neusetzen der von der Torbreite abhaengigen Flaggen
  theFlagPositions[LEFT_GOAL_TOP_FLAG_ID] =
    Vector( -Field::theFieldLength / 2, Field::theGoalWidth / 2 );
  theFlagPositions[LEFT_GOAL_BOTTOM_FLAG_ID] =
    Vector( -Field::theFieldLength / 2, -Field::theGoalWidth / 2 );
  theFlagPositions[RIGHT_GOAL_TOP_FLAG_ID] =
    Vector( Field::theFieldLength / 2, Field::theGoalWidth / 2 );
  theFlagPositions[RIGHT_GOAL_BOTTOM_FLAG_ID] =
    Vector( Field::theFieldLength / 2, -Field::theGoalWidth / 2 );
}

const Vector& Field::posOfPoint(int anId) const
{
  if ( anId >= MIN_FLAG_ID && anId <= MAX_EXISTING_FLAG_ID ) {
    return theFlagPositions[anId - MIN_FLAG_ID];
  }
  if ( anId >= MIN_GOAL_ID && anId <= MAX_EXISTING_GOAL_ID ) {
    return theGoalPositions[anId - MIN_GOAL_ID];
  } 
  else {
    throwException( InvalidIdException(
      "Id of point was neither a flag-id nor a goal-id." ) );
  }
}

bool Field::isInsideField( const Vector& point )
{
  if ( ( fabs( point.x() ) > (double)fieldLength() / 2.0 ) || 
       ( fabs( point.y() ) > fieldWidth() / 2 )  ) {
    return NO;
  }
  else {
    return YES;
  }
}

ViewPoint Field::triangulateViewPoint( const Array<StableObject>& seenLines,
	           const Array<StableObject>& seenFlags ) const
{
  double angle;
  // Aus den gesehenen Linien den absoluten Blickwinkel berechnen
  angle = computeFaceDir( seenLines );

  if ( IS_ACTIVE( Log::WORLD_MODEL, 8 ) ) {
    globalLog->info( "Zahl der gesehenen Linien: " + 
                     String( seenLines.count() ) + "\n" );
    globalLog->info( "Zahl der gesehenen Flaggen: " + 
                     String( seenFlags.count() ) + "\n" );
  }
  INFO( Log::WORLD_MODEL, 7, "Berechneter Blickwinkel: " + 
        String( degFromRadiant( angle ) ) + "\n" );

  int flagCount = seenFlags.count();
  if( flagCount == 0 ) {
    throwException( FailedPositionCalculationException() );
  }

  // Schwerpunkt berechnen von allen Positionen, die jeweils aus einer
  // Flagge und dem Blickwinkel bestimmt werden.
  // Gewichtet wird mit dem Reziproken der Entfernung der Flagge.
  Vector pos;
  Vector average;
  Vector weightedPositionsSum(0, 0);
  double sumOfWeights = 0;
  double weight;
 
  int i;
  for( i = 0; i < flagCount; i++ ) {
    pos = computePosFromFlag( seenFlags[i], angle );
    INFO( Log::WORLD_MODEL, 13, "Absolute Position aus Flagge: " + 
          String( i ) + pos.toString() + "\n" );
    
    weight = 1 / seenFlags[i].pos().length();
    
    weightedPositionsSum += (weight * pos);
    sumOfWeights += weight;
    // Hier folgt auskommentierter Code, der die maximale Abweichung
    // der einzelnen Positionsberechnungen vom Durchschnitt ermittelt.
    //    average = (1/sumOfWeights) * weightedPositionsSum;
    //    
    //    if ( (pos - average).length() > maxDiff.length() ) {
    //      maxDiff = pos - average;
    //      maxDiffId = seenFlags[i].id();
    //      INFO( Log::WORLD_MODEL, 15, "Neue maximale Abweichung: " +
    //            String( maxDiff.length() ) + "m (Flagge " +
    //            String( maxDiffId ) + ")\n" );
    //    }
  }
  
  average = (1/sumOfWeights) * weightedPositionsSum;
  INFO( Log::WORLD_MODEL, 7, "Gewichtetes Mittel: " + average.toString() +
        "\n" );
  //  INFO( Log::WORLD_MODEL, 15, "Neue maximale Abweichung: " + 
  //        String( maxDiff.length() ) + "m (Flagge " + String( maxDiffId ) + 
  //        "\n" );
  
  ViewPoint result(average, angle);
  return result;
}


double Field::computeFaceDir( const Array<StableObject>& seenLines )  
{
  double angle = 0;
  int lineId = 0;
  // zuerst die Linie auswaehlen, deren Winkel ausgewertet wird.
  switch ( seenLines.count() ) {
    case 0:
      WARNING( Log::WORLD_MODEL, 
               "Cannot compute face dir. Number of seen lines is 0." );
      throwException( FailedTriangulationException() );
      break;
    case 1:
      // Position ist innerhalb des Spielfeldes.
      angle = seenLines[0].pos().angle();
      lineId = seenLines[0].id();
      break;
    default:
      // Die zweite Linie wird von innen gesehen, daher
      // kann der Winkel berechnet werden als waere man im inneren
      // des Feldes.
      int i;
      if( seenLines[0].pos().length() > seenLines[1].pos().length() ) {
	i = 0;
      }
      else {
	i = 1;
      }
      angle = seenLines[i].pos().angle();
      lineId = seenLines[i].id();
      break;
  }
  // Jetzt abhaengig von der LinienId die absolute Blickrichtung berechnen 
  INFO( Log::WORLD_MODEL, 10, "Gewaehlte Linie: " + String( lineId ) +
        " Winkel: " + String( degFromRadiant( angle ) ) + ")\n" );
  switch ( lineId ) {
    case UPPER_LINE_ID:
      if ( angle >= 0 ) {
	return PI - angle;
      }
      else {
	return -angle;
      } 
      break;
    case LOWER_LINE_ID:
      if ( angle > 0 ) {
	return 2.0*PI - angle;
      }
      else {
	return PI - angle;
      }
      break;
    case LEFT_LINE_ID:
      if ( angle > 0 ) {
	return 1.5*PI - angle;
      }
      else if ( angle < 0 ) {
	return PI/2 - angle;
      } 
      else {
        WARNING( Log::WORLD_MODEL, 
                 "Couldn't compute face dir. Special case 1 is not treated." );
	throwException( FailedTriangulationException() );
      }
      break;
    case RIGHT_LINE_ID:
      if ( angle > 0 ) {
	return PI/2 - angle;
      }
      else if ( angle < 0 ) {
	return 1.5*PI - angle;
      } 
      else {
        WARNING( Log::WORLD_MODEL, 
                 "Couldn't compute face dir. Special case 2 is not treated." );
	throwException( FailedTriangulationException() );
      }
      break;
    default:
      WARNING( Log::WORLD_MODEL, "Unknown line id in computeFaceDir" );
      throwException( FailedTriangulationException() );
      break;
  }
}

Vector Field::computePosFromFlag( const StableObject& flag,
				  double faceDir ) const
{
  Vector relPos;
  Vector absPos;
  
  relPos.setPolar( flag.pos().length(), faceDir + flag.pos().angle() );
  absPos = posOfPoint( flag.id() ) - relPos;
  
  return absPos;
}
      

  
  
