
/*
 * Aktueller Stand
 *
 *   $RCSfile: GoalKickOption.cc,v $
 *   $Revision: 1.24 $
 *   $Date: 1998/08/14 13:20:38 $
 *   $Author: myritz $ 
 *
 */
 
#include "GoalKickOption.h"
#include "Constraints.h"
#include "WorldModel.h"
#include "Skills.h"
#include "StraightLine.h"

GoalKickOption::GoalKickOption() : KickOption() 
{
  theId = GOAL_KICK_OPTION_ID;
}  


int GoalKickOption::baseUtility() const 
{
  INFO( Log::ACTIVE_OPTIONS, 2, 
      "ENTER GoalKickOption::baseUtility\n" );
  int bestUtility = KickOption::baseUtility();
  if ( bestUtility == MAX_UTILITY ) {
    // Es liegt eine Kick-Situation vor. Jetzt genaue Werte ausrechnen.
    
    // Szenario vorbereiten
    Situation scenario( globalWorldModel->situation() );
    ModelledBall ball;
    ModelledPlayer player = scenario.playerWithId( 
        globalWorldModel->id() );
    // Mich selbst bei der Simulation ignorieren
    scenario.removePlayerWithId (globalWorldModel->id() );
    // Alle Spieler hinter mir rausschmeissen
    int i;
    for ( i = scenario.playerCount() - 1 ; i >= 0; i-- ) {
      if ( scenario.playerWithIndex( i ).pos().x() < player.pos().x() ) {
        scenario.removePlayerWithIndex( i );
      }
    }

    // Wir koennen den Ball kicken.
    // Jetzt drei moegliche Schussrichtungen aufs Tor (linke Ecke, Mitte,
    // rechte Ecke) ausprobieren
    double stepSize = 2.5;
    double x = globalWorldModel->field().fieldLength()/2;
    Vector speed;
    int stepsUntilIntercepted;
    int possibleUtility;
    Vector toPossibleTarget;
    bestUtility = MIN_UTILITY;
    
    for ( double y = -2*stepSize; y <= 2*stepSize; y += stepSize ) {
      Vector possibleTarget( x, y );
      INFO( Log::ACTIVE_OPTIONS, 8, "Scenario mit Schuss zum Ziel " +
          possibleTarget.toString() + "\n" );

      // Ausrechnen, wo der Ball ist, wenn ihn der naechste Spieler erreicht
      ball = globalWorldModel->ball();
      toPossibleTarget = possibleTarget - player.pos();
      double shotLength = toPossibleTarget.length();
      if ( shotLength > globalConf->maxGoalKickDist ) {
        // Das Tor ist zu weit weg.
        INFO( Log::ACTIVE_OPTIONS, 8, "Tor ausser Reichweite (" +
            String( toPossibleTarget.length() ) + "\n" );
        continue;
      }
      // Kollision mit Spieler ist ko-Kriterium
      StraightLine trajectory( player.pos(), possibleTarget );
      bool isCollisionFound = NO;
      for ( i = scenario.playerCount() - 1 ; i >= 0 ; i-- ) {
        if ( trajectory.distTo( scenario.playerWithIndex( i ).pos() ) <
             player.size() + ball.size() + 0.5 ) {
          isCollisionFound = YES;
          INFO( Log::ACTIVE_OPTIONS, 8, 
              "target rejected because of collision\n" );
          break;
        }
      }
      if ( isCollisionFound == YES ) {
        continue;
      }
      
      // angleFactor: 1, wenn der Schuss waagrecht ist, 
      // 0, wenn der Schuss senkrecht ist
      double angleFactor = fabs( normalizeAngle( toPossibleTarget.angle() ) );
      if ( angleFactor > PI/2 ) {
        // Schuss geht nach hinten, das muss Quatsch sein
        continue;
      }
      angleFactor /= ( PI / 2 );
      if ( angleFactor > 0.7 ) {
        // Der Winkel ist einfach zu schlecht
        continue;
      }
      // Jetzt fuer die spaetere Utility-Berechnung anpassen
      angleFactor = 1 - angleFactor*angleFactor;
          
      toPossibleTarget.normalize();
      speed = globalConf->maxEffectiveBallSpeed * toPossibleTarget;
      toPossibleTarget = shotLength * toPossibleTarget;
      ball.setSpeed( speed );
      ball.simulateSteps( globalConf->reactionTimeSteps );
      scenario.setBall( ball );
      stepsUntilIntercepted = scenario.stepsOfNearestPlayer();
      if ( stepsUntilIntercepted > globalConf->maxStepsOfBallMovement ) {
        // Der Ball wird von niemandem erreicht.
        stepsUntilIntercepted = globalConf->maxStepsOfBallMovement;
      }
      INFO( Log::ACTIVE_OPTIONS, 11, "stepsUntilIntercepted: " +
          String( stepsUntilIntercepted ) + "\n" );
      ball.simulateSteps( stepsUntilIntercepted );
      INFO( Log::ACTIVE_OPTIONS, 11, "Position des Balls zu dieser Zeit: " +
          ball.pos().toString() + "\n" );

      // interceptFactor: 1, wenn der Bal	l erst bei Stillstand intercepted
      // wird, 0, wenn der Ball sofort intercepted wird  
      double interceptFactor = stepsUntilIntercepted / 
          globalConf->maxStepsOfBallMovement;
      if ( ball.pos().x() > globalWorldModel->enemyGoal().x() ) {
        interceptFactor = 1;
      }
      // distFactor: 0, wenn der Schuss ueber die maximale Distanz geht,
      // 1, wenn der Schuss eine Laenge von 0 Metern hat
      
      double distFactor = shotLength / globalConf->maxShotLength;
      
      INFO( Log::ACTIVE_OPTIONS, 11, "% of averageShotLength : " + 
          String( distFactor * 100 ) + "\n" );
      distFactor = 1 - min( 1.0, distFactor * distFactor );
      //  centerFactor: 1, wenn der Ball in die Tormitte gezielt ist,
      // 0, wenn der Ball auf einen Pfosten gezielt ist
      double halfGoalWidth = globalWorldModel->field().goalWidth() / 2.0;
      double distToCenter = ( possibleTarget - 
        globalWorldModel->enemyGoal() ).length() / halfGoalWidth;
      double centerFactor;
      // Abschnittsweise grobe Annaeherung an eine Gausskurve.
      if ( distToCenter > 5/7 * halfGoalWidth ) {
        centerFactor = 0.0;
      } else if ( distToCenter > 3/7 * halfGoalWidth ) {
        centerFactor = 1 - ( distToCenter / halfGoalWidth - 3/7 ) * 7/2; 
      } else {
        centerFactor = 1.0;
      }
              
      possibleUtility = (int)( globalConf->maxGoalKickUtility * (
          globalConf->goalKickUtilityWeights[0] * interceptFactor  + 
          globalConf->goalKickUtilityWeights[1] * distFactor + 
          globalConf->goalKickUtilityWeights[2] * centerFactor +
          globalConf->goalKickUtilityWeights[3] * angleFactor ) / 
          globalConf->sumOfGoalKickUtilityWeights  );
      
      if ( IS_ACTIVE( Log::ACTIVE_OPTIONS, 10 ) ) {
        globalLog->info( "interceptFactor: " + String( interceptFactor ) + 
            "\n" );
        globalLog->info( "distFactor: " + String( distFactor ) + "\n" );
        globalLog->info( "centerFactor: " + String( centerFactor ) + "\n" );
        globalLog->info( "angleFactor: " + String( angleFactor ) + "\n" );
      }
      INFO( Log::ACTIVE_OPTIONS, 10, "Utility dieses Schusses: " + 
            String( possibleUtility ) + "\n" );
      
      if ( possibleUtility > bestUtility ) {
        INFO( Log::ACTIVE_OPTIONS, 9, "Schuss angenommen!\n" );
        // Das waere der bisher beste Schuss
        toTarget = possibleTarget - player.pos();
        bestUtility = possibleUtility;
      }
    } // of for
    INFO( Log::ACTIVE_OPTIONS, 5, "Ausgewaehltes Ziel: " +
        ( toTarget + player.pos() ).toString() + "\n" );
  }
  else {
    // nichts machen. Die default-Werte sind ausreichend.
  }
  
  INFO( Log::ACTIVE_OPTIONS, 5, "Utility " + String( bestUtility ) + "\n" );
  return bestUtility;
}


Plan GoalKickOption::generatePlan() const 
{
  INFO( Log::ACTIVE_OPTIONS, 3,  "ENTER KickOption::generatePlan\n" );
  Plan plan;
  if ( isPossible() == NO ) {
    WARNING( Log::ACTIVE_OPTIONS, "Generating Plan for impossible Option.\n" );
  INFO( Log::ACTIVE_OPTIONS, 3,  "LEAVE KickOption::generatePlan\n" );
    return plan;
  }
  double direction;

  INFO( Log::ACTIVE_OPTIONS, 10, 
    "Kick vector: " + toTarget.toString() + "\n" );
  globalSkills->resetSituation();
  direction = adjustedToTarget( toTarget ).angle();

  globalSkills->resetSituation();
  plan.addPlan( globalSkills->setViewWidth( VIEW_WIDTH_NORMAL ) );
  plan.addPlan( globalSkills->kickBall( 100.0, direction ) );
  INFO( Log::ACTIVE_OPTIONS, 3,  "LEAVE KickOption::generatePlan\n" );
  return plan;
}

Vector GoalKickOption::adjustedToTarget( const Vector& toTarget ) const
{
  Vector originalTarget = globalWorldModel->player().pos() + toTarget;
  Plan kickPlan = globalSkills->kickBall( 100.0, toTarget.angle() );
  
  // Den Kickplan ausfuehren
  INFO( Log::ACTIVE_OPTIONS, 11, "Kick plan to target " +
      originalTarget.toString() + " : \n " + kickPlan.toString() + "\n" );
  int actionIndex;
  ModelledBall ball = globalWorldModel->ball();
  ModelledPlayer player = globalWorldModel->player();
  try {
    for ( actionIndex = 0; actionIndex < kickPlan.length(); actionIndex++ ) {
      const Action& currentAction = kickPlan.actionAt( actionIndex );
      if ( currentAction.id() == KICK_ACTION ) {
        ball.accelerate( currentAction.firstParameter(), 
            currentAction.secondParameter(), player );
        ball.simulateSteps();
        player.simulateSteps();
        INFO( Log::ACTIVE_OPTIONS, 11, "Ball after kick action:" + 
            ball.toString() + "\n" );
        INFO( Log::ACTIVE_OPTIONS, 11, "Player->ball:" + 
            ( ball.pos() - player.pos() ).toString() + "\n" );
      }            
    }
  }
  catch ( ModelledBall::KickImpossibleException ) {
    // Ball kann nicht gekickt werden. Dann ist der Kickplan falsch,
    // da kann ich hier auch nichts machen. Ich ignoriere den
    // Kick, da er vom Server auch nicht ausgefuehrt 
    WARNING( Log::ACTIVE_OPTIONS, "Illegal action " + String( actionIndex ) +
        "\n" );
    return originalTarget;
  }
  
  // Ball zum Tor simulieren. 
  int simSteps = 0;
  while ( ball.pos().x() < globalWorldModel->enemyGoal().x() && 
      simSteps < globalConf->maxStepsOfBallMovement ) {
    ball.simulateSteps();
    simSteps++;
  }
  INFO( Log::ACTIVE_OPTIONS, 11, "Ball pos behind goal " + 
      ball.pos().toString() + "\n" );
  
  Vector adjustedTarget( originalTarget.x(), 
      2 * originalTarget.y() - ball.pos().y() );
  
  INFO( Log::ACTIVE_OPTIONS, 5, "Adjusted Target: " + 
      adjustedTarget.toString() + "\n" );
  return ( adjustedTarget - player.pos() );
}
    



StorableObject* GoalKickOption::copy() const 
{
  return new GoalKickOption( *this );
}

String GoalKickOption::toString() const  
{
  return String( "( GoalKickOption utility: " ) +
    String( theExpectedUtility ) + String( " ) " );
}



