
#include "Log.h"
#include "String.h"
#include "Configuration.h"

Log *globalLog;

Log::Log( int debugLevel = 0, ostream& target = cout )
{
  theTargetStream =  &target;
  theModeActivation = Array<int>(MODE_COUNT);
  for( int i = 0; i < MODE_COUNT; i++ ) {
    theModeActivation.addObject( debugLevel );
  }
}

void Log::init( Configuration *aConfigurationObject )
{
  int maxActivation = 0;
  for( int i = 1; i < MODE_COUNT; i++ ) {
    theModeActivation[i] = aConfigurationObject->debugModes[i-1];
    maxActivation = max( maxActivation, theModeActivation[i] );
  }
  theModeActivation[0] = maxActivation;
}
	
bool Log::isActive( int mode, int debugLevel ) 
{
  if( debugLevel <= theModeActivation[mode] ) {
    return YES;
  }
  else {
    return NO;
  }
}

void Log::error( const String& message )
{
  String wholeMessage = "\nError:   " + message + "\n";
  info( wholeMessage );
  cerr << wholeMessage << flush;
}

void Log::warning( const String& message )
{
  info( "\nWarning: " + message + "\n" );
}

void Log::info( const String& message )
{
  *theTargetStream << message << flush;
}

String Log::toString() const
{
  return String( String( "Log Information: \tModi: (" ) +
                 "\nModus: Communicator: " + 
                 String( theModeActivation[COMMUNICATOR] ) +
                 "\nModus: Configuration: " +
                 String( theModeActivation[CONFIGURATION] ) +
                 "\nModus: Sensors: " +
                 String( theModeActivation[SENSORS] ) +
                 "\nModus: Effectors: " +
                 String( theModeActivation[EFFECTORS] ) +
                 "\nModus: Situation: " +
                 String( theModeActivation[SITUATION] ) +
                 "\nModus: WorldModel: " +
                 String( theModeActivation[WORLD_MODEL] ) +
                 "\nModus: Timer: " +
                 String( theModeActivation[TIMER] ) +
                 "\nModus: Deliberator: " +
                 String( theModeActivation[DELIBERATOR] ) +
                 "\nModus: Constraints: " +
                 String( theModeActivation[CONSTRAINTS] ) +
                 "\nModus: PassiveOptions: " +
                 String( theModeActivation[PASSIVE_OPTIONS] ) +
                 "\nModus: ActiveOptions: " +
                 String( theModeActivation[ACTIVE_OPTIONS] ) +
                 "\nModus: ModelledObjects: " +
                 String( theModeActivation[MODELLED_OBJECTS] ) +
                 "\nModus: Skills: " +
                 String( theModeActivation[SKILLS] ) +

                 " )\n" );
}


