#ifndef _MOVINGPLAYER_H_
#define _MOVINGPLAYER_H_

#include "MovingObject.h"

/**
  Ein gesehender Spieler.
  Ein MovingPlayer hat zusaetzlich zu den Eigenschaften eines MovingObject noch
  eine Blickrichtung. Die Blickrichtung gibt an, in welche Richtung der Spieler
  guckt, also auch in welche Richtung er beim naechsten Dash beschleunigen
  wird.
  
  MovingPlayer wird in \Ref{Sensors} eingesetzt, um die wahrgenommenen Spieler
  zu repr\"asentieren.

  {\small {\bf Quelltext}

  [\URL[Header]{src/MovingPlayer.h} | [\URL[Implementierung]{src/MovingPlayer.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: MovingPlayer.h,v $
  *   $Revision: 1.12 $
  *   $Date: 1998/08/14 13:21:00 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class MovingPlayer : public MovingObject
{
public:

  /** @name Erzeugen und Zerst\"oren */
  //@{  
    /** Default-Konstruktor. Alle Werte werden auf undef gesetzt. */
    MovingPlayer();

    /** Neuer MovingPlayer mit Id, Position, Geschwindigkeitu und
      Blickrichtung. Geschwindigkeit und Blickrichtung sind optional.
      Sind sie nicht angegeben, werden die Werte auf undef gesetzt.
      @param id Die Id des Spielers
      @param aPos Die Position.
      @param aSpeed Die Geschwindigkeit.
      @param aFaceDir Die Blickrichtung. */
    MovingPlayer( int id, const Vector& aPos, const Vector& aSpeed = Vector(),
		  double aFaceDir = UNDEF_DOUBLE );
  
    /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
        darauf zurueck. */
    virtual StorableObject* copy() const;
  //@}

  /** @name Werte abfragen */
  //@{
    /** Liefert die Blickrichtung. */
    double faceDir() const;

    /** Beschreibung als String fuer die Ausgabe auf einem Terminal.
      Es werden ID, Position, Geschwindigkeit und FaceDir (in Grad) 
      ausgegeben. */
    virtual String toString() const;
  //@}
  
  /** @name Werte setzen */
  //@{
    /** Setzt die Blickrichtung neu. */
    void setFaceDir( double newFaceDir );
  //@}

    
private:
  /** Die Blickrichtung als absoluter Winkel im Bogenmass. */
  double theFaceDir;
};

inline double MovingPlayer::faceDir() const {
  return theFaceDir;
}

inline void MovingPlayer::setFaceDir(double newFaceDir) {
  theFaceDir = newFaceDir;
}

#endif
