
/*
 * Aktueller Stand
 *
 *   $RCSfile: Option.cc,v $
 *   $Revision: 1.18 $
 *   $Date: 1998/08/14 13:21:01 $
 *   $Author: myritz $ 
 *
 */
 
#include "Option.h"

#include "WorldModel.h"
#include "Skills.h"

Option::Option() : Constraint(), 
  requestedConstraints ( Array<Constraint *>(2) ),
  acceptedConstraints ( Array<Constraint *>(2) )
{
  theId = UNDEF_OPTION_ID;
  setIsOptionPossible( NO );
}  

int Option::computeExpectedUtility() const
{
  if( IS_ACTIVE( Log::PASSIVE_OPTIONS, 2 ) ||
      IS_ACTIVE( Log::ACTIVE_OPTIONS, 2 ) ) {
    globalLog->info( "ENTER Option::computeExpectedUtility\n" );
  }
  int utility;
  utility = baseUtility();
  acceptedConstraints.removeAll();
  int i;
  for ( i = 0; i < requestedConstraints.count(); i++ ) {
    int constraintUtility;
    constraintUtility = requestedConstraints[i]->expectedUtility();
    constraintUtility -= constraintCost( requestedConstraints[i] );
    if( IS_ACTIVE( Log::PASSIVE_OPTIONS, 8 ) ||
        IS_ACTIVE( Log::ACTIVE_OPTIONS, 8 ) ) {
      globalLog->info( requestedConstraints[i]->toString() );
    }
    if ( constraintUtility > 0 ) {
      acceptedConstraints.addObject( requestedConstraints[i] );
      utility += constraintUtility;
      if( IS_ACTIVE( Log::PASSIVE_OPTIONS, 8 ) ||
          IS_ACTIVE( Log::ACTIVE_OPTIONS, 8 ) ) {
        globalLog->info( " -- accepted\n" );
      }
    } 
    else {
      if( IS_ACTIVE( Log::PASSIVE_OPTIONS, 8 ) ||
          IS_ACTIVE( Log::ACTIVE_OPTIONS, 8 ) ) {
        globalLog->info( " -- rejected\n" );
      }

    }
  }
  return utility;
}

int Option::baseUtility() const
{
  return NEUTRAL_UTILITY;
}

int Option::constraintCost( const Constraint *constraint ) const
{
  return NEUTRAL_COST;
}

bool Option::forceContinuation() const
{
  return NO;
}

int Option::cancellationCost() const
{
  return NEUTRAL_COST;
}


bool Option::isPossible() const
{
  return theIsOptionPossible;
}


void Option::adaptToPreviousIntention( const Option *previousIntention )
{
}

Plan Option::generatePlan() const
{
  Plan plan;
  int i;
  const ModelledPlayer& player = globalWorldModel->player();
  
  globalSkills->resetSituation();
  double preferredDirection = globalWorldModel->preferredKickDirection( 
      player.pos() ).angle();
  if ( !isAlmostEqual( player.faceDir(), preferredDirection ) ) {
    plan.addPlan( globalSkills->turnTo( preferredDirection ) );
  }
  for ( i = plan.length(); i < globalConf->maxLengthOfPlan; i++ ) {
    plan.addAction( globalSkills->simpleEmptyAction() );
  }
  return plan;
}

bool Option::isConstraintAccepted( int constraintId ) const
{
  for (int i = 0; i < acceptedConstraints.count(); i++ ) {
    if ( acceptedConstraints[i]->id() == constraintId ) {
      return YES;
    }
  }
  return NO;
}

StorableObject* Option::copy() const 
{
  return new Option( *this );
}

String Option::toString() const 
{
  return String( "( Option utility: " ) +
    String( expectedUtility() ) + String( " ) " );
}



 
