
/*
 * Aktueller Stand
 *
 *   $RCSfile: RasteredPlayground.cc,v $
 *   $Revision: 1.6 $
 *   $Date: 1998/08/14 13:21:07 $
 *   $Author: myritz $ 
 *
 */
 
#include "RasteredPlayground.h"
#include "Configuration.h"
#include "Field.h"
#include "Common.h"
#include "WorldModel.h"
#include "Timer.h"
#include <fstream.h>


RasteredPlayground::RasteredPlayground() 
{
  Colouring::setGlobalDecrement( 0 );

  // Koordinatenursprung (0,0) links unten, nur positive Koordinaten
  // und aufrunden
  rows    = ( int ) ( Field::fieldWidth() / globalConf->yResolution + .5);
  columns = ( int ) ( Field::fieldLength() / globalConf->xResolution + .5);

  halfFieldLength = ( int ) Field::fieldLength() / 2;
  halfFieldWidth = ( int ) Field::fieldWidth() / 2;

  // ein Colouring-Objekt wird rows-mal in eine Column eingefuegt
  Colouring tempColour; 
  ColumnT  tempColumn;

  for( int i = 0; i < rows; i++ ) {
    tempColumn.addObject( tempColour );
  }

  // diese Spalte wird jetzt in alle Spalten dupliziert 
  for ( int i = 0; i < columns; i++ ) {
    colours.addObject( tempColumn );
  }
}


RasteredPlayground::~RasteredPlayground()
{
}


Vector RasteredPlayground::realFromRasteredCoordinates( 
    const Vector& rasteredVector ) 
{
  double realX = ( rasteredVector.x() + 0.5 ) * 
    globalConf->xResolution - halfFieldLength;
  double realY = ( rasteredVector.y() + 0.5 ) * 
    globalConf->yResolution - halfFieldWidth;

  return Vector(realX, realY);
}


Vector RasteredPlayground::rasteredFromRealCoordinates( 
       const Vector &realVector ) 
{
  double rasteredX = ( realVector.x() + halfFieldLength) /
    globalConf->xResolution;
  if ( rasteredX < 0 ) {
    rasteredX = 0;
  }
  if ( rasteredX > columns - 1 ) {
    rasteredX = columns - 1;
  }

  double rasteredY = ( realVector.y() + halfFieldWidth) /
    globalConf->yResolution;
  if ( rasteredY < 0 ) {
    rasteredY = 0;
  }
  if ( rasteredY > rows - 1 ) {
    rasteredY = rows - 1;
  }

  return Vector( floor( rasteredX ), floor( rasteredY ) );
}


Vector RasteredPlayground::positionOnEdge( 
    Vector firstVector, Vector secondVector )
{
  INFO( Log::PASSIVE_OPTIONS, 4, "BEGIN positionOnEdge\n");

  // eine gute passive Position (reelle Koord.)
  Vector thePositionOnEdge; 

  // Position auf den Kanten bestimmen: 
  // wenn beide Schwerpunkte vom eigenen Tor aus hoechstens die Entfernung von
  // "manGuardingBorder" haben, dann
  // im Abstand "distFromOpponent" von demjenigen SP stellen, 
  // der am naechsten dem eigenen Tor ist.
  if ( ( firstVector.x() < ( int ) - halfFieldLength +
         globalConf->manGuardingBorder ) &&
       ( secondVector.x() < ( int ) - halfFieldLength +
         globalConf->manGuardingBorder ) ) {
    // Nach dem naechsten Test ist firstVector der Schwerpunkt des Gegners,
    // der naeher am eigenen Tor steht, sodass dieser Gegner gedeckt
    // wird.
    if ( ( firstVector  - Vector ( - halfFieldLength, 0) ).length() >
    	 ( secondVector - Vector ( - halfFieldLength, 0) ).length() ) {
      Vector tmpVector = firstVector;
      firstVector = secondVector; 
      secondVector = tmpVector;
    }
    // Berechnung der Position fuer die Manndeckung.
    return Vector ( firstVector.x() + 
		    ( double ) globalConf->distFromOpponent /
		    ( secondVector - firstVector ).length() *
		    ( secondVector.x() - firstVector.x() ),
		    firstVector.y() + 
		    ( double ) globalConf->distFromOpponent /
		    ( secondVector - firstVector ).length() *
		    ( secondVector.y() - firstVector.y() ) );
  }
  else {
    // wenn einer der beiden Schwerpunkte weiter von der Torlinie entfernt ist
    // als "manGuardingBorder", dann auf die Mitte der Kante postieren
    return Vector ( ( firstVector.x() + secondVector.x() ) / 2.0, 
		    ( firstVector.y() + secondVector.y() ) / 2.0 ) ;
  }
}


void RasteredPlayground::storePosOfTeammates( Array <Vector> &teammatePos ) 
{
  const Situation& theSituation = globalWorldModel->situation();
  for ( int playerIndex = 0; 
	playerIndex < theSituation.playerCount(); 
	playerIndex++ ) {
    int playerId = theSituation.playerWithIndex( playerIndex ).id();
    if ( playerId >= MIN_TEAMMATE_ID && 
	 playerId <= MAX_EXISTING_TEAMMATE_ID &&
	 playerId != SELF_ID ) {   
      teammatePos.addObject( 
	theSituation.playerWithIndex( playerIndex ).pos() );
      // testTeammate nur fuer den Test
      Vector testTeammate = theSituation.playerWithIndex( playerIndex ).pos();
    }
  }    
}


bool RasteredPlayground::teammateCloserTo( 
    const Vector &thePositionOnEdge, Array <Vector> &teammatePos)
{
  for ( int indexOfTeammate = 0; 
	indexOfTeammate < teammatePos.count();
	indexOfTeammate++ ) {
    // wenn ein Teammate naeher dran ist als der Agent selbst
    if ( ( teammatePos[indexOfTeammate] - thePositionOnEdge ).length()  <
	 ( globalWorldModel->player().pos() - thePositionOnEdge ).length() ) {
      return YES;
    }
  }  
  return NO;
}


Vector RasteredPlayground::goodPassivePosition() 
{
  INFO( Log::PASSIVE_OPTIONS, 4, "BEGIN GoodPassivePosition\n" );

  // abhaengig von der Rolle des Agenten lernt er oder nicht.
  switch ( globalWorldModel->role() ) {
    case LEARNING_FORWARD_ROLE:
    case LEARNING_MIDFIELD_ROLE:
    case LEARNING_DEFENDER_ROLE:
      {
	// Ein Array von guten passiven Positionen (reelle Koord.).
	// Diese werden in computePositions()
	// * berechnet, 
	// * auf ihre Erreichbarkeit geprueft, 
	// * getestet, ob kein eigener Spieler naeher dran ist
	// und in position eingefuegt
	Array <PositionTyp> position;
	
	// Ein Array von Schwerpunkten der Gegner (in Rasterkoordinaten)
	// Diese werden in foundFocuses() berechnet und
	// in computePositions() ausgewertet.
	Array <Vector>  playerFocus;
	
/*
	if( IS_ACTIVE( Log::PASSIVE_OPTIONS, 9 ) ) {
	  globalLog->info( 
	    "aktueller Zustand: " +
	    String( globalWorldModel->gameState() ) + "\n" +
	    "initiale Position: " +
	    globalWorldModel->initialPosition().toString() +
	    "  (real)\n" );
	  Vector rasteredInitial = globalWorldModel->initialPosition(); 
	  globalLog->info( 
	    rasteredFromRealCoordinates( rasteredInitial ).toString() +
	    "  (rastered)  " 
	    "\naktuelle Position: " +
	    globalWorldModel->player().pos().toString() +
	    "  (real)     " );
	  Vector rasteredCurrent = globalWorldModel->player().pos(); 
	  globalLog->info( 
	    rasteredFromRealCoordinates( rasteredCurrent ).toString() +
	    "  (rastered)  " 
	    "\n\nBASIS Positionen: " 
	    "\n  BEFORE_KICK_OFF, OUR_TEAM: " +
	    globalWorldModel->basePosition(
	      WorldModel::BEFORE_KICK_OFF, OUR_TEAM_ID).toString() +
	    "\n  BEFORE_KICK_OFF, THEIR_TEAM: " +
	    globalWorldModel->basePosition(
	      WorldModel::BEFORE_KICK_OFF, THEIR_TEAM_ID).toString() +
	    "\n  PLAY_ON: " +
	    globalWorldModel->basePosition(
	      WorldModel::PLAY_ON, UNDEF_INT).toString() +
	    "\n\nINITIAL Positionen: " 
	    "\n  BEFORE_KICK_OFF, OUR_TEAM: " +
	    globalWorldModel->initialPosition( 
	      WorldModel::BEFORE_KICK_OFF, OUR_TEAM_ID).toString() +
	    "\n  BEFORE_KICK_OFF, THEIR_TEAM: " +
	    globalWorldModel->initialPosition( 
	      WorldModel::BEFORE_KICK_OFF, THEIR_TEAM_ID ).toString() +
	    "\n  PLAY_ON: " +
	    globalWorldModel->initialPosition(
	      WorldModel::PLAY_ON, UNDEF_INT).toString() );
	}
*/
// 	for ( int x = 0; x < columns; x++ ) {
// 	  for ( int y = 0; y < rows; y++ ) {
// 	      colours[x][y].reallyDecPlayerColour();
// 	      colours[x][y].reallyDecBallColour();
// 	  }
// 	}
	Colouring::setGlobalDecrement( 0 );
  
	if ( foundFocuses( playerFocus ) >= 2 && 
	     computePositions( position, playerFocus ) == YES) {    
	  Vector bestPos = bestPosition( position );
	  // beste passive Position ins Weltmodell eintragen...
	  // ... fuer PLAY_ON
	  globalWorldModel->setBasePosition( 
	    WorldModel::PLAY_ON, UNDEF_INT, bestPos ); 
	  // ... fuer eigenen Anstoss
    /*
	  if ( bestPos.x() > -1.0 ) {
	    bestPos = Vector ( -1.0, bestPos.y() );
	  }
	  globalWorldModel->setBasePosition( 
	    WorldModel::BEFORE_KICK_OFF, OUR_TEAM_ID, bestPos );
	  // ... fuer gegnerischen Anstoss 
	  if ( bestPos.length() < Field::kickoffCircleRadius() + 1.0 ) { 
	    bestPos = Vector ( - ( Field::kickoffCircleRadius() + 1.0 ), 
	    bestPos.y() ); 
	  }
	  globalWorldModel->setBasePosition( 
	    WorldModel::BEFORE_KICK_OFF, THEIR_TEAM_ID, bestPos ); 
    */
	  INFO( Log::PASSIVE_OPTIONS, 4,  "\nEnde GoodPassivePosition()\n" );
	  return ( globalWorldModel->basePosition() );
	}
	else {
	  INFO( Log::PASSIVE_OPTIONS, 4,  "\nEnde GoodPassivePosition()\n" );
	  // alte passive Position zurueckgeben
	  return ( globalWorldModel->initialPosition() );
	}
	break;
      }
    default:
      return ( globalWorldModel->basePosition()  );
      break;
  }
}


int RasteredPlayground::foundFocuses( Array <Vector> & playerFocus ) 
{
  // das argument playerFocus ist in Rasterkoordinaten

  // Alle Rasterfelder werden darauf hin geprueft, ob die Player-Faerbung
  // groesser als ein gewisser Schwellwert ist. Wenn das der Fall ist, 
  // werden alle diese SPs in das Array playerFocus eingefuegt.
  for( int x = 0; x < columns; x++ ) {
    for( int y = 0; y < rows; y++ ) {
      if ( colours[x][y].playerColour() >= globalConf->threshold ) {
	playerFocus.addObject( Vector( x, y ) );
      }
    }
  }
  return ( playerFocus.count() );
}


int RasteredPlayground::getBallAppearance ( Vector firstVector, 
    Vector secondVector ) 
{
  // die Argumente firstVector und secondVector sind Rasterkoordinaten
  int theBallAppearance = 0;
  // Anstieg der Kante
  double rise;
  if ( firstVector.x() != secondVector.x() ) {
    rise = ( secondVector.y() - firstVector.y() ) /
      ( secondVector.x() - firstVector.x() );
  } 
  else {
    // "unendlich" grosser Anstieg, wenn die x-Koord. gleich sind.
    rise = INT_MAX;
  }
  // wenn Anstieg der Kante > 1 oder x-Koord. gleich  
  if ( ( abs( rise ) > 1 ) ) { 
    //  sicherstellen, dass y-Koord. des 1.SP < y-Koord. des 2.SP
    if ( firstVector.y() > secondVector.y() ) {
      Vector tmpVector = firstVector; 
      firstVector = secondVector; 
      secondVector = tmpVector;
    }
    //  wenn x-Koord. der Schwerpunkte gleich
    //  durchlaufe die y-Koordinate des Rasters vom 1. SP bis zum 2. SP
    //  summiere die Ballfaerbung entlang dieser Kante auf.
    if ( firstVector.x() == secondVector.x() ) {
      for ( int y = ( int ) firstVector.y(); 
	    y <= ( int ) secondVector.y(); 
	    y++ ) {
	theBallAppearance +=
	  colours[ ( int ) firstVector.x() ] [y].ballColour();
      }
    }
    else {
      //   durchlaufe die y-Koordinate des Rasters vom 1. SP bis zum 2. SP
      //   berechne die x-Koordinaten des Ball-Vektors 
      //   zwischen den beiden SPs, 
      //   summiere die Ballfaerbung entlang dieser Kante auf.
      for ( int y = ( int ) firstVector.y();
	    y <= ( int ) secondVector.y(); 
	    y++ ) {
	double xReal = firstVector.x() + 1 / rise * ( y - firstVector.y() );
	theBallAppearance += colours[(int) xReal] [y].ballColour();
	// auch Nachbar-Zellen werden mitbetrachtet
	if ( ( int ) (xReal) != ( int ) ceil ( xReal ) ) {
	  theBallAppearance +=
	    colours[(int) ceil (xReal)][y].ballColour();
	}
      }
    }
  }
  else {
    //  wenn der Anstieg der Kante <= 1 
    //  sicherstellen, dass x-Koord. des 1. SP < x-Koord. des 2. SP
    if ( firstVector.x() > secondVector.x() ) {
      Vector tmpVector = firstVector; 
      firstVector = secondVector; 
      secondVector = tmpVector;
    }
    // durchlaufe die x-Koordinate des Rasters vom 1. SP bis zum 2. SP,
    // berechne die y-Koordinaten des Ball-Vektors 
    // zwischen den beiden SPs, 
    // summiere die Ballfaerbung entlang dieser Kante auf.
    for ( int x = ( int ) firstVector.x(); 
	  x <= ( int ) secondVector.x(); 
	  x++ ) {
      double yReal = firstVector.y() + rise * ( x - firstVector.x() );
      theBallAppearance += colours[x][(int) yReal].ballColour();
      // auch Nachbar-Zellen werden mitbetrachtet
      if ( ( int ) (yReal) != ( int ) ceil ( yReal ) ) {
	theBallAppearance += colours[x][(int) ceil ( yReal )].ballColour();
      }
    }
  }
  return theBallAppearance;
}

 
bool RasteredPlayground::computePositions( 
  Array <PositionTyp> & position, Array <Vector> & playerFocus ) 
{
  INFO( Log::PASSIVE_OPTIONS, 4, "\n\nStart computePositions\n" );
  // die Positionen aller Mitspieler (in reellen Koord.)
  Array <Vector>  teammatePos;
  storePosOfTeammates( teammatePos );
  
  // Es werden alle Paare von Schwerpunkten betrachtet
  for ( int indexOfFirstFocus = 0; 
	indexOfFirstFocus < playerFocus.count() - 1; 
	indexOfFirstFocus++ ) {      
    for ( int indexOfSecondFocus = indexOfFirstFocus + 1; 
	  indexOfSecondFocus < playerFocus.count();
	  indexOfSecondFocus++ ) {
      Vector firstFocus = 
	realFromRasteredCoordinates( playerFocus[indexOfFirstFocus] );
      Vector secondFocus = 
	realFromRasteredCoordinates( playerFocus[indexOfSecondFocus] );
      // falls Kante nicht zu kurz und nicht zu lang
      if ( ( ( secondFocus - firstFocus ).length() >= 
	     globalConf->minEdgeLength ) &&
	   ( ( secondFocus - firstFocus ).length() <= 
	     globalConf->maxEdgeLength ) ) {
	// Hier wird die moegliche passive Position auf der Kante
	// bestimmt.
	Vector thePositionOnEdge = positionOnEdge( firstFocus, secondFocus ); 
	// falls die berechnete Position zu weit von der initialen Position
	if ( ( ( globalWorldModel->initialPosition() - 
		 thePositionOnEdge ).length() <
	       globalConf->maxDistFromHome ) && 
	     !teammateCloserTo ( thePositionOnEdge, teammatePos ) ) {
	  PositionTyp positionElem;
	  // Die Spielfeldkoord. (reelle Koord.) der passiven Pos.
	  positionElem.thePosition = thePositionOnEdge;
	  // Ballhaeufigkeit zwischen den Rasterkoord. der Schwerpunkte
	  positionElem.ballAppearance = getBallAppearance 
	    ( rasteredFromRealCoordinates( firstFocus ),
	      rasteredFromRealCoordinates( secondFocus ) );
	  
	  // Faerbung der Kante bzgl. Gegner
	  // ( = arithm. Mittel (Faerbung von SP1, SP2) )	      
	  positionElem.edgeColour = ( int )
	    ( colours[( int ) playerFocus[indexOfFirstFocus].x()]
	      [( int ) playerFocus[indexOfFirstFocus].y()].
	      playerColour() + 
	      colours[( int ) playerFocus[indexOfSecondFocus].x()]
	      [( int ) playerFocus[indexOfSecondFocus].y()].
	      playerColour() ) / 2;
	  position.addObject( positionElem );
	}
      }
    }
  }
  return( position.count() > 0 ? YES : NO);
}


Vector RasteredPlayground::bestPosition( Array <PositionTyp> & position ) 
{
  INFO( Log::PASSIVE_OPTIONS, 4, "\n\nStart bestPosition()\n" );

  int distFromCurrentPos, distFromHomePos;
  int indexOfBestPos = 0, totalWeight, previousTotalWeight = INT_MIN;
  Vector realPos, realBestPos;
  
  if( IS_ACTIVE( Log::PASSIVE_OPTIONS, 9 ) ) {
    globalLog->info( 
      "\n\nhome-Position des Spielers: " +
      globalWorldModel->initialPosition( 
	globalWorldModel->gameState(), 
	globalWorldModel->hasInitiative( OUR_TEAM_ID ) ).toString() +
      "\n\naktuelle Position des Spielers: " +
      globalWorldModel->player().pos().toString() + 
      "\n" 
      "Bewertung gefundener Positionen:\n" );
  }
  // Bestimmung der besten Position ueber gewichtete Summe
  
  for ( int indexOfPos = 0; 
	indexOfPos < position.count(); 
	indexOfPos++ ) {
    if ( globalWorldModel->role() ==  LEARNING_DEFENDER_ROLE ) {
      totalWeight = - ( int ) ( position[indexOfPos].thePosition - 
			Vector ( - halfFieldLength, 0) ).length();
    }
    else {
      distFromHomePos = ( int )
	( globalWorldModel->initialPosition()  -
	  position[indexOfPos].thePosition ).length();
      distFromCurrentPos = 
        ( int ) ( globalWorldModel->player().pos() - 
		  position[indexOfPos].thePosition ).length();
      totalWeight = ( int )
        ( ( globalConf->weightEdgeColour * position[indexOfPos].edgeColour ) +
	  ( globalConf->weightBallAppearance * 
	    position[indexOfPos].ballAppearance )  - 
	  ( globalConf->weightDistFromHome  * distFromHomePos ) -
	  ( globalConf->weightDistFromCurrent * distFromCurrentPos ) );
    }
    if ( totalWeight > previousTotalWeight ) {
      previousTotalWeight = totalWeight;
      indexOfBestPos = indexOfPos;
    }      
    if( IS_ACTIVE( Log::PASSIVE_OPTIONS, 9 ) ) {
      globalLog->info(
	"Position Nr. " +
	String( indexOfPos ) +
	": " +
	position[indexOfPos].thePosition.toString() +
	"\ndistFromHomePos : " +
	String( distFromHomePos ) +
	"\ndistFromCurrentPos : " +
	String( distFromCurrentPos ) +
	"\nballAppearance : " +
	String( position[indexOfPos].ballAppearance ) +
	"\nedgeColour : " +
	String( position[indexOfPos].edgeColour ) +
	"\ntotalWeight : " +
	String( totalWeight ) +
	"\n" );
    }
  }
    
  if( IS_ACTIVE( Log::PASSIVE_OPTIONS, 9 ) ) {
    globalLog->info(
      "\n\nbeste Position: " 
      "\nNr. " +
      String( indexOfBestPos ) +
      "\n" +
      position[indexOfBestPos].thePosition.toString() +
      "\ndistFromHome : " +
      String( ( int ) ( globalWorldModel->initialPosition( 
	globalWorldModel->gameState(),
	globalWorldModel->hasInitiative(OUR_TEAM_ID) )  -
	position[indexOfBestPos].thePosition ).length() ) +
      "\ndistFromCurrent : " +
      String( ( int ) ( globalWorldModel->player().pos() - 
	position[indexOfBestPos].thePosition ).length() ) +
      "\nFarbe der Kante : " +
      String( position[indexOfBestPos].edgeColour ) +
      "\nBallhaeufigkeit auf der Kante : " +
      String( position[indexOfBestPos].ballAppearance ) +
      "\nprevioustotalWeight : " +
      String( previousTotalWeight ) +
      "\n\nEnde bestPosition()\n" );
  }

  return ( position[indexOfBestPos].thePosition );
}


void RasteredPlayground::acquireKnowledgeAbout( const int aboutId ) 
{  
  const Situation& theSituation = globalWorldModel->situation(); 
  // aboutId ist PLAYER_ID (lerne Spieler-Positionen) 
  //       oder BALL_ID (lerne Ball-Positionen)
  switch( aboutId ) {
    case PLAYER_ID:
      // nur fuer sichtbare Spieler der gegnerischen Mannschaft 
      // mit bekannter Nummer in der aktuellen Situation wird gelernt:
      for( int playerIndex = 0;
	   playerIndex < theSituation.playerCount();
	   playerIndex++ ) {
	int playerId = theSituation.playerWithIndex( playerIndex ).id();
	// falls ein Gegner mit bekannter nummer:
	if( playerId >= MIN_OPPONENT_ID &&
	    playerId <= MAX_EXISTING_OPPONENT_ID &&
	    ( theSituation.playerWithIndex( 
	      playerIndex ).posReliability().get() <= Reliability::GOOD ) ) {
	  const Vector& playerPos = 
	      theSituation.playerWithIndex( playerIndex ).pos();
	  // und entsprechendes Rasterfeld faerben 
	  // (dabei Spielfeldkoordinaten in Rasterkoordinaten umrechnen)
	  // Der Torwartbereich ( 6 ) wird nicht mitgelernt
	  if ( ( playerPos.x() > - halfFieldLength ) &&
	       ( playerPos.x() <   halfFieldLength - 
		 globalConf->learnNoGoalieDist ) &&  
	       ( playerPos.y() > - halfFieldWidth )  &&
	       ( playerPos.y() <   halfFieldWidth ) ) {
	    int indexPosX = ( int ) 
	      ( rasteredFromRealCoordinates( playerPos ).x() );
	    int indexPosY = ( int ) 
	      ( rasteredFromRealCoordinates( playerPos ).y() );
	    colours[indexPosX][indexPosY].incPlayerColour();
	  }
	}
      }
      break;
    case BALL_ID: 
      { /* this is a seperate block to scope ballPos */
	// position ermitteln
	const Vector& ballPos = theSituation.ball().pos();
	// und entsprechendes Rasterfeld faerben
	// (dabei Spielfeldkoordinaten in Rasterkoordinaten umrechnen)
	if ( ( ballPos.x() > - halfFieldLength ) &&
	     ( ballPos.x() <   halfFieldLength ) &&  
	     ( ballPos.y() > - halfFieldWidth )  &&
	     ( ballPos.y() <   halfFieldWidth ) ) {
	  int indexPosX = ( int ) 
	    ( rasteredFromRealCoordinates( ballPos ).x() );
	  int indexPosY = ( int ) 
	    ( rasteredFromRealCoordinates( ballPos ).y() );
	  colours[indexPosX][indexPosY].incBallColour(); 
	}
      }
      break;
    default:
      // do nothing
      break;
}

  Colouring::decrementAllColours();
  for ( int x = 0; x < columns; x++ ) {
    for ( int y = 0; y < rows; y++ ) {
      colours[x][y].reallyDecPlayerColour();
      colours[x][y].reallyDecBallColour();
    }
  }
  Colouring::setGlobalDecrement( 0 );

  if( IS_ACTIVE( Log::PASSIVE_OPTIONS, 9 ) ) {
    globalLog->info(
      "\n\ngefaerbtes Feld:" 
      "\nfuer Gegner:\n\n     " );
    for ( int x = 0; x < columns; x++ ) {
      globalLog->info( String( x ) );
      if ( x < 10 ) {
	globalLog->info( "  " );
      }
      else {
	globalLog->info( " " );
      }
    }
    globalLog->info( "\n" ); 
    for ( int y = rows-1; y >= 0; y-- ) {
      globalLog->info( "\n" + String( y ) );
      if ( y < 10 ) {
	globalLog->info( "    " );
      }
      else {
	globalLog->info( "   " );
      }
      for ( int x = 0; x < columns; x++ ) {
	globalLog->info( String( colours [x][y].playerColour() ) );
        if (colours[x][y].playerColour() < 10 ) {
          globalLog->info( "  " );
	}
        else {
          globalLog->info( " " );
	}
      }
    }
    globalLog->info( "\nfuer Ball:\n\n     " );
    for ( int x = 0;x < columns; x++ ) {
      globalLog->info( String( x ) );
      if ( x < 10 ) {
	globalLog->info( "  " );
      }
      else {
	globalLog->info( " " );
      }
    }
    globalLog->info( "\n" ); 
    for ( int y = rows-1; y >=0; y-- ) {
      globalLog->info( "\n" );
      globalLog->info( String( y ) );
      if ( y < 10 ) {
	globalLog->info( "    " );
      }
      else {
	globalLog->info( "   " );
      }
      for ( int x = 0; x < columns; x++ ) {
	globalLog->info( String( colours [x][y]. ballColour() ) );
	if ( colours[x][y].ballColour() < 10 ) {
	  globalLog->info( "  " );
	}
	else {
	  globalLog->info( " " );
	}
      }
    }
  }
  INFO ( Log::PASSIVE_OPTIONS, 4, "\nEnde AquireKnowledgeAbout\n" );
}
