
/*
 * Aktueller Stand
 *
 *   $RCSfile: Skills.cc,v $
 *   $Revision: 1.46 $
 *   $Date: 1998/08/14 13:21:16 $
 *   $Author: myritz $ 
 *
 */
 
#include "Skills.h"
#include "Configuration.h"

Skills *globalSkills;

Skills::Skills() {
  minKickRadius = globalConf->playerSize + globalConf->ballSize;
  maxKickRadius = globalConf->kickRange;
  optimalKickRadius = minKickRadius + globalConf->kickMargin * 
    globalConf->optimalKickDist;
  maxSpeedOfPlayer = globalConf->playerSpeedMax;
  maxSpeedOfBall = globalConf->ballSpeedMax;
  maxLengthOfPlan = globalConf->maxLengthOfPlan;
  kickPowerRate = globalConf->kickPowerRate;
  outerKickCircleTolerance = globalConf->outerKickCircleTolerance;
  innerKickCircleTolerance = globalConf->innerKickCircleTolerance;
  intermediateKickTolerance = globalConf->intermediateKickTolerance;
  resetSituation();
}


//
// beamToPosition
//

Plan Skills::beamToPosition( const Vector& position )
{
  Plan plan = Plan(maxLengthOfPlan);
  plan.addAction( simpleMoveTo( position ) );
  return plan;
}


//
// gotoRegion und Hilfsroutinen
//

Plan Skills::gotoRegion( 
    const Vector& centerOfTarget, 
    double targetRadius, 
    const double staminaLimit, 
    const int specialBehaviors,
    int desiredMaxLengthOfPlan )
{
  bool avoidObstacles = ( specialBehaviors & AVOID_OBSTACLES ) ? YES : NO;
  bool runBackwards = ( specialBehaviors & RUN_BACKWARDS ) ? YES : NO;
  INFO( Log::SKILLS, 3, "ENTER Skills::gotoRegion\n" );
  INFO( Log::SKILLS, 8, "at simStep " + String( situation().simStep() ) + 
      "\n" );
  INFO( Log::SKILLS, 10, "centerOfTarget: " + centerOfTarget.toString() +
      "\ntargetRadius: " + String( targetRadius ) +
      "\nstaminaLimit: " + String( staminaLimit ) +
      "\navoidObstacles: " + ( (avoidObstacles==YES)?"YES":"NO" ) + 
      "\nrunBackwards: " + ( (runBackwards==YES)?"YES":"NO" ) + "\n" );
 
  if( desiredMaxLengthOfPlan == UNDEF_INT ) {
    desiredMaxLengthOfPlan = maxLengthOfPlan;
  }
 
  static const double INTERMEDIATE_RADIUS = TRIANGULATION_ACCURACY;
  Plan plan = Plan( desiredMaxLengthOfPlan );
  double radius;
  if( targetRadius < TRIANGULATION_ACCURACY ) {
    targetRadius = TRIANGULATION_ACCURACY;
  }

  double distToIntermediateTarget;
  double deviationFromTarget;
  double staminaLeft;
  Vector intermediateTarget;
  if ( avoidObstacles == YES ) {
    radius = INTERMEDIATE_RADIUS;
    // Hindernisse umlaufen: Parameter einstellen
    globalSkills->initEmptySectorSearch( centerOfTarget, 
        radFromDegree( 7 ), 6.0 , 0.75*PI, radFromDegree( 20.0 ) );
  }
  else {
    intermediateTarget = centerOfTarget;
    radius = targetRadius;
  }
  
  while( plan.length() < desiredMaxLengthOfPlan ) {
    INFO( Log::SKILLS, 8, "Berechne Aktion " + String( plan.length()+1 ) 
        + "\n" );
    
    // Zielbestimmung
    if ( avoidObstacles == YES ) {
      intermediateTarget = findEmptySector( situation() ) + player().pos();
      INFO( Log::SKILLS, 8, "Neues Zwischenziel: " + 
          intermediateTarget.toString() + "\n" );
      if ( intermediateTarget == centerOfTarget) {
        radius = targetRadius;
        INFO( Log::SKILLS, 8, "Neues Zwischenziel ist entgueltiges Ziel\n" );
      }
    }
    distToIntermediateTarget = player().distTo( intermediateTarget );
    INFO( Log::SKILLS, 10, "distToIntermediateTarget: " +
        String( distToIntermediateTarget ) + "\n" );

    INFO( Log::SKILLS, 10, "distToCenterOfTarget: " +
        String( player().distTo( centerOfTarget ) ) + "\n" );
    if ( player().distTo( centerOfTarget ) <= targetRadius ) {
      INFO( Log::SKILLS, 8, "Am endgueltigen Ziel angekommen\n" );
      break;
    }
    
    // Bestimmung der Drehnotwendigkeit
    staminaLeft = player().stamina() - staminaLimit;
    INFO( Log::SKILLS, 9, "staminaLeft: " + String( staminaLeft ) + "\n" );
    
    deviationFromTarget = player().angleTo( intermediateTarget );
    if ( runBackwards == YES ) {
      deviationFromTarget = normalizeAngle( deviationFromTarget + PI );
    }
    INFO( Log::SKILLS, 9, "deviationFromTarget: " + 
        String( degFromRadiant( deviationFromTarget ) ) + "\n" );
    if( fabs( deviationFromTarget ) >= radFromDegree( 10.0 ) ||
      staminaLeft < 0 ) { 
      // drehen
      INFO( Log::SKILLS, 8, "Es wird gedreht" );
      if ( staminaLeft < 0 ) {
        INFO( Log::SKILLS, 8, "( weil stamina zu niedrig ist )\n");
      }
      else {
        INFO( Log::SKILLS, 8, "( Richtungsabweichung zu hoch ist )\n");
      }
      
      Plan turnPlan = turnBy( deviationFromTarget );
      if ( turnPlan.length() == 0 ) {
        // turn nicht notwendig, weil Richtung bereits genau stimmt
        plan.addAction( simpleEmptyAction() );
      }
      else {
        plan.addPlan( turnPlan );
      }
    }
    else { // beschleunigen
      INFO( Log::SKILLS, 8, "Es muss gedasht werden\n" );
      double power = player().maxEffectiveDashPower();
      power = player().dashPowerForSend( power );
      power = min( staminaLeft, power );
      if ( runBackwards == YES ) {
        power =  player().normalizeDashPower( -power );
      } 
      plan.addAction( simpleDash( power ) );
    }
  }
  INFO( Log::SKILLS, 7, "gotoRegionPlan: " + plan.toString() + "\n" );
  INFO( Log::SKILLS, 3, "LEAVE Skills::gotoRegion\n" );
  
  return plan;
}

void Skills::initEmptySectorSearch( 
    const Vector& aFinalTarget,
    const double aSectorWidth,
    const double aSectorLength,
    const double aMaxDeviation,
    const double aStepAngle )
{
  theFinalTarget = aFinalTarget;
  theSectorAngle = aSectorWidth / 2;
  theSectorLength = aSectorLength;
  theMaxDeviation = aMaxDeviation;
  theStepAngle = aStepAngle;
}

Vector Skills::findEmptySector( const Situation& aSituation ) const
{
  INFO( Log::SKILLS, 4, "ENTER Skills::findEmptySector\n" );
  INFO( Log::SKILLS, 8, "Szenario vorbereiten\n" );
  setComparisonAccuracy( 0.02 ); 
  Situation scenario( aSituation );
  ModelledPlayer unmodifiedPlayer = scenario.playerWithId( 
      globalWorldModel->id() );
  Vector toTarget( theFinalTarget - unmodifiedPlayer.pos() );
  // mich selbst und alle Spieler, die weit weg sind, rausnehmen
  INFO( Log::SKILLS, 10, "uebergebene Situation:\n" + aSituation.toString() +
      "\n" );
  scenario.removePlayerWithId( globalWorldModel->id() );
  INFO( Log::SKILLS, 10, "Entferne Spieler " + String(
      globalWorldModel->id() ) + " (mich selbst)\n" );
  double relativeAngle;
  Vector otherRelPos;
  double effectiveSectorLength = min( toTarget.length(), sectorLength() );
  int i;
  INFO( Log::SKILLS, 12, 
      "finalTarget: " + finalTarget().toString() + 
      "\nsectorAngle: " + String( degFromRadiant( sectorAngle() ) ) + 
      "\nmaxDeviation: " + String( degFromRadiant( maxDeviation() ) ) + 
      "\neffectiveSectorLength: " + String( effectiveSectorLength ) + 
      "\nstepAngle: " + String( degFromRadiant( stepAngle() ) ) + "\n" );

  for ( i = scenario.playerCount() - 1; i >= 0; i-- ) {
    INFO( Log::SKILLS, 15, "Situation vor Checken von Spieler " +
        String( i ) + "(" + String( scenario.playerWithIndex( i ).id() ) +
        ") : " + scenario.toString() + "\n" );
    otherRelPos = scenario.playerWithIndex( i ).pos() - 
        unmodifiedPlayer.pos();
    if ( otherRelPos.length() > effectiveSectorLength ) {
      INFO( Log::SKILLS, 10, "Entferne Spieler " + String(
          scenario.playerWithIndex( i ).id() ) + " (zu weit weg: " +
          String( otherRelPos.length() ) + ")\n" );
      scenario.removePlayerWithIndex( i );
    } 
    else {
      relativeAngle = normalizeAngle( otherRelPos.angle() - 
          toTarget.angle() );
      if ( fabs( relativeAngle ) > ( PI/2 + sectorAngle() * 2 ) ) {
        INFO( Log::SKILLS, 10, "Entferne Spieler" + String(
            scenario.playerWithIndex( i ).id() ) + " (zu weit rechts/links: " +
            String( degFromRadiant( relativeAngle ) ) + ")\n" );
        scenario.removePlayerWithIndex( i );
      }
    }
  }
  
  INFO( Log::SKILLS, 9, "Szenario: " + scenario.toString() + "\n" );
  double turnAngle = 0.0; // Um diesen Winkel wird jeweils weitergedreht
  double sectorDirection = toTarget.angle();
  double adjustedPlayerRadius = globalConf->playerSize *2.1;
  double shiftLength = adjustedPlayerRadius / sin( sectorAngle() );
  ModelledPlayer player;
  Vector shift;
  double foundAngle = UNDEF_DOUBLE;
  bool isSectorBlocked;

  // Zuerst die Sektoren links von der Vorzugsrichtung probieren
  INFO( Log::SKILLS, 8, "Winkel links durchgehen\n" );
  while ( fabs( turnAngle )  < maxDeviation() ) {
    player = unmodifiedPlayer;
    INFO( Log::SKILLS, 13, "Meine Pos: " + player.pos().toString() +
        "\n" );
    // Pos nach hinten verschieben, damit im Sektor die Spielerbreiten
    // ignoriert werden koennen
    shift = Vector( shiftLength, sectorDirection + PI, POLAR );
    INFO( Log::SKILLS, 11, "Getestete Richtung: " + 
        String( degFromRadiant( sectorDirection ) ) + "\n" );
    INFO( Log::SKILLS, 12, "Shift: " + shift.toString() + "\n" );
    player.setPos( player.pos() + shift );
    INFO( Log::SKILLS, 12, "Geshiftete Pos: " + 
        player.pos().toString() + "\n" );
    isSectorBlocked = NO;
    for ( i = scenario.playerCount() - 1; i >= 0; i-- ) {
      INFO( Log::SKILLS, 9, "Checke Spieler " + 
          String( scenario.playerWithIndex( i ).id() ) + "\n" );
      otherRelPos = scenario.playerWithIndex( i ).pos() - player.pos();

      relativeAngle = normalizeAngle( otherRelPos.angle() -
          sectorDirection );
      INFO( Log::SKILLS, 11, "Winkel von shiftedPos aus: " + 
          String( degFromRadiant( relativeAngle ) ) + "\n" );
      if ( ( fabs( relativeAngle ) < 
          sectorAngle() ) && otherRelPos.length() > shiftLength ) {
        // Es steht jemand in diesem Sektor. 
        // (Beim Vergleich muss der Spieler mehr als shiftLength weg sein,
        //  sonst steht er in Wirklichkeit hinter uns.)
        INFO( Log::SKILLS, 11, "Sektor blockiert.\n" );
        turnAngle += stepAngle();
        INFO( Log::SKILLS, 11, "Neuer Drehwinkel:" + 
            String( degFromRadiant( turnAngle ) ) + "\n" );

        sectorDirection = toTarget.angle() + turnAngle;
        isSectorBlocked = YES;
        break;
      }
    } // of for
    if ( isSectorBlocked == NO ) {
      foundAngle = turnAngle;
      INFO( Log::SKILLS, 11, "Winkel gefunden: " + 
          String( degFromRadiant( turnAngle ) ) + "\n" );
      break; // while verlassen
    }
  } // of while

  // Jetzt die andere Seite. Hier muss man von neuem alle Spieler
  // beruecksichtigen, aber nur soweit testen, bis die neue Richtung
  // keinesfalls besser ist als die bisher schon gefundene.
  INFO( Log::SKILLS, 8, "Winkel rechts durchgehen\n" );
  turnAngle = -stepAngle();
  sectorDirection = toTarget.angle();
  while ( fabs( turnAngle ) < fabs( foundAngle ) &&
      ( fabs( turnAngle ) < maxDeviation() ) ) {
    player = unmodifiedPlayer;
    INFO( Log::SKILLS, 13, "Meine Pos: " + player.pos().toString() +
        "\n" );
    // Pos nach hinten verschieben, damit im Sektor die Spielerbreiten
    // ignoriert werden koennen
    shift = Vector( shiftLength, sectorDirection + PI, POLAR );
    INFO( Log::SKILLS, 11, "Getestete Richtung: " + 
        String( degFromRadiant( sectorDirection ) ) + "\n" );
    player.setPos( player.pos() + shift );
    INFO( Log::SKILLS, 12, "Shift: " + shift.toString() + "\n" );
    INFO( Log::SKILLS, 12, "Geshiftete Pos: " + 
        player.pos().toString() + "\n" );
    isSectorBlocked = NO;
    for ( i = scenario.playerCount() - 1; i >= 0; i-- ) {
      INFO( Log::SKILLS, 9, "Checke Spieler " + 
          String( scenario.playerWithIndex( i ).id() ) + "\n" );
      otherRelPos = scenario.playerWithIndex( i ).pos() - player.pos();
      double relativeAngle = normalizeAngle( otherRelPos.angle() - 
        sectorDirection );
      INFO( Log::SKILLS, 11, "Winkel von shiftedPos aus: " + 
          String( degFromRadiant( relativeAngle ) ) + "\n" );
      INFO( Log::SKILLS, 11, "Entfernung von shiftedPos aus: " + 
          String( otherRelPos.length() ) + "\n" );

      if ( fabs( relativeAngle ) < 
          sectorAngle() && otherRelPos.length() > shiftLength ) {
        // Es steht jemand in diesem Sektor.
        // (Beim Vergleich muss der Spieler mehr als shiftLength weg sein,
        //  sonst steht er in wirklichkeit hinter uns.)
        INFO( Log::SKILLS, 11, "Sektor blockiert.\n" );
        turnAngle -= stepAngle();
        INFO( Log::SKILLS, 11, "Neuer Drehwinkel:" + 
            String( degFromRadiant( turnAngle ) ) + "\n" );
        sectorDirection = toTarget.angle() + turnAngle;
        isSectorBlocked = YES;
        break;
      }            
    } // of for
    if ( isSectorBlocked == NO ) {
      foundAngle = turnAngle;
      INFO( Log::SKILLS, 11, "Winkel gefunden: " + 
          String( degFromRadiant( turnAngle ) ) + "\n" );
      break; // while verlassen
    }
  } // of while
  if ( foundAngle==UNDEF_DOUBLE ) {
    // Es wurde kein Winkel zum Ausweichen .
    // Spieler versucht direky in Richtung ziel zu laufen
    foundAngle=0.0;
    INFO( Log::SKILLS, 8, "kein leerer Sektor gefunden\n");  
  }
  sectorDirection = toTarget.angle() + foundAngle;
  toTarget.set( effectiveSectorLength, sectorDirection, POLAR );
  INFO( Log::SKILLS, 5, "Entscheidung:Sektor-Richtung: " +
      String( degFromRadiant( sectorDirection ) ) +
      "\nAbweichung von der Idealrichtung: "  +
      String( degFromRadiant( foundAngle ) ) +
      "\nRelatives Ziel: " + toTarget.toString() + "\n" );
  return toTarget;
}


//
// kickBall und Hilfsroutinen
//

Vector Skills::getPassVector( const Vector& desiredKick, 
                              const ModelledBall& ball, 
                              const ModelledPlayer& player ) const
{
  return ( desiredKick - ( ball.pos() - player.pos() ) );
}

bool Skills::checkBodyContact( const Vector& desiredKick, 
    const ModelledBall& cBall, 
    const ModelledPlayer& cPlayer ) const
{

  ModelledBall ball = cBall;
  ModelledPlayer player = cPlayer;
  
  ball.accelerate( desiredKick.length(), desiredKick.angle(), player );
  ball.simulateSteps(1);
  player.simulateSteps(1);
 
  /* wenn im naechsten Schritt Ball und Spieler Kontakt haetten,
     wird YES zurueckgeliefert */ 
  if ( ( ( ball.pos() - player.pos() ).length() ) < 
      (minKickRadius + innerKickCircleTolerance ) ) {   
    return YES;  
  }
  else {
    return NO;
  }
}

Vector Skills::bestKickVector( const Vector& desiredKick,
    const Vector& calculatedKick,
    const ModelledBall& aBall, 
    const ModelledPlayer& aPlayer ) const
{
  double alpha;
  double beta;
  double length;

  ModelledBall ball = aBall;
  ModelledPlayer player = aPlayer;
  player.simulateSteps(1);
  Vector oldDistFromPlayer = ( player.pos() - ball.pos() );
  ball.accelerate( calculatedKick.length(), calculatedKick.angle(), player );
  ball.simulateSteps(1);

  if( oldDistFromPlayer.length() > minKickRadius + innerKickCircleTolerance ) {
    /* betrachte den Ziel-Vektor als Tangente an den Kreis (Spieler) mit dem
       Radius minKickRadius, berechne mit Sinussatz den Winkel zum
       alten Vektor, alpha(+) / beta(-)  */

       alpha =  desiredKick.angle() +
                (asin( ( minKickRadius  + innerKickCircleTolerance ) / 
          oldDistFromPlayer.length())); 
    
       beta  =  desiredKick.angle() -
                (asin( ( minKickRadius + innerKickCircleTolerance ) /
          oldDistFromPlayer.length())); 
  
       length = minKickRadius;
  }
  /* Kollision im naechsten Schritt */
  else {
       alpha = oldDistFromPlayer.angle() + (PI / 2);
       beta  = oldDistFromPlayer.angle() - (PI / 2);
       length = minKickRadius;
  }

  Vector new1 = Vector( length, alpha, POLAR, RADIAN );
  Vector new2 = Vector( length, beta, POLAR, RADIAN );

  /* den neuen Vektor mit der geringeren Abweichung zur vorher berechneten
     Richtung zurueckgeben */
  if ( ( ( ( aBall.pos() + new1 ) - ball.pos() ).length() ) <
       ( ( ( aBall.pos() + new2 ) - ball.pos() ).length() ) ) {
    new1 = getKickVector( new1 , aBall, aPlayer );
    return new1;
  }
  else {
    new2 = getKickVector( new2 , aBall, aPlayer );
    return new2;
  }  
}

Vector Skills::getKickVector( const Vector& desiredKick, 
    const ModelledBall& ball, 
    const ModelledPlayer& player ) const
{

  double absMaxEKPower = ( ball.maxEffectiveKickPower( player ) * 
                           kickPowerRate );
  Vector targetVector;

  if ( ( desiredKick - ball.speed() ).length() <= 
         absMaxEKPower ) {
    targetVector = desiredKick - ball.speed();
  }
  else {
    /* die maximale Beschleunigung berechnen, Schnittpunkte der
       Gerade durch den Zielvektor mit dem Kreis umd den Geschwindig-
       keitsvektor des Balls mit dem Radius der maximalen Kickstaerke;
       t = targetvector (gew"unschter Zielvektor),
       s = speedvector (Geschwindigkeit des Balles),
       m = Anstieg der Gerade t (targetVektor) ( t.y() / t.x()),
       n = konstantes Glied der Geradengleichung nach Verschiebung des
           Koordinatenursprunges um s (speedVektor)
       Geradengleichung + Vektor zum Kreismittelpunkt = s:
       y = m * x + (-s.y() + m * s.x()); 
       r=sqrt(x0^2+y0^2), 
       x:=x0=s.x(), y:=y0=s.y() ; Ziel: ax^2+bx+c */
    Vector s  = ball.speed();
    Vector t  = desiredKick;
    double m  = ( t.y() / t.x() );
    double n = ( -s.y() + ( m * s.x() ) );
    double r  = absMaxEKPower;

    double a  = ( ( m * m ) + 1);
    double b  = ( 2 * m * n );
    double c  = ( ( n * n ) - ( r * r ) );
    double d = ( ( b * b ) - ( 4 * a * c ) );
    
    if ( d > 0 ) {
      /* quadratische Gleichung mit zwei Loesungen */
      double x1 = ( ( ( -b ) +  sqrt( d ) ) / ( 2 * a ) );
      double y1 = ( m * x1 + n );
      double x2 = ( ( ( -b ) -  sqrt( d ) ) / ( 2 * a ) );
      double y2 = ( m * x2 + n);
      
      /* waehle Loesung die weniger abweicht */
      if ( ( desiredKick - Vector( x1, y1, CARTESIAN ) ).length() < 
           ( desiredKick - Vector( x2, y2, CARTESIAN ) ).length() ) {
        /* ermittle den Vektor zwischen Ball.speed und Loesung */
        
        targetVector = ( ( Vector( x1, y1, CARTESIAN ) ) );
      }
      else { 
        
        targetVector = ( ( Vector( x2, y2, CARTESIAN ) ) );
      }
    }

    else if (d == 0) {
      /* quadratische Gleichung mit einer Loesung */
      double x1 = ( (-b) / ( 2 * a ) );
      double y1 = ( m * x1 + n );      

      targetVector = ( ( Vector( x1, y1, CARTESIAN ) ) ); 
    }    

    else {  // d < 0
      /* ohne Loesung - Exeption??? */

      targetVector =  Vector( absMaxEKPower, ( ball.speed().angle() - PI ), 
                              POLAR, RADIAN );
    }    
  }
  return targetVector;
}

Vector Skills::getKickToRadius( const Vector& desiredKick, 
                                const ModelledBall& aBall, 
                                const ModelledPlayer& aPlayer ) const
{
  Vector targetVector;
  ModelledPlayer player = aPlayer;
  player.simulateSteps();

  Vector s  = ( player.pos() - aBall.pos() );
  Vector t  = desiredKick;
  double m  = ( t.y() / t.x() );
  double n = ( -s.y() + ( m * s.x() ) );
  double r  = ( maxKickRadius - outerKickCircleTolerance );

  double a  = ( ( m * m ) + 1);
  double b  = ( 2 * m * n );
  double c  = ( ( n * n ) - ( r * r ) );
  double d = ( ( b * b ) - ( 4 * a * c ) );

  if ( d >= 0 ) {
    /* quadratische Gleichung mit zwei Loesungen */
    double x1 = ( ( ( -b ) +  sqrt( d ) ) / ( 2 * a ) );
    double y1 = ( m * x1 + n );
    double x2 = ( ( ( -b ) -  sqrt( d ) ) / ( 2 * a ) );
    double y2 = ( m * x2 + n);

    /* waehle Loesung die weniger abweicht */
    if ( ( desiredKick - ( Vector( x1, y1, CARTESIAN ) + s ) ).length() < 
        ( desiredKick - ( Vector( x2, y2, CARTESIAN ) + s ) ).length() ) {
      /* ermittle den Vektor zwischen Ball.speed und Loesung */

      targetVector = ( Vector( x1, y1, CARTESIAN ) + s);
    }
    else {
      targetVector = ( Vector( x2, y2, CARTESIAN ) + s);
    }
  }
  else {
    targetVector = desiredKick;
  }
  return targetVector;
}

Vector Skills::makeRelativeKick( const Vector& desiredKick, 
                                 const ModelledBall& ball, 
                                 const ModelledPlayer& player ) const
{
  return Vector( ball.kickPowerForSend( ( desiredKick.length() /
                 kickPowerRate ), player ), ( desiredKick.angle() -
                 player.faceDir() ), POLAR, RADIAN );
}


Plan Skills::kickBall( double power, const double direction )
{
  INFO( Log::SKILLS, 3, "ENTER Skills::kickBall\n" );
  INFO( Log::SKILLS, 10, "Desired power: " + String( power ) + "\n" );
  INFO( Log::SKILLS, 10, "Desired direction: " + String( 
      degFromRadiant( direction ) ) + "\n" );
  Plan plan = Plan();
  
  Vector kickVector;
  int actionCount = 0;

  if ( power > globalConf->maxPower ) {
    power = globalConf->maxPower;
  }
  double kickAngle = direction;
  Vector realKick = Vector( ( power * ( maxSpeedOfBall / 100 ) ), kickAngle, 
                              POLAR, RADIAN );
  Vector intermediateKick;  
  Vector desiredKick;
  INFO( Log::SKILLS, 10, "Resulting realKick: " + 
      realKick.toString() + "\n" );

  INFO( Log::SKILLS, 11, "distToBall: " + 
      String( player().distTo( ball().pos() ) ) + "\n" ); 
  while ( ( player().distTo( ball().pos() )  < maxKickRadius ) &&
        ( actionCount <= maxLengthOfPlan ) ) {
    // Der Ball ist wahrscheinlich noch innerhalb des Radius
    // Zur Sicherheit wird ein etwas groesserer Radius angenommen
      
    intermediateKick = getKickToRadius( realKick, ball(), player() );
    INFO( Log::SKILLS, 10, "intermediateKick: " + 
        intermediateKick.toString() + "\n" );

    /* "Uberpr"ufe, ob es sinnvoll ist, einen Zwischenkick auszuf"uhren,
       wenn nicht, Kicke so gut es geht, sonst teste ob ein zweiter 
       Schritt m"oglich ist.
       Anmerkung: der zweite Schritt ist in Verbindung mit der while-Schleife
       nicht ganz sauber, mathematisch laesst sich eventuell auch eine bessere
       L"osung fiden, aber i.A. werden nicht mehr als zwei Kicks ausgef"uhrt. 
       */
    if ( ( intermediateKick.length() < intermediateKickTolerance ) ||
              intermediateKick.length() < ball().speed().length() ) {
      desiredKick = realKick;
      kickVector = getKickVector( desiredKick, ball(), player() );
      kickVector = makeRelativeKick( kickVector, ball(), player() );
    }
    else {
      desiredKick =  intermediateKick;
      kickVector = getKickVector( desiredKick, ball(), player() );

      if( ( desiredKick - ( ball().speed() + kickVector ) ).length() >
          intermediateKickTolerance ) {
        desiredKick = ( ( 1 / ( 1 + globalConf->ballDecay ) ) * 
            ( desiredKick - ( ball().speed() + kickVector ) ) );
        kickVector = getKickVector( desiredKick, ball(), player() );        
      }
      kickVector = makeRelativeKick( kickVector, ball(), player() );      
    }
    if ( ( checkBodyContact( kickVector, ball(), player() ) ) == YES ) {
      INFO( Log::SKILLS, 10, "Ball makes body contact.\n" );
      kickVector = bestKickVector( desiredKick, kickVector, ball(), 
          player() );
      kickVector = makeRelativeKick( kickVector, ball(), player() );      
          INFO( Log::SKILLS, 10, "New kickVector: " + kickVector.toString() + 
              "\n" );
    }
    
    plan.addAction( simpleKick( kickVector.length(), kickVector.angle() ) );
    actionCount++;
  }
  INFO( Log::SKILLS, 7, "kickPlan: " + plan.toString() + "\n" );
  INFO( Log::SKILLS, 3, "LEAVE Skills::kickBall\n" );

  return plan;
}


//
// dribble und Hilfsroutinen
//

Plan Skills::trivialDribble( double staminaLimit, const double direction, 
    const int side )
{
  INFO( Log::SKILLS, 3, "ENTER Skills::trivialDribble\n" );
  Plan plan = Plan();
  double staminaLeft = player().stamina() - staminaLimit;
  if ( staminaLeft > 30 ) {
    plan.addPlan( kickBall( 25, direction ) );
  }
  INFO( Log::SKILLS, 7, "dribblePlan: " + plan.toString() + "\n" );
  INFO( Log::SKILLS, 3, "LEAVE Skills::trivialDribble\n" );
  return plan;
}

// berechnet den Kickvector aus dem WegVector des balls (weg in steps Schritten)
Vector Skills::calcKickToRelPos(  const Vector& aRelativeVector, 
                                 const ModelledBall& aBall, 
                                 const ModelledPlayer& aPlayer,   
                                 int steps) const 
{
  double sum = 1.0;

  double ballDecay=globalConf->ballDecay;

  Vector kickVector;
  if (steps==0) { /* throwException*/ steps=1;};
  for ( int j = 1; j < steps; j++) {
    sum += ballDecay;
    ballDecay *= ballDecay;
  }
  
  kickVector = 1/sum * aRelativeVector;
  //INFO( Log::SKILLS,15, "kickvectorformove angepasster kickvecktor: " + kickVector.toString()+ "\n" );
  //return kickVector;
  if ( aPlayer.distTo( aBall.pos()) > maxKickRadius) {
    cout << "cannot kick ball ( out of range ): " << aPlayer.pos() << aBall.pos() << endl;
  }

  kickVector = getKickVector( kickVector, aBall, aPlayer );
    
  return makeRelativeKick( kickVector, aBall, aPlayer);
}



bool Skills::checkBallInDribbleKickRange( const Vector& desiredKick, 
    const ModelledBall& cBall, 
    const ModelledPlayer& cPlayer, int side ) const
{

  ModelledBall ball = cBall;
  ModelledPlayer player = cPlayer;
  if ( desiredKick.length() > 0.0) {
    ball.accelerate( desiredKick.length(), desiredKick.angle(), player );
  }
  ball.simulateSteps(1);
  player.simulateSteps(1);
 
  // wenn im naechsten Schritt der Ball den Kickbereich nicht verlaesst und vor  uns liegt,
  //   wird YES zurueckgeliefert 
  if ( ( ( player.pos() ).length() ) < ( maxKickRadius) &&
       isAlmostEqual( ball.speed().angle(), normalizeAngle( player.faceDir() ) ) )
    {
    return YES;  
  }
  else {
    return NO;
  }
}


bool Skills::checkBallInKickRange( const Vector& desiredKick, 
    const ModelledBall& cBall, 
    const ModelledPlayer& cPlayer ) const
{

  ModelledBall ball = cBall;
  ModelledPlayer player = cPlayer;
  if ( desiredKick.length() > 0.0) {
    ball.accelerate( desiredKick.length(), desiredKick.angle(), player );
  }
  ball.simulateSteps(1);
  player.simulateSteps(1);
 
  // wenn im naechsten Schritt der Ball den Kickbereich nicht verlaesst und vor  uns liegt,
  //     wird YES zurueckgeliefert 
  if ( player.distTo( ball.pos() )  < ( maxKickRadius) ) {
    return YES;  
  }
  else {
    return NO;
  }
}

Plan Skills::dribble( double staminaLimit, const double direction, 
    const int side ) 
{ 
  Vector realKickVector, kickVector, ballToPlayerVector, desiredBallPos;

  Plan plan = Plan();
  ModelledPlayer dribblePlayer = player();
  ModelledPlayer playerAtKickPos;
  ModelledBall dribbleBall = ball();
  double dashPower, maxTurnAngle, sideFactor = (side == LEFT_SIDE_ID)? 1:-1;
  double deltaKickAngle;
  ModelledBall ballAtKickPos=dribbleBall;
  bool isBodyContact;
 
  INFO( Log::SKILLS, 3, "ENTER Skills::dribble\n" );
  /* calcDashPower */
  do {
    double staminaLeft = player().stamina() - staminaLimit + globalConf->playerStaminaInc;
    if (staminaLeft > 30.0 ) {
      dashPower = min( staminaLeft/3 , 100.0) ;
    } else {
      dashPower = 0.0;
    }
    INFO( Log::SKILLS, 15, "Berechneter DashPower :" + String( dashPower) + "\n");
    INFO( Log::SKILLS, 15, "Entfernung Ball - Spieler" + String( ( ball().pos() - player().pos() ).length() ) + "\n" );
    
    while ( fabs( normalizeAngle( player().faceDir()) - direction) > PI/18 ) {            
      // Fall 1 : Dribbeln nicht in Blickrichtung

      dribblePlayer = player();
      dribbleBall = ball();
      
      INFO( Log::SKILLS, 11, "\nnicht in Blickrichtung dribbeln\n");
      INFO( Log::SKILLS, 15, "normalisierte Blickrichtung :" + String( normalizeAngle( player().faceDir()))   + "\n"); 
      INFO( Log::SKILLS, 15, "Direction  :" + String(direction)   + "\n"); 
      INFO( Log::SKILLS, 15, "ball  :" + ball().toString()   + "\n");
      INFO( Log::SKILLS, 15, "player  :" + player().toString()   + "\n");
      INFO( Log::SKILLS, 15, "Entfernung Ball - Spieler" + String( ( ball().pos() - player().pos() ).length() ) + "\n\n" );

      maxTurnAngle = direction - normalizeAngle( player().faceDir());
      maxTurnAngle = signum( maxTurnAngle)* min( fabs(maxTurnAngle), player().maxEffectiveTurnAngle()); 
      
      dribblePlayer.turn( player().turnAngleForSend( maxTurnAngle) );
    
      if( checkBallInKickRange( Vector(0,0), ball(), dribblePlayer)){
        
        INFO( Log::SKILLS, 11, "mache nur einfachen Turn : " + String( player().turnAngleForSend( maxTurnAngle) ) + "\n");
        plan.addAction( simpleTurn( player().turnAngleForSend( maxTurnAngle) ) );
      }
      else {
        INFO( Log::SKILLS, 11, "nur Turn nicht moeglich stoppe ball ab \n");
        
        if ( player().distTo( ball().pos() ) > maxKickRadius ) {
          WARNING( Log::SKILLS, " can't kick Ball ( out Of Range ) \n");
        }
        
        realKickVector = makeRelativeKick( Vector( ball().speed().length(), 
                                                   normalizeAngle( PI + ball().speed().angle() ), POLAR, RADIAN ) + player().speed(), 
                                           ball(), player() );
        if ( realKickVector.length() < 5 ) {
          INFO( Log::SKILLS, 15, "Hole den Ball etwas naeher ran \n");
          realKickVector = realKickVector + 1/( 4 * kickPowerRate) * ( player().pos() - ball().pos() );
        } 
        INFO( Log::SKILLS, 15, "kicke Ball mit " + realKickVector.toString() + "\n");
        plan.addAction( simpleKick( realKickVector.length(), realKickVector.angle() ) );
      }
    }

    dribblePlayer = player();
    dribbleBall = ball();
    ballAtKickPos = dribbleBall;
    playerAtKickPos = dribblePlayer;


    INFO( Log::SKILLS, 11, "\n Weiter dribbeln in Blickrichtung \n");
    INFO( Log::SKILLS, 15, " ball  :\n" + ball().toString()   + "\n");
    INFO( Log::SKILLS, 15, " player  :\n" + player().toString()   + "\n");
    INFO( Log::SKILLS, 15, " Entfernung spieler ball  :" + String( player().distTo( ball().pos() ) )   + "\n");

    dribblePlayer.accelerate( player().dashPowerForSend( dashPower) );
    dribblePlayer.simulateSteps();// nur dash
    dribbleBall.simulateSteps();// simulierter ball nach dem dash
    /* ball ist immernoch vor player*/
    // nach dem dash liegt der ball noch vor dem spieler
    INFO( Log::SKILLS , 15, " Winkel des Balls zur BlickRichtung :" + String( dribblePlayer.angleTo( dribbleBall.pos())) + "\n\n");

    if ( fabs( dribblePlayer.angleTo( dribbleBall.pos())) < PI/2 && 
         //Ball ist noch in Kickrange
         dribblePlayer.distTo( dribbleBall.pos() ) < maxKickRadius  &&
         // keine Kollision
         dribblePlayer.distTo( dribbleBall.pos() ) > minKickRadius &&
         dashPower > 0.0 ) {
      
      ballToPlayerVector = dribblePlayer.pos() - dribbleBall.pos();
      INFO( Log::SKILLS , 11,"dash reicht \n " );
      INFO( Log::SKILLS, 15," Ballposition: " + dribbleBall.toString() + "\n Spielerposition  : " + 
            dribblePlayer.toString() + "\n");
      plan.addAction( simpleDash( player().dashPowerForSend( dashPower ) ) );
    } else {

      dribblePlayer = player();
      dribblePlayer.simulateSteps();//kick
      dribblePlayer.accelerate( dribblePlayer.dashPowerForSend( dashPower) );
      dribblePlayer.simulateSteps();//dash

      ballToPlayerVector = dribblePlayer.pos() - ball().pos();
      // ball soll immer genau links oder rechts von uns ankommen

      desiredBallPos = Vector( minKickRadius + 0.4, 
                               normalizeAngle( dribblePlayer.faceDir() + 
                                               sideFactor * PI /2), POLAR, RADIAN );
      

      INFO(Log::SKILLS,15,"Desired ballposition : " + desiredBallPos.toString() + "\n");

      kickVector = desiredBallPos + ballToPlayerVector;

      INFO(Log::SKILLS,15,"kickVector : " + kickVector.toString() + "\n");
      realKickVector = calcKickToRelPos( kickVector, ball(), player(), 2);
      INFO(Log::SKILLS,15,"realkickVector : " + realKickVector.toString() + "\n");

      if ( checkBodyContact( realKickVector, ball(), player() ) ) {
        // BodyContact nach Kick
        INFO(Log::SKILLS,11,"BodyContact nach Kick \n");
        // auf richtige Seite ist abhaengig vom RealKickVector und 
        // Beschleunigung nach vorn von eigener Geschwindigkeit
        kickVector = Vector( 0.0, realKickVector.y() ) + 
      1/kickPowerRate * player().speed(); 
        INFO( Log::SKILLS, 15, "berechneter Kick:" + kickVector.toString() + "\n" );

        if ( checkBodyContact( kickVector, ball(), player() ) ) {
          double turnFactor = 0.0;
          deltaKickAngle = DELTA_KICK_ANGLE;
          do {
            do {
              INFO( Log::SKILLS, 15,"TURNFACTOR " + String( turnFactor) + " \n");
              kickVector.setPolar( player().distTo( ball().pos() ) / kickPowerRate , 
                                   normalizeAngle( realKickVector.angle() 
                                                   + sideFactor * turnFactor), 
                                   RADIAN);
              INFO( Log::SKILLS, 15, "KickVecotr: "+ kickVector.toString() + "\n");  
              turnFactor += deltaKickAngle;
              isBodyContact = checkBodyContact( kickVector, ball(), player() );
            }while  ( isBodyContact && fabs( kickVector.angle()) < PI/2);
            deltaKickAngle = - DELTA_KICK_ANGLE;
            if ( turnFactor > 0) {
              turnFactor = deltaKickAngle;
            }
          }while ( isBodyContact && fabs( turnFactor) < PI);
          INFO( Log::SKILLS, 11,"Kicke um mich herum \n");
          INFO( Log::SKILLS, 15, "Kickvector:" + kickVector.toString() + "\n" );
                    
          kickVector.setPolar( min( kickVector.length(), 50.0), 
        kickVector.angle(), RADIAN );

          while ( !checkBallInKickRange( kickVector, ball(), player() ) && 
        kickVector.length() > 10.0 ) {
            kickVector.setPolar( 0.7 * kickVector.length(), kickVector.angle(), 
          RADIAN);
            INFO( Log::SKILLS, 15, "Kick zu stark. Neuer Kick:" + 
          kickVector.toString() + "\n" );
          }
          plan.addAction( simpleKick( kickVector.length() , kickVector.angle() )  );
        }
        else {
          INFO( Log::SKILLS, 11, " Versuche auf die richtige Seite zu spielen \n");

          kickVector.setPolar( min( kickVector.length(), 100.0), 
        kickVector.angle(), RADIAN );
          //  INFO( Log::SKILLS, 15, "kick zu stark neuer Kick:" + kickVector.toString() + "\n" );

          while ( !checkBallInKickRange( kickVector, ball(), player() )
        && kickVector.length() > 10.0 ) {
            kickVector.setPolar( 0.5 * kickVector.length(), kickVector.angle(), 
          RADIAN ) ;
            INFO( Log::SKILLS, 15, "Kick zu stark. Neuer Kick:" + 
          kickVector.toString() + "\n" );
          }
          plan.addAction( simpleKick( kickVector.length(), kickVector.angle() ) );
        }
        // kann zu ballverlust fuehren
      }
      else if ( checkBallInKickRange( realKickVector, ball(), player() ) ) {
        // im Ballbesitz nach Kick
        INFO(Log::SKILLS,11,"Ich kicke \n");
        INFO( Log::SKILLS, 15,"  mit: " + realKickVector.toString() + "\n");
        plan.addAction( simpleKick( realKickVector.length(), 
                                    realKickVector.angle() ) );
        if ( checkBodyContact( Vector( 0,0), ball(), player() ) ) {
          // BodyContact nach dash -> nicht dashen
          INFO(Log::SKILLS,11,"BodyContact nach Dash: ich laufe nicht hinterher \n");
        }
        else if ( checkBallInKickRange ( Vector( 0,0), ball(), player() ) ) {
          // im Ballbesitz nach dash
          INFO( Log::SKILLS,11,"ich laufe\n" );
          INFO( Log::SKILLS, 15, "  mit: " + String( player().dashPowerForSend( dashPower ) ) + " \n");
          plan.addAction( simpleDash( player().dashPowerForSend( dashPower ) ) );
        }
        else {
          // kein Ballbesitz mehr nach dash -> nicht oder weniger oder mehr dashen?
          if ( fabs( player().angleTo( ball().pos() ) ) < PI/4 ) {
            INFO( Log::SKILLS, 11, "Ballverlust nach dash, aber Ball vor mir dashe maximal hinterher");
            if ( staminaLeft > 0 ) {
              dashPower = min( staminaLeft,100);
              plan.addAction( simpleDash( player().dashPowerForSend( dashPower ) ) );
            }
          }
          else {
            INFO(Log::SKILLS,11,"Ballverlust nach dash: ich laufe lieber nicht \n");
          }
        }
      } 
      else {
        // kein Ballbesitz mehr nach kick  -> weniger kicken
        while ( !checkBallInKickRange( realKickVector, ball(), player() )
            && kickVector.length() > 10.0 ) {
          realKickVector.setPolar( realKickVector.length() * 0.8, 
        realKickVector.angle(), RADIAN );
        }
        INFO(Log::SKILLS,11, "Ballverlust nach kick. \nIch kicke nur mit: " + 
            String( realKickVector.length()/2 ) + "\n");
        plan.addAction( simpleKick( realKickVector.length()/2, 
            realKickVector.angle() ) );
      } 
    }
    
    INFO( Log::SKILLS , 15, "player - ball nach Ausfuehrung : " + 
        String((player().pos() - ball().pos()).length()) + "\n");
    
    if ( player().distTo( ball().pos() ) > maxKickRadius) {
      INFO( Log::SKILLS , 11, " habe den Ball verloren \n");
      break;
    }
  } while ( plan.length() < MIN_DRIBBLE_PLAN_LENGTH );
  
  INFO( Log::SKILLS, 7, "dribblePlan: " + plan.toString() + "\n" );
  INFO( Log::SKILLS, 3, "LEAVE Skills::dribble\n" );
  return plan;
}

//
// rufen
//


Plan Skills::sayWorldModel()  
{
  INFO( Log::SKILLS, 4, "ENTER Skills::sayWorldModel\n" );
  int ourMessageId = globalConf->ourMessageIdentifier;
  Plan plan = Plan();
  const Situation& scenario = situation();
  const ModelledBall& ball = scenario.ball();

  // Ausgabe :
  // " ourMessageId ourMessageIdEncoded simStepEncoded 
  //   { p ballPosX ballPosY ballPosReliability }
  //   { s ballSpeedX ballSpeedY ballSpeedReliability }
  //   { + playerID { o playerPosX playerPosY playerPosReliability}
  //                { t stamina staminaReliability } }* -

  INFO( Log::SKILLS, 11, "Scenario to say: " + scenario.toString() + "\n" );

  String message;

  char headerInfoString[256] = "";
  char ballPosString[64] = "";
  char ballSpeedString[64] = "";
  sprintf( headerInfoString, "%i %s %s ", ourMessageId, 
           globalWorldModel->encode( ourMessageId ).cString(),
           globalWorldModel->encode(scenario.simStep()).cString());

  if( ball.posReliability().get() <= 9 ) {
    sprintf( ballPosString, "p %.2f %.2f %li ", ball.pos().x(),
             ball.pos().y(), ball.posReliability().get() );
  }
  if( ball.speedReliability().get() <= 9) {
    sprintf( ballSpeedString, "s %.2f %.2f %li ", ball.speed().x(), 
      ball.speed().y(), ball.speedReliability().get());
  }

  message = headerInfoString + String( ballPosString ) + ballSpeedString;

  for (int i = 0; i < scenario.playerCount(); i++) {
    const ModelledPlayer& player = scenario.playerWithIndex( i );
    if( player.isPlayerNumberKnown() ) {
      char playerInfoString[256] = "+ ";
      char playerIdString[64] = "";
      char playerPosString[64] = "";
      char playerStaminaString[64] = ""; 
      sprintf( playerIdString, "%i ", player.id());
      strcat( playerInfoString, playerIdString );
      
      if( ( player.posReliability().get() <= 9 ) ) {
        sprintf(playerPosString, " o %.0f %.0f %li ",
		player.pos().x(), player.pos().y(), 
		player.posReliability().get());
        strcat( playerInfoString, playerPosString );
      }      
      
      if( ( player.team() == OUR_TEAM_ID ) && 
	  ( player.staminaReliability().get() <= 9 ) ) {
        sprintf(playerStaminaString, " t %.0f %li ",
                player.stamina(), player.staminaReliability().get());
        strcat( playerInfoString, playerStaminaString);
      }
      message = message + playerInfoString;
    }
  }           
  message = message + "-";
  
  plan.addAction( simpleSay( message ) );
  INFO( Log::SKILLS, 8, "spoken message: " + message + "\n");

  return plan;
}


//
// turn 
//

Plan Skills::turnBy( double turnAngle )
{  
  INFO( Log::SKILLS, 4, "ENTER Skills::turnBy\n" );
  Plan plan = Plan();
  turnAngle = player().normalizeTurnAngle( turnAngle );
  double maxAngle = player().maxEffectiveTurnAngle();
  INFO( Log::SKILLS, 10, "maxAngle: " + String( degFromRadiant( maxAngle ) ) +
      "\n" );
  
  while( fabs(turnAngle) > maxAngle ) {
    double simpleTurnAngle = signum(turnAngle) * maxAngle;
    INFO( Log::SKILLS, 10, "toTurn: " + String( degFromRadiant( turnAngle ) ) +
          "\n" );
    INFO( Log::SKILLS, 10, "maxAngle: " + String( 
        degFromRadiant( maxAngle ) ) + "\n" );
    INFO( Log::SKILLS, 10, "simpleTurnAngle: " + String( 
        degFromRadiant( simpleTurnAngle ) ) + "\n" );
    turnAngle -= simpleTurnAngle;
    plan.addAction( simpleTurn( 
        player().turnAngleForSend( simpleTurnAngle ) ) );
    maxAngle = player().maxEffectiveTurnAngle();
  }
  INFO( Log::SKILLS, 10, "toTurn: " + String( degFromRadiant( turnAngle ) ) +
        "\n" );
  if( fabs(turnAngle) >= radFromDegree( MIN_TURN_ANGLE_IN_DEGREE ) ) {
    plan.addAction(simpleTurn( player().turnAngleForSend( turnAngle ) ));
  }
  INFO( Log::SKILLS, 7, "turnPlan: " + plan.toString() + "\n" );
  INFO( Log::SKILLS, 4, "LEAVE Skills::turnBy\n" );
  return plan;
}


Plan Skills::turnTo( const double finalDirection ) 
{
  return turnBy( finalDirection - player().faceDir() );
}

//
// setViewWidth
//

Plan Skills::setViewWidth( const desiredViewWidth )
{
  Plan plan;
  if ( player().viewWidth() != desiredViewWidth ) {
    plan.addAction( simpleChangeView( desiredViewWidth, VIEW_QUALITY_HIGH ) );
  }
  return plan;
}
//
// doNothing
//

Plan Skills::doNothing( int steps )
{
  if ( steps == UNDEF_INT ) {
    steps = maxLengthOfPlan;
  }
  Plan plan = Plan( steps );
  for( int i = 0; i < steps; i++ ) {
    plan.addAction( simpleEmptyAction() );
  }
  return plan;
}


//
// gainBallControl und Hilfsroutinen
//

Plan Skills::gainBallControl()
{
  INFO( Log::SKILLS, 3, "ENTER Skills::gainBallControl\n" ); 
  Plan plan = Plan( maxLengthOfPlan );
  // Szenario vorbereiten

  int i;
  ModelledPlayer otherPlayer;
  Vector relPos;
  for (i = 0; i < situation().playerCount(); i++ ) {
    otherPlayer = situation().playerWithIndex( i );
    if ( ( otherPlayer.id() != globalWorldModel->id() ) &&
        ( ball().isInKickRangeOfPlayer( otherPlayer ) == YES ) ) {
      INFO( Log::SKILLS, 8, "Gegner ist Spieler " + 
            String( otherPlayer.id() ) + "\n" );
      break;
    }
  }
  relPos = otherPlayer.pos() - player().pos();
  Vector preferredDir = globalWorldModel->preferredKickDirection( 
      player().pos() );
  INFO( Log::SKILLS, 10, "Seine relative Position: " + relPos.toString() + 
        "\n" );
  
  double angle = normalizeAngle( relPos.angle() - preferredDir.angle() );
  INFO( Log::SKILLS, 12, "Winkel zur Vorzugsrichtung: " + 
        String( degFromRadiant( angle ) ) + "\n" );
  if ( fabs( angle + PI/2 ) < fabs( angle - PI/2 ) ) {
    angle = relPos.angle() + PI/2;
    INFO( Log::SKILLS, 12, "Ausweichen nach links " );
  }
  else {
    angle = relPos.angle() - PI/2;
    INFO( Log::SKILLS, 12, "Ausweichen nach rechts " );
  }
  INFO( Log::SKILLS, 12, "in Richtung " + String( degFromRadiant( angle ) ) +
        "\n" );
  //plan.addPlan( turnTo( angle ) );
  plan.addAction( simpleKick( 50, angle ) );
  //plan.addAction( simpleDash( 100 ) );
  INFO( Log::SKILLS, 7, "gainBallControlPlan: " + plan.toString() + "\n" );
  INFO( Log::SKILLS, 3, "LEAVE Skills::gainBallControl\n" );
  return plan;
}

void Skills::resetSituation( Situation scenario )
{
  theSituation = scenario;
}

Action Skills::simpleChangeView( const int width,
    const int quality )
{
  ModelledPlayer player( player() );
  player.setViewWidth( width );
  situation().setPlayer( player );
  return Action( CHANGE_VIEW_ACTION, situation().simStep(), 
      width, quality );
}

Action Skills::simpleMoveTo( const Vector& position )
{
  ModelledPlayer player( player() );
  player.setPos( position );
  situation().setPlayer( player );
  situation().simulateSteps();
  return Action( MOVE_ACTION, situation().simStep() - 1, 
      position.x(), position.y() );
}

Action Skills::simpleTurn( const double angle )
{
  ModelledPlayer player( player() );
  player.turn( angle );
  situation().setPlayer( player );
  situation().simulateSteps();
  return Action( TURN_ACTION, situation().simStep() - 1, angle );
}

Action Skills::simpleDash( const double power )
{
  ModelledPlayer player( player() );
  player.accelerate( power );
  situation().setPlayer( player );
  situation().simulateSteps();
  return Action( DASH_ACTION, situation().simStep() - 1, power );
}

Action Skills::simpleKick( const double power,
    const double angle )
{
  ModelledBall ball( situation().ball() );
  ball.accelerate( power, angle, player() );
  situation().setBall( ball );
  situation().simulateSteps();
  return Action( KICK_ACTION, situation().simStep() - 1, power, angle );
}

Action Skills::simpleSay( const String& message )
{
  return Action( SAY_ACTION, situation().simStep(), message );
}

Action Skills::simpleCatch( const double angle )
{
  // Mehr kann bisher im Weltmodell nicht registriert werden
  situation().simulateSteps();
  return Action( CATCH_ACTION, situation().simStep() - 1, angle );
}
  
Action Skills::simpleSenseBody( )
{
  return Action( SENSE_BODY_ACTION, situation().simStep() );
}

Action Skills::simpleEmptyAction( )
{
  situation().simulateSteps();
  return Action( EMPTY_ACTION, situation().simStep() - 1 );
}

