#ifndef _CONFIGURATION_H_
#define _CONFIGURATION_H_

#include "String.h"
#include "Log.h"
#include "Common.h"
#include "Array.h"
#include "Vector.h"

/**
  Konfiguration des Servers und des Agenten.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Configuration.h} | [\URL[Implementierung]{src/Configuration.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Configuration.h,v $
  *   $Revision: 1.66 $
  *   $Date: 1998/08/14 13:20:05 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class Configuration
{
public:
  
  /** Konstruktor */
  Configuration();
  
  /** Liest eine Konfigurationsdatei ein. agent.conf und server.conf werden
  gleich behandelt. */
  void readConfigurationFile(String configurationFile);
  
  /** Schreibt aktuelle Parameter in das Log.
    @see Log */
  void logCurrentParameterValues();
  
#define MAX_FILE_LEN 256

  /** @name Parameter des SoccerServers */
  //@{
    /// goal width
    double goalWidth;               
    /// player size
    double playerSize;              
    /// player decay
    double playerDecay;             
    /// player rand
    double playerRand;              
    /// player weight
    double playerWeight;            
    /// player speed max
    double playerSpeedMax;          
    /// player stamina max
    double playerStaminaMax;        
    /// player stamina inc
    double playerStaminaInc;        
    /// player recovery decriment threshold
    double playerRecoverDecThr;     
    /// player recovery min
    double playerRecoverMin;        
    /// player recovery decriment
    double playerRecoverDec;        
    /// player dash effort decriment threshold
    double playerEffortDecThr;      
    /// player dash effrot min
    double playerEffortMin;         
    /// player dash effort decriment
    double playerEffortDec;         
    /// player dash effort incriment threshold
    double playerEffortIncThr;      
    /// player dash effort incriment
    double playerEffortInc;         
    /// ball size
    double ballSize;                
    /// ball decay
    double ballDecay;               
    /// ball rand
    double ballRand;                
    /// ball weight
    double ballWeight;              
    /// ball speed max
    double ballSpeedMax;            
    /// dash power rate
    double dashPowerRate;           
    /// kick power rate
    double kickPowerRate;           
    /// kickable margin
    double kickMarginServer;
    /// control radius
    double controlRadius;           
    /// max power
    double maxPower;                
    /// min power
    double minPower;                
    /// max moment
    double maxMoment;               
    /// min moment
    double minMoment;               
    /// visible angle
    double visibleAngleInDegree;    
    /// visible Dist
    double visibleDistance;         
    /// wind direction
    double windDirection;           
    /// wind force
    double windForce;               
    /// wind random
    double windRand;                
    /// goalie catchable area length
    double catchAreaLength;         
    /// goalie catchable area width
    double catchAreaWidth;          
    /// goalie catchable possibility
    double catchProbability;         
    /// corner kick margin
    double cornerKickMargin;        
    /// wind factor is none
    bool isWindFactor;              
    /// wind factor is random
    bool windRandom;                
    /// port number
    int portNumber;                 
    /// coach port number
    int coachPortNumber;            
    /// half time
    int halfTimeInSecs;             
    /// simulator step interval msec
    int simulatorStep;              
    /// udp send step interval msec
    int sendStep;                   
    /// udp recv step intercal msec
    int receiveStep;                
    /// string size of say message
    int sayMessageSize;             
    /// player hear_capacity_max
    int hearCapacityMax;            
    /// player hear_capacity_inc
    int hearCapacityInc;            
    /// player hear_capacity_decay
    int hearCapacityDecay;          
    /// goalie catch ban cycle
    int catchCycle;                 
    /// flag for using off side rule
    bool useOffSide;                
    /// flag for forbid kick off offside
    bool forbidKickOffOffside;      
    /// audio cut off Dist
    double audioCutOffDistance;     
    /// quantize step of Dist
    double distQuantStep;       
    /// quantize step of Dist for landmark
    double landmarksQuantStep;      

    /// server log file name
    char logFile[MAX_FILE_LEN];     
    /// log file name for recording
    char recFile[MAX_FILE_LEN];     
    /// flag for log recording
    bool recLog;
    /// flag for record log
    int recordVersion;
    /// log file name for replay
    char replay[MAX_FILE_LEN];      
  
    /// coach mode
    bool coachMode;                 
    /// coach with referee mode
    bool coachWithRefereeMode;      
    /// old format for hear command (coach)
    bool oldCoachHear;              
    /// flag for verbose mode
    bool verbose;
    /// intertia moment for turn
    double inertiaMoment;
    /// offside active area size
    double offsideArea;
  //@}

  // double playerRecoverInit;       /* player recovery init */
  // double playerEffortInit;        /* player dash effort init */
  // double controlRadiusWidth;      /* (control radius) - (player size) */

    
  /** @name Parameter des Agent-Files */
  //@{
    /// 0 meint direkt 1 weit weg vom Koerper
    double optimalKickDist;
    /// maximale Kick-Entfernung zum Ball
    double maxKickDist;
    /// Zahl der gespeicherten Situationen
    int historyCount;            
    /// die maximale Laenge eines Plans
    int maxLengthOfPlan;         
    /// zu tolerierende Kosten zur Stabilitaet der Wuensche
    int toleratedStabilityCost;  
    /// Nutzen, ab dem ein Constraint gefordert wird
    int minConstraintUtility;    
    /// Nutzen, ab dem ein Desire eventuell umgesetzt wird
    int minDesireUtility;
    /// Minimale Distanz, die ein Spieler ohne Initiative vom Ball haben mu\3.
    double enforcedMinDistToBall;
    /// Zahl von Takten nach einem Dash, die fuer die Spielersimulation noch
    /// beruecksichtigt werden.
    int maxStepsOfPlayerMovement;
    /// Zahl von Takten nach einem Kick, die fuer die Ballsimulation noch
    /// beruecksichtigt werden.
    int maxStepsOfBallMovement;
    /// Effektive Ballgeschwindigkeit, die ein Spieler mit einem kickBall erzielen
    /// kann
    double maxEffectiveBallSpeed;
    /// Zahl von Takten, die ein sich freilaufender Spieler ber\"ucksichtigt
    /// um festzustellen ob er im Einflu\3bereich des kickenden Spielers steht.
    int averageStepsOfBallMovement;
    /// Rollen der Spieler
    Array<int> roles;
    /// Aufstellung, wenn wir Anstoss haben
    Array<Vector> kickoffPositionsOffensive;
    /// Aufstellung, wenn der Gegner Anstoss hat
    Array<Vector> kickoffPositionsDefensive;
    /// Aufstellung waehrend des Spiels
    Array<Vector> homePositions;
    /// maximale baseUtility fuer GoalKickOption
    int maxGoalKickUtility;
    /// Gewichte fuer GoalKickUtilityBerechnung
    Array<int> goalKickUtilityWeights;
    /// maximale baseUtility fuer ForwardPassOption
    int maxForwardPassUtility;
    /// minimale baseUtility fuer ForwardPassOption
    int minForwardPassUtility;
    /// maximale baseUtility fuer DirectPassOption
    int maxDirectPassUtility;
    /// maximale baseUtility fuer DoublePassOption
    int maxDoublePassUtility;
    /// maximale baseUtility fuer DribbleOption
    int maxDribbleUtility;
    /// Laenge des beim Dribbeln betrachteten Sektors (in m)
    double dribbleSectorLength;    
    /// Breite des beim Dribbeln betrachteten Sektors (in Grad)
    double dribbleSectorWidth;
    /// maximale baseUtility fuer ControlBallOption
    int maxControlBallUtility;
    /// minimaler Nutzen fuer CollectDataOption
    int minCollectDataUtility;
    /// normaler Nutzen fuer CollectDataOption
    int normalCollectDataUtility;
    /// maximaler Nutzen fuer CollectDataOption
    int maxCollectDataUtility;
    /// normaler Nutzen fuer GoToBaseOption
    int normalGoToBaseUtility;
    /// minimaler Nutzen fuer GoToBaseOption
    int minGoToBaseUtility;
    /// Abstand vom Ball in der X-Koordinate, den ein Verteidiger bei
    /// Nachruecken zum Ball versucht einzuhalten
    double defenderMoveForwardBallDistActivation;
    /// maximale X-Position die ein nachrueckender Verteidiger einnimmt.
    double defenderMoveForwardStopLine;
    /// Abstand vom Ball in der X-Koordinate, den ein Mittelfeldspieler bei
    /// Nachruecken zum Ball versucht einzuhalten
    double midfieldMoveForwardBallDistActivation;
    /// maximale X-Position die ein nachrueckender Mittelfeldspieler einnimmt.
    double midfieldMoveForwardStopLine;
    /// Genauigkeit mit der ein Spieler auf seine BasePosition laufen soll
    double gotoBasePositionAccuracy;
    /// Genauigkeit mit der ein Spieler auf ein InterceptPosition laufen soll
    double getIntoSpaceAccuracy;
    /// Genauigkeit mit der ein Spieler zum Ball laufen soll
    double interceptBallAccuracy;
    /// normaler Nutzen fuer die InterceptBallOption
    int normalInterceptBallUtility;
    /// normaler Nutzen fuer die GetIntoSpaceOption
    int normalGetIntoSpaceUtility;
    /// Distanz von der Baseposition bis zu der sich der Spieler freilaeuft.
    double leaveBasePositionRange;
    /// Distanz vom Abfangpunkt des Balles ab der sich ein Spieler freilaeuft
    int getIntoSpaceDist;
    /// minimale Distanz vom Abfangpunkt bei der freigelaufen werden soll.
    double getIntoSpaceMinDist;
    /// Simulationsschritte bis zum Abfangpunkt ab der sich ein Spieler
    /// freilaeuft
    int getIntoSpaceSimSteps;
    /// Entfernung des Spielers von seiner BasePostion, bis zu der er sich bei
    /// Aktivierung der GetIntoSpaceOption zum Ball dreht und diesen hinterher
    /// schaut
    double getIntoSpaceAvoidBasePositionDist;
    /// Distanz von einem gegenerischen Spieler, ab der sich der Spieler
    /// anfaengt aktiv freizulaufen.
    double avoidCoverActivationDist;
    /// Entfernung die der Spieler vom Gegner haben will, wenn er sich beim
    /// freilaufen vor diesen stellt.
    double avoidCoverBestPositionDist;
    /// maximale Zeit die ein Spieler einen Plan verfolgen soll
    int maxForceContinuationTime;
    /// Entfernung zum Ball, ab dem er die Situation genauer betrachtet
    double distToBallForEvaluation;
    /// Schrittezahl ab der, die Berechnung von calculatedStepsToBall abbricht
    int maxStepsToCalculate;
    /// Aktivierung der einzelnen Modi f\"ur die Debug-Ausgabe
    Array<int> debugModes; 
    /** Stamina-Wert unterhalb dessen das ConserveStaminaConstraint
      aktiv wird. */ 
    int conserveStaminaActivation;
    /** Vektor, der die Steilheit (Steigung) des linearen Anstiegs
      des Nutzens des ConserveStaminaConstraint festlegt.  x: stamina, y:
      utility. y mu\3 kleiner als conserveStaminaActivation sein. */
    Vector conserveStaminaUtilityIncrease;
    /// maximaler Nutzen des ConserveStaminaConstraint
    int maxConserveStaminaUtility;
    /// maximaler Nutzen des AvoidOffsideConstraint
    int maxAvoidOffsideUtility;
    /// Bonus f\"ur das Staminalimit, wenn das AvoidOffsideConstraint aktiviert
    /// wird
    double avoidOffsideStaminaLimitBonus;
    /// Spielt der Agent ein echtes Spiel?
    bool tournamentMode;
    /// Variable die aussagt ob Kommunikation genutzt werden soll.
    bool useCommunication;
    /// Timeout fuer Kommunikation mit dem Server in Sekunden
    int communicationTimeout;
    /// Staminalimit fuers Dribbeln
    double dribbleStaminaLimit;
    /// Staminalimit fuers Ball abfangen
    double interceptBallStaminaLimit;
    /// Staminalimit fuer die Rueckkehr zur Basisposition
    double gotoBaseStaminaLimit;
    /// Staminalimit fuers Freilaufen
    double getIntoSpaceStaminaLimit;
    /// Toleranz bei wieviel Schritten Differenz zum schnellsten Spieler,
    /// dennoch zum Ball gerannt werden soll
    int interceptionTolerance;
    /// Distanz vom Tor, in der sich der Torwart normalerweise aufhaelt
    double goalieDistFromGoal;
    /// Distanz, ab der der Torwart den Ball oder andere Spieler als relevent
    /// ansieht
    double goalieDangerZone;
    /// Wenn ein nicht freundlicher Spieler naeher als diese Distanz kommt,
    /// wird nicht mehr gedribbelt (sondern hoffentlich gepasst).
    double emergencyPassDist;
    /// Voraussichtliche Reaktionszeit des gegnerischen Torwarts in Takten 
    int reactionTimeSteps;
    /// Bonus fuer das Dribbeln, wenn der Agent sich in Abseitsverdaechtiger
    /// Position befindet
    int dribbleInOffsideBonus;
  //@}

  /** @name Parameter fr das Online-Wissensakquirieren */
  //@{
    /// Aufloesung des gerasterten Spielfeldes, zum Lernen
    /// X-Koordinate (die laengere Seite des Spielfeldes)
    int xResolution;
    /// Y-Koordinate
    int yResolution;
    /// Staerke der Faerbung des Spielers
    int incPlayer;
    /// Staere der Faerbung des Balls
    int incBall;
    /// Staerke des Verblassens des Spielers
    int decPlayer;
    /// Staere des Verblassens des Balls
    int decBall;
    /// Schwellwert, ab dem ein Rasterkaestchen als Schwerpunkt gilt
    int threshold;
    /// minimale Laenge einer Schwerpunkt-Kante
    int minEdgeLength;
    /// maximale Laenge einer Schwerpunkt-Kante
    int maxEdgeLength;
    /// Abstand vom eigenen Tor-Aus, innerhalb dem Manndeckung erfolgt
    int manGuardingBorder;
    /// Abstand des Agenten von dem Schwerpunkt, der naeher dem eigenen Tor ist
    int distFromOpponent;
    /// Gewichte der Kriterien fuer die Bewertung der ermittelten Position:
    /// fuer den Abstand der home-Position davon
    double weightDistFromHome;
    /// fuer den Abstand von der eigenen aktuellen Position davon
    double weightDistFromCurrent;
    /// fuer die Haeufigkeit mit der der Ball auf der Schwerpunkt-Kante langlief
    double weightBallAppearance; 
    /// fuer die Farbe der Schwerpunkt-Kante auf der sich diese Position 
    /// befindet 
    double weightEdgeColour;   
    /// Abstand Agent - Ball, ab dem "GoToBaseOption" MAX_UTILITY zurueckgibt
    int maxUtilityDist;
    /// Abstand Agent - Ball, unter dem "GoToBaseOption" MIN_UTILITY zurueckgibt
    int minUtilityDist;
    /// Abstand Agent-Ball, unter dem der Ball gelernt wird,
    int ballLearnDist;
    /// maximaler Abstand des Agenten von der Home-Position zur Akzeptanz 
    /// einer ermittelten basePosition
    int maxDistFromHome;
    /// Abstand vom gegnerischen Tor-Aus, innerhalb dem der Gegner nicht
    /// gelernt
    /// wird, um nicht den gegnerischen Torwart zu lernen
    int learnNoGoalieDist;
    /// Abstand des Verteidigers vom Ball bei Einwuerfen ec
    double defenderFreeKickDist;
    /// Betrag um den verteidigende Spieler bei Einwuerfen an den Ball 
    /// nachruecken.
    double defenderFreeKickMoveUp;
    /// Abstand des Balles vom eigenen Tor, ab dem eine zu verteidigende 
    /// Situation vorliegt.
    double dangerousRangeFromGoal;
    /// Maximale Abweichung/Toleranz f\"ur Kickroutine am Aussenradius.
    double outerKickCircleTolerance;
    /// Maximale Abweichung/Toleranz f\"ur Kickroutine am Innenradius.
    double innerKickCircleTolerance;
    /// Maximale Abweichung/Toleranz f\"ur Kickroutine fuer Zwischenziele.
    double intermediateKickTolerance;
    /// Bezeichner f\"ur unsere Nachrichtenzeichenketten beim Weltmodelruf.
    int ourMessageIdentifier;
    /// Vorsprung in Takten, den unsere Spieler bei einem Pass haben muessen, damit
    /// er als erfolgreich bewertet wird
    int passLeadForTeammates;
    /// Entfernung innerhalb der ein nicht zuordenbarer Spieler an den Rand
    /// des Sichtbereiches des Spielers verschoben wird.
    double escapeFieldOfViewDist;
    /// Maximale Dist bei der noch Spieler aufeinander abgebildet werden.
    double maximalMergeDist;
    /// Distanz bis zu der ein Spieler einen anderen Spieler was zuruft.
    double distToPlayerForSay;
    /// Winkeldifferenz zwischen eigenen Blickwinkel und des Partners 
    /// Vorzugsrichtung die fuer einen Ruf akzeptiert wird (im Gradmass).
    double sayAngleDifference;
    
    double maxGoalKickDist;
  //@}
    
  /** @name Abhaengige, automatisch berechnete Parameter */
  //@{
    /// korigiertes kickMargin (kickMarginServer * maxKickDist)
    double kickMargin;
    /// (halfTimeInSecs * 1000 / simulatorStep)
    int halfTime;                   
    /// (radFromDegree(visibleAngleInDegree))
    double visibleAngle;            
    /// (sendStep /= 4)
    int sendStepNorm;               
    /// (kickMargin+playerSize+ballSize)
    double kickRange;
    ///  maxEffectiveBallSpeed * ( ( ballDecay^maxStepsOfBallMovement - 1 ) / 
    ///                   (ballDecay - 1 ) )
    double maxShotLength;
    ///  maxEffectiveBallSpeed * ( ( ballDecay^averageStepsOfBallMovement - 1 ) / 
    ///                   (ballDecay - 1 ) )
    double averageShotLength;
    ///  dashPowerRate * ( ( playerDecay^maxStepsOfPlayerMovement - 1 ) /
    ///                    ( playerDecay - 1 ) )
    double distPerStaminaPoint;
    /// Summe der goalKickUtility-Gewichte
    int sumOfGoalKickUtilityWeights;
    /// linearer Anstieg des ConserveStaminaConstraints
    double conserveStaminaGradient;
  //@}
    
private:
  ///
  void calculateDependentParameters();

  /** Liest eine Zeile aus einer geoeffneten Datei.
    Dabei werden Zeilen, die mit einem Backslash (\) gefolgt von einem
    newline (\n) enden, aneinandergehaengt. Leerzeilen und Zeilen, die mit einem
    (#) beginnen, werden ignoriert. Die Maximallaenge einer Zeile ist 
    \Ref{MAX_MESSAGE}.
    @param filePointer Zeiger auf Dateistruktur, so wie sie z.B. von fopen
      geliefert wird.
    @return Zeiger auf einen Buffer, der die naechste Zeile enthaelt, oder NULL,
      wenn das Dateiende erreicht worden ist. */
  char *readLine( FILE *filePointer );
  
  /** Scant ein Array von Vektoren ein. 
    @param buffer Der Speicherbereich, in dem der String steht.
    @param ArrayPtr Zeiger auf das Array, das initialisiert werden soll.
  */
  int scanArrayOfVector( char *buffer, Array<Vector> *ArrayPtr );

  /** Scant ein Array von Integerwerten ein. 
    @param buffer Der Speicherbereich, in dem der String steht.
    @param ArrayPtr Zeiger auf das Array, das initialisiert werden soll.
  */
  int scanArrayOfInt( char *buffer, Array<int> *ArrayPtr );
  
  


  ///
  typedef struct {
    char optionName[64];
    void *pointer;
    int type;
    double value;
  } options;
  
  ///
  options *serverOptions; 

#define INT               1
#define DOUBLE            2
#define BOOL              3
#define STRING            4
#define OPTIONAL          5
#define VECTOR            6
#define ARRAY_OF_VECTOR   7
#define ARRAY_OF_INT      8
#define NONE              99
  
};

  
/** Global verfuegbares Configuration-Object */
extern Configuration *globalConf;

#endif




