#ifndef _DRIBBLE_OPTION_H_
#define _DRIBBLE_OPTION_H_

#include "KickOption.h"
#include "Field.h"

const int DRIBBLE_OPTION_ID = 707;


/**
  Diese Klasse steht fuer den Wunsch, mit dem Ball in die  Vorzugsrichtung zu
  laufen.
  
  Dabei werden Spieler im Weg umgangen.
    
  Die zu dieser Option gehoerende Faehigkeit wird von der Praktikumsgruppe 
  'offline' implementiert.

  {\small {\bf Quelltext}

  [\URL[Header]{src/DribbleOption.h} | [\URL[Implementierung]{src/DribbleOption.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: DribbleOption.h,v $
  *   $Revision: 1.5 $
  *   $Date: 1998/08/14 13:20:25 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class DribbleOption : public KickOption
{
public:
  /** Erzeugt eine DribbleOption. */
  DribbleOption();

  /** Beschreibung als String fuer die Ausgabe auf einem Terminal. */
  virtual String toString() const;

  /** Erzeugt einen Plan zur Umsetzung der Absicht. 
    @return Plan, der die ersten Schritte zur Umsetzung der Absicht enthaelt. */
  Plan generatePlan() const;
  
  /** Prueft, ob die Absicht tatsaechlich umsetzbar ist.
    @return YES oder NO. */
  virtual bool isPossible() const;


  /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. Diese Methode muss fuer JEDE Unterklasse neu definiert 
      werden. */
  virtual StorableObject* copy() const;
  
  
protected:  
  /** Der Grundnutzen des Wunsches. Dieser Teil der Nutzenberechnung ist nur 
      vom Weltmodell abhaengig.
      @return Im Allgemeinen ein Wert zwischen MAX_UTILITY und MIN_UTILITY. */
  int baseUtility() const;

  /** Liefert die Kosten, die durch die Einhaltung einer Nebenbedingung   
      entstehen. 
      @param constraint Zeiger auf die einzuhaltende Nebenbedingung
      @return Im allgemeinen Fall ein Wert zwischen NEUTRAL_COST und MAX_COST. */
  virtual int constraintCost( const Constraint *constraint ) const;
  
  /** Waehlt das Dribble-Ziel aus.
    Dabei wird die interne Instanzvariable toTarget gesetzt. */
  void chooseTarget() const;
  
  /** Waehlt die Seite aus, auf der der Ball gefuehrt wird.
    Ist ein Gegner vor uns in der Naehe, wird der Ball auf der ihm abgelegenen Seite
    gefuehrt. Andernfalls wird die Seite ausgewaehlt, auf der der Ball schon
    liegt. 
    @param direction Die Richtung, in die gedribbelt werden soll */
  int chooseDribbleSide( double direction ) const;
};

#endif



