
/*
 * Aktueller Stand
 *
 *   $RCSfile: GetIntoSpaceOption.cc,v $
 *   $Revision: 1.25 $
 *   $Date: 1998/08/14 13:20:35 $
 *   $Author: myritz $ 
 *
 */
 
#include "GetIntoSpaceOption.h"
#include "Constraints.h"
#include "WorldModel.h"
#include "Skills.h"
#include "StraightLine.h"

GetIntoSpaceOption::GetIntoSpaceOption() : PassiveOption() 
{
  theId = GET_INTO_SPACE_OPTION_ID;
}  


int GetIntoSpaceOption::baseUtility() const 
{
  INFO( Log::PASSIVE_OPTIONS, 2, "ENTER GetIntoSpaceOption::baseUtility\n" );
  const ModelledPlayer& player = globalWorldModel->player();
  int utility = PassiveOption::baseUtility();

  if( utility == MAX_UTILITY ) {
    utility = NEUTRAL_UTILITY;
    if( globalWorldModel->gameState() != WorldModel::BEFORE_KICK_OFF &&
        globalWorldModel->gameState() != WorldModel::GOAL ) {
      if( globalWorldModel->ball().posReliability().get() < 
          Reliability::VERY_GOOD ) {
        if( globalWorldModel->initiative() == BOTH_TEAMS_ID ||
            globalWorldModel->initiative() == OUR_TEAM_ID ) {

          if( globalWorldModel->player().distTo( 
                  globalWorldModel->ball().pos() ) <=
              globalConf->distToBallForEvaluation ) {
            if( !globalWorldModel->amIOneOfTheNearestPlayerToBall() ) {
              // ich bin nicht einer der naechsten am Ball
              const Situation& situation = globalWorldModel->situation();
              
              int nearestPlayerSteps = situation.stepsOfNearestPlayer();
              //int nearestTeammateSteps =
              //  situation.stepsOfNearestPlayer( OUR_TEAM_ID );
	      //int nearestOpponentSteps =
              //  situation.stepsOfNearestPlayer( THEIR_TEAM_ID );
              const ModelledPlayer& nearestPlayer = 
                situation.nearestPlayer();
              ModelledBall ball = situation.ball();

              nearestPlayer.calculatedStepsToBall( ball,
                  nearestPlayerSteps );

              if( player.distTo( ball.pos() ) <= 
                  globalConf->getIntoSpaceDist ) {

    // Freilaufen wenn, man gedeckt ist , nicht zu nahe an den Ball ran 
		ModelledPlayer opponent;
		bool foundOpponent = NO;
		Vector basePosition = globalWorldModel->basePosition();
		basePosition = closestNonOffsideTarget( basePosition );
		const Vector& myPos = player.pos();
		if( player.distTo( basePosition ) < 
		    globalConf->leaveBasePositionRange &&
		    player.distTo( ball.pos() ) >
		    globalConf->getIntoSpaceMinDist  && 
		    nearestPlayerSteps <= globalConf->getIntoSpaceSimSteps) {
		  for( int i = 0; i < situation.playerCount(); i++ ) {
		    const ModelledPlayer& possibleOpponent = 
		      situation.playerWithIndex( i );
		    if( possibleOpponent.team() != OUR_TEAM_ID ) {
		      if( ( myPos - possibleOpponent.pos() ).length() <=
			  globalConf->avoidCoverActivationDist ) {
			foundOpponent = YES;
			opponent = possibleOpponent;
			break;
		      }
		    }
		  }
		}
    // freilaufen, weil Gegner gefunden - Abdeckpunkt finden
		if( foundOpponent ) {
		  utility = globalConf->normalGetIntoSpaceUtility;
		  StraightLine opponentToBall( opponent.pos(), ball.pos() );
		  Vector bestPosition = opponentToBall.calcPoint(
                      globalConf->avoidCoverBestPositionDist );
		  setTarget( bestPosition );
		}
		else {
		  if( ( myPos - basePosition ).length() <
		      globalConf->getIntoSpaceAvoidBasePositionDist ) {
		    utility = globalConf->normalGetIntoSpaceUtility;
		    setIsOptionPossible( YES );
		  }
		}
		/*
		if( nearestPlayerSteps == nearestTeammateSteps ) {
		  // aktive Einflussposition
                  
		  utility = globalConf->normalGetIntoSpaceUtility;
                  setIsOptionPossible( YES );
		  if( globalWorldModel->decisionReason() == 
                      WorldModel::DECISION_REASON_SEE_INFO ) {
                    setForceContinuationTime( nearestPlayerSteps ); 
                  }
                }
                else {
                  // passive Einflussposition
                }
		*/
              }
            }
          }
        }
        else { // initiative() == THEIR_TEAMS_ID
          if( isDefenseSituation() ) {
            Vector goalieHomePosition( globalWorldModel->ourGoal().x() +
                globalConf->goalieDistFromGoal, 0 );
            StraightLine ballToGoalieHome( globalWorldModel->ball().pos(),
                goalieHomePosition );
            Vector bestPosition = ballToGoalieHome.calcPoint( 
                globalConf->defenderFreeKickDist );
            double myDistToBestPosition = 
              player.distTo( bestPosition );
            const Situation& situation = globalWorldModel->situation();

            utility = globalConf->normalGetIntoSpaceUtility;
            for( int i = 0; i < situation.playerCount(); i++ ) {
              const ModelledPlayer& teammate = situation.playerWithIndex( i );
              if ( teammate.team() == OUR_TEAM_ID &&
                   teammate.id() != player.id()   &&
                   isDangerousPos( teammate.pos() ) 
                   // und ist nicht Torwart
                   ) {
                if ( myDistToBestPosition > 
                     teammate.distTo( bestPosition ) ) {
                  utility = NEUTRAL_UTILITY;
                  break;
                }
              }
            }

            if ( utility != NEUTRAL_UTILITY ) {
              setTarget( bestPosition );
              INFO( Log::PASSIVE_OPTIONS, 8,
                    "target: " + target().toString() + "\n" );
            }
          }
          else { // !isDefenseSituation()
            utility = MIN_UTILITY;
          }
    
        }
      }
    }
  }

  INFO( Log::PASSIVE_OPTIONS, 6, "utility: " + String( utility ) + "\n" );
  INFO( Log::PASSIVE_OPTIONS, 2, "LEAVE GetIntoSpaceOption::baseUtility\n" );

  return utility;
}

Plan GetIntoSpaceOption::generatePlan() const 
{
  Plan plan;
  if( !isTargetComputed() ) {
    // Jan: Hier sp"ater bessere Funktionalitaet
    setTarget( globalWorldModel->player().pos() );
  }
  // immer Abseits beachten
  setTarget( closestNonOffsideTarget( target() ) );
    
  globalSkills->resetSituation();
  plan.addPlan( globalSkills->setViewWidth( VIEW_WIDTH_NORMAL ) );
  // Kommunikation
  plan.addPlan( isSomethingToSay() );
  if( globalWorldModel->player().distTo( target() ) <=
      globalConf->getIntoSpaceAccuracy ) {
    // bereits am Ziel angekommen, zum Ball nachdrehen
    double angle = (globalWorldModel->ball().pos() - 
                    globalWorldModel->player().pos()).angle();
    plan = globalSkills->turnTo( angle );
    plan.addPlan( globalSkills->doNothing() );
    setIsPlanCompleted( YES );
  }
  else {
    double staminaLimit = globalConf->getIntoSpaceStaminaLimit;
    if( isConstraintAccepted( AVOID_OFFSIDE_CONSTRAINT_ID ) ) {
      staminaLimit -= globalConf->avoidOffsideStaminaLimitBonus;
    }

    plan = globalSkills->gotoRegion( target(), 
        globalConf->getIntoSpaceAccuracy, staminaLimit,
        Skills::AVOID_OBSTACLES );
    if( plan.length() == 0 ) {
      setIsPlanCompleted( YES );
    }
  }
  return plan;
}


StorableObject* GetIntoSpaceOption::copy() const 
{
  return new GetIntoSpaceOption( *this );
}

String GetIntoSpaceOption::toString() const  
{
  return String( "( GetIntoSpaceOption" ) + PassiveOption::toString() +
    String( " ) " );
}


 


