
/*
 * Aktueller Stand
 *
 *   $RCSfile: InterceptBallOption.cc,v $
 *   $Revision: 1.38 $
 *   $Date: 1998/08/14 13:20:42 $
 *   $Author: myritz $ 
 *
 */
 
#include "InterceptBallOption.h"
#include "Constraints.h"
#include "WorldModel.h"
#include "Skills.h"

InterceptBallOption::InterceptBallOption() : PassiveOption() 
{
  theId = INTERCEPT_BALL_OPTION_ID;
}  


int InterceptBallOption::baseUtility() const 
{
  INFO( Log::PASSIVE_OPTIONS, 2, "ENTER InterceptBallOption::baseUtility\n" );
  int utility = PassiveOption::baseUtility();

  /*
  if ( globalWorldModel->role() == GOALIE_ROLE && 
       globalWorldModel->gameState() != WorldModel::FREE_KICK ) {
    utility = MIN_UTILITY;
    INFO( Log::PASSIVE_OPTIONS, 12, "No normal Intercept because is Goalie\n");
    INFO( Log::PASSIVE_OPTIONS, 6, "utility: " + String( utility ) + "\n" );
    INFO( Log::PASSIVE_OPTIONS, 2, 
          "LEAVE InterceptBallOption::baseUtility\n" );
    return utility;
  }
  */

  if( utility == MAX_UTILITY ) {
    utility = NEUTRAL_UTILITY;
    if( globalWorldModel->gameState() != WorldModel::BEFORE_KICK_OFF &&
        globalWorldModel->gameState() != WorldModel::GOAL ) {
      if( globalWorldModel->ball().posReliability().get() < 
          Reliability::VERY_GOOD ) {
        if( globalWorldModel->initiative() == BOTH_TEAMS_ID ||
            globalWorldModel->initiative() == OUR_TEAM_ID ) {

          if( globalWorldModel->player().distTo( 
                  globalWorldModel->ball().pos() ) <=
              globalConf->distToBallForEvaluation ) {
            if ( globalWorldModel->gameState() == WorldModel::FREE_KICK &&
                globalWorldModel->initiative() == OUR_TEAM_ID &&
                globalWorldModel->role() != GOALIE_ROLE ) {
              // Der Torwart macht nie den Abstoss, also muss ein anderer
              // Spieler hinrennen
              
              Situation situation = globalWorldModel->situation();
              // Torwart raus. Das heisst wir nehmen an, 11 ist der Torwart.
              try {
                situation.removePlayerWithId( 430 );
              }
              catch ( Situation::NoPlayerWithThisIdException ) {
                // Wenn er nicht drin ist, muss er auch nicht rausgeschmissen
                // weden
              }
              // ich selbst bin naechster Spieler unseres Teams am Ball
              if ( situation.isOneOfTheNearestTeammatesToBall( 
                  globalWorldModel->id() ) ) {
                setTarget( situation.ball().pos() );
                utility = globalConf->normalInterceptBallUtility;
              }
            }
            else {
              // Normaler Intercept
              if( globalWorldModel->amIOneOfTheNearestPlayerToBall() ||
                  globalWorldModel->role() == GOALIE_ROLE ) {
                const Situation& situation = globalWorldModel->situation();
                int nearestPlayerSteps = situation.stepsOfNearestPlayer();
                int myStepsToBall = situation.stepsOfPlayerWithId( 
                    globalWorldModel->id() );
                const ModelledPlayer& player = situation.playerWithId(
                    globalWorldModel->id() );
                ModelledBall ball = situation.ball();

                if( myStepsToBall > 6 ) {
                  ball.simulateSteps();
                }
                if( myStepsToBall > 12 ) {
                  ball.simulateSteps();
                }

                player.calculatedStepsToBall( ball, myStepsToBall );

                utility = globalConf->normalInterceptBallUtility;
                setTarget( ball.pos() );              
                if( globalWorldModel->decisionReason() == 
                    WorldModel::DECISION_REASON_SEE_INFO ) {
                  setForceContinuationTime( nearestPlayerSteps ); 
                }
              }
            }
          }
        }
      }
    }
  }
  INFO( Log::PASSIVE_OPTIONS, 6, "utility: " + String( utility ) + "\n" );
  INFO( Log::PASSIVE_OPTIONS, 2, "LEAVE InterceptBallOption::baseUtility\n" );

  return utility;
}

void InterceptBallOption::adaptToPreviousIntention( 
  const Option *previousIntention ) 
{
}

Plan InterceptBallOption::generatePlan() const 
{
  Plan plan;
  if( !isTargetComputed() ) {
    throwException( Exception( "Target should have been already computed" ) );
  }
  double interceptBallAccuracy;

  globalSkills->resetSituation();

  if( globalWorldModel->initiative() == OUR_TEAM_ID ) {
    interceptBallAccuracy = globalConf->kickRange;
    if ( globalWorldModel->gameState() == WorldModel::KICK_IN ) {
      setTarget( computeKickInTarget() );
      interceptBallAccuracy = 0.1;
    }
    else {
      noCancelByReasons( CANCEL_REASON_NEW_BALL_INFO |
			 CANCEL_REASON_CONT_TIME_EXPIRED );
    }
    plan.addPlan( globalSkills->setViewWidth( VIEW_WIDTH_WIDE ) );
  }
  if( globalWorldModel->initiative() == BOTH_TEAMS_ID ) {
    interceptBallAccuracy = globalConf->interceptBallAccuracy;
    plan.addPlan( globalSkills->setViewWidth( VIEW_WIDTH_NORMAL ) );
  }

  if( globalWorldModel->player().distTo( target() ) > 
      interceptBallAccuracy ) {  
    plan.addPlan( globalSkills->gotoRegion( target(), interceptBallAccuracy,
        globalConf->interceptBallStaminaLimit, Skills::AVOID_OBSTACLES ) );
  }
  else {
    if( globalWorldModel->initiative() == OUR_TEAM_ID ) {
      Plan turnPlan;
      Vector ballPos = globalWorldModel->ball().pos();
      if( globalWorldModel->gameState() == WorldModel::KICK_IN ) {
	turnPlan = globalSkills->turnTo( 
            ( Vector( ballPos.x(), -ballPos.y() ) - ballPos ).angle() );
      }
      else { // in Vorzugsrichtung schauen
	turnPlan = globalSkills->turnTo( 
            globalWorldModel->preferredKickDirection( ballPos ).angle() );
      }
      if( turnPlan.length() > 0 && 
	  fabs( turnPlan.actionAt( 0 ).firstParameter() ) >=
	  radFromDegree( 10.0 ) ) {
	plan.addPlan( turnPlan );
      }
      if( plan.length() == 0 ) {
	// Position erreicht, Blickwinkel korrekt -> auf neuen Input warten.
	setCancelReasons();
      }
      plan.addPlan( globalSkills->doNothing() );
    }
    else {
      setIsPlanCompleted( YES );
    }
  }
  return plan;
}

Vector InterceptBallOption::computeKickInTarget() const
{
  const ModelledPlayer& player = globalWorldModel->player();
  const ModelledBall& ball= globalWorldModel->ball();
  Vector target;
  
  if ( player.distTo( ball.pos() ) > 5 ) {
    // Auf Position auf der Linie laufen, von der aus wir den Ball sicher
    // umrunden koennen
    target.set( ball.pos().x() + 3, ball.pos().y() + 
        signum( ball.pos().y() ) * 3 );
  }
  else {
    if ( ball.pos().y() > 0 ) {
      // Einwurf von oben
      target.set( ball.pos().x(), 
          ball.pos().y() + player.size() + ball.size() + 0.1 );
    }
    else {
      // Einwurf von unten
      target.set( ball.pos().x(), 
          ball.pos().y() - player.size() - ball.size() - 0.1 );
    }
  }
  return target;
}
      

StorableObject* InterceptBallOption::copy() const 
{
  return new InterceptBallOption( *this );
}

String InterceptBallOption::toString() const  
{
  return String( "( InterceptBallOption" ) + PassiveOption::toString() +
    String( " ) " );
}


 

