#ifndef _INTERCEPT_BALL_OPTION_H_
#define _INTERCEPT_BALL_OPTION_H_

#include "PassiveOption.h"

const int INTERCEPT_BALL_OPTION_ID = 701;


/**
  Diese Klasse steht fuer den Wunsch den Ball abzufangen. Dieser Wunsch
  bekommt eine hohe Aktivierung, wenn der Spieler den Ball am schnellsten
  bzgl. seiner Teamkameraden erreichen kann, oder nur minimal langsamer
  ist als der schnellste Teamkamerad. Bei Auswahl dieses Wunsches wird
  ein Plan erzeugt um den Ball zu erreichen.

  {\small {\bf Quelltext}

  [\URL[Header]{src/InterceptBallOption.h} | [\URL[Implementierung]{src/InterceptBallOption.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: InterceptBallOption.h,v $
  *   $Revision: 1.10 $
  *   $Date: 1998/08/14 13:20:43 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class InterceptBallOption : public PassiveOption
{
public:
  /** Erzeugt eine InterceptBallOption. */
  InterceptBallOption();
  
  /** Anpassung an die bisherige Absicht. */
  void adaptToPreviousIntention( const Option *previousIntention );

  /** Erzeugt einen Plan zur Umsetzung der Absicht.  */
  Plan generatePlan() const;

  /** Beschreibung als String fuer die Ausgabe auf einem Terminal. */
  virtual String toString() const;

  /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. Diese Methode muss fuer JEDE Unterklasse neu definiert 
      werden. */
  virtual StorableObject* copy() const;
  
private:
  
  /** Der Grundnutzen des Wunsches. Dieser Teil der Nutzenberechnung ist nur 
    vom Weltmodell abhaengig.
    @return Ein Wert zwischen MAX_UTILITY und MIN_UTILITY. */
  int baseUtility() const;

  /** Berechnet Zwischenziele fuer Einwurf */
  Vector computeKickInTarget() const;
};

#endif

