#ifndef _KICK_OPTION_H_
#define _KICK_OPTION_H_

#include "Option.h"
#include "Field.h"


/**
  Diese Klasse ist Superklasse fuer alle Wuensche, die den Ball kicken wollen.
  
  Alle diese Wuensche liegen das Ziel bereits bei der baseUtility-Berechnung
  fest, und auch die Ausfuehrung (\Ref{generatePlan}(), die Kostenberechnungen
  (\Ref{constraintCost}(), \Ref{cancellationCost}()) und einige andere Methoden
  sind identisch.

  {\small {\bf Quelltext}

  [\URL[Header]{src/KickOption.h} | [\URL[Implementierung]{src/KickOption.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: KickOption.h,v $
  *   $Revision: 1.6 $
  *   $Date: 1998/08/14 13:20:46 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class KickOption : public Option
{
public:
  
  /** Gibt die Abbruchkosten an. 
    @return NEUTRAL_COST */
  virtual int cancellationCost() const;

  /** Prueft, ob die Absicht tatsaechlich umsetzbar ist.
    @return YES oder NO. */
  virtual bool isPossible() const;

  /** Erzeugt einen Plan zur Umsetzung der Absicht. Wurde die Umsetzung des
    Plans bereits teilweise erledigt, so stellt dies generatePlan() fest und
    erzeugt die naechsten Schritte ausgehend von der aktuellen Situation im
    Weltmodell.
    @return Plan, der die ersten Schritte zur Umsetzung der Absicht enthaelt.
    Die Zahl der Schritte laesst sich konfigurieren (Variable maxLengthOfPlan
    in agent.conf). */
  virtual Plan generatePlan() const;
  
protected:
  /** Protected Konstruktor macht die Klasse abstrakt.
    Erzeugt eine KickOption. */
  KickOption();

  /** Der Grundnutzen des Wunsches. 
      Diese Funktion testet, ob die Grundvoraussetzungen f\"ur einen Kick 
      vorliegen, d.h. die eigene und die Ballposition sind ausreichend
      zuverl\"assig und der Ball ist innerhalb der KickRange.
      @return NEUTRAL_UTILITY, wenn die Positionen nicht ausreichend genau sind;
      MIN_UTILITY, wenn der Ball ausserhalb der KickRange ist; MAX_UTILITY, wenn
      gekickt werden kann. */
  virtual int baseUtility() const;

    /** Der Vektor von der eigenen Position zum Kickziel. 
    Wird bereits von baseUtility ausgerechnet und gecacht. */
  mutable Vector toTarget;
  
  /** Liefert die Kosten, die durch die Einhaltung einer Nebenbedingung   
      entstehen. 
      Praktisch alle Kick-Optionen machen keinen Dash,  daher wird das hier
      fuer alle vorgegeben. Wer Sonderw\"unsche hat, mu\3 diese Methode
      \"uberschreiben.
      @param constraint Zeiger auf die einzuhaltende Nebenbedingung
      @return Im allgemeinen Fall ein Wert zwischen NEUTRAL_COST und MAX_COST. */
  virtual int constraintCost( const Constraint *constraint ) const;
};

#endif



