#ifndef _MOVINGOBJECT_H_
#define _MOVINGOBJECT_H_

#include "StableObject.h"

/**
  Ein sich (moeglicherweise) bewegendes Objekt.
  Ein MovingObject hat wie ein StableObject eine Id und eine Position,
  zusaetzlich aber noch eine Geschwindigkeit.

  {\small {\bf Quelltext}

  [\URL[Header]{src/MovingObject.h} | [\URL[Implementierung]{src/MovingObject.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: MovingObject.h,v $
  *   $Revision: 1.15 $
  *   $Date: 1998/08/14 13:20:58 $
  *   $Author: myritz $ 
  *
  **************************************************
  \end{verbatim} }*/
class MovingObject : public StableObject
{
public:

  /** @name Erzeugen und Zerst\"oren */
  //@{
    /** Default-Konstruktor. */
    MovingObject();

    /** Neues MovingObject mit Id, Position und Geschwindigkeit.
      @param id Die Objekt-Id. M\"ogliche Werte sind in den \Ref{Objekt-Ids}
        aufgelistet.
      @param newPosition Die Position.
      @param newSpeed Die Geschwindigkeit. */
    MovingObject( int id, const Vector& aPos, const Vector& aSpeed = Vector() );

    /** Legt eine Kopie des Objekts auf dem Heap an und liefert einen Zeiger
      darauf zurueck. */
    virtual StorableObject* copy() const;
	//@}
    
  /** @name Werte abfragen */
  //@{  
    /** Prueft, ob das Objekt komplett definiert ist. 
        @return YES falls die Id verschieden von UNDEF_ID ist und sowohl Position
        als auch Geschwindigkeit complete sind, sonst NO */
    bool isComplete() const;

    /** Liefert Geschwindigkeit */
    const Vector& speed() const;
    
    /** Beschreibung als String fuer die Ausgabe auf einem Terminal.
        Es werden ID Position und Geschwindigkeit ausgegeben. */
    virtual String toString() const;
  //@}
    
  /** @name Werte setzen */
  //@{  
    /** Setzt Geschwindigkeit neu. */
    void setSpeed( const Vector& newSpeed );
  //@}  
  
private:
  /** Geschwindigkeitsvektor. */
  Vector theSpeed;
};

inline const Vector& MovingObject::speed() const {
  return theSpeed;
}

inline void MovingObject::setSpeed(const Vector& newSpeed) {
  theSpeed = newSpeed;
}

inline bool MovingObject::isComplete() const
{
  return (StableObject::isComplete() && theSpeed.isValid() );
}

#endif

