
/*
 * Aktueller Stand
 *
 *   $RCSfile: Situation.cc,v $
 *   $Revision: 1.37 $
 *   $Date: 1998/08/14 13:21:13 $add
 *   $Author: myritz $ 
 *
 */
 
#include "Situation.h"
#include "WorldModel.h"
#include "Log.h"
#include "InvalidIdException.h"
#include "Timer.h"

double Situation::theMinDistPlayerBall = UNDEF_DOUBLE;

Situation::Situation( int aSimStep ) 
  : theSimStep( aSimStep ), thePlayers( Array<ModelledPlayer>(22) ),
    areStepsToBallEstimated( NO ),
    stepsToBallOfPlayers( Array<int>(22) ),
    nearestAllPlayerSorted( Array<int>(22) ),
    nearestTeammatesSorted( Array<int>(11) ),
    nearestEnemiesSorted( Array<int>(11) ),
    theOffsideLineX( UNDEF_DOUBLE ),
    theOffsideLineXReliability( Reliability::RANDOM )
{
  if ( theMinDistPlayerBall == UNDEF_DOUBLE ) {
    theMinDistPlayerBall = globalConf->playerSize + globalConf->ballSize;
  }
}


void Situation::setPlayer( const ModelledPlayer& newPlayer )
{
  try {
    int index = indexOfPlayerWithId( newPlayer.id() );
    setPlayer( newPlayer, index );
  }
  catch( NoPlayerWithThisIdException ) {
    addPlayer( newPlayer );
  }
}

int Situation::indexOfPlayerWithId( int aPlayerId ) const
{
  int i;
  for( i = 0; i < thePlayers.count(); i++ ) {
    if ( thePlayers[i].id() == aPlayerId ) {
      return i;
    }
  }
  throwException( NoPlayerWithThisIdException() );
}

void Situation::removePlayerWithIndex( int index )
{
  if ( playerWithIndex( index ).team() == THEIR_TEAM_ID ) {
    theOffsideLineX = UNDEF_DOUBLE;
  }
  thePlayers.removeObjectAt( index );
}  

void Situation::removePlayerWithId( int aPlayerId )
{
  int playerIndex = indexOfPlayerWithId( aPlayerId );
  removePlayerWithIndex( playerIndex );
}

const ModelledPlayer& Situation::playerWithId( int aPlayerId ) const
{
  return thePlayers[indexOfPlayerWithId( aPlayerId )];
}

const int Situation::stepsOfPlayerWithId( int aPlayerId ) const
{
  if( areStepsToBallEstimated ) {
    return stepsToBallOfPlayers[indexOfPlayerWithId( aPlayerId )];
  }
  else {
    return thePlayers[indexOfPlayerWithId( aPlayerId )].estimatedStepsToBall(
        theBall );
  }
}


const ModelledPlayer& Situation::playerWithIndex( int index ) const
{
  return thePlayers[index];
}

bool Situation::containsPlayerWithId( int aPlayerId ) const
{
  try {
    indexOfPlayerWithId( aPlayerId );
    return YES;
  }
  catch( NoPlayerWithThisIdException ) {
    return NO;
  }
}
  
void Situation::simulateSteps( unsigned int stepCount ) 
{
  if( stepCount > 10 ) {
    WARNING( Log::SITUATION, 
        "Situation::simulateSteps: more than 10 steps." );
  }
   
  int playerIndex;
  Vector playerToBall;
  bool isBallValid;
  if ( theBall.posReliability().get() >= Reliability::RANDOM ) {
    isBallValid = NO;
  }
  else {
    theBall.simulateSteps( stepCount );
    isBallValid = YES;
  }
  
  for( playerIndex = playerCount() - 1; playerIndex >= 0; playerIndex-- ) {
    thePlayers[playerIndex].simulateSteps( stepCount );
    playerToBall = theBall.pos() - thePlayers[playerIndex].pos();
    if ( isBallValid == YES &&
        playerToBall.length() < theMinDistPlayerBall ) {
      INFO( Log::SITUATION, 10, 
          "Situation::simulateSteps collision with ball detected\n" );
      playerToBall.setPolar( theMinDistPlayerBall, playerToBall.angle() );
      INFO( Log::SITUATION, 10, "Moving ball from " + 
          theBall.pos().toString() + " to " );
      theBall.setPos( thePlayers[playerIndex].pos() + playerToBall );
      INFO( Log::SITUATION, 10, theBall.pos().toString() + "\n" );
      theBall.setSpeed( -0.1 * theBall.speed() );
    } 
  }
  
  theSimStep += stepCount;
  areStepsToBallEstimated = NO;
  theOffsideLineX = UNDEF_DOUBLE;
}

bool Situation::isOneOfTheNearestTeammatesToBall( int id ) const 
{
  int nearestPlayerSteps = stepsOfNearestPlayer( OUR_TEAM_ID );
  int stepsOfAskedPlayer = stepsOfPlayerWithId( id );

  if( stepsOfAskedPlayer - nearestPlayerSteps <= 
      globalConf->interceptionTolerance ) {
    return YES;
  }
  else {
    return NO;
  }
}

int Situation::indexOfNearestPlayer( int teamId ) const
{
  if( !areStepsToBallEstimated ) {
    estimateForAllStepsToBall();
  }
  try {
    if( teamId == BOTH_TEAMS_ID ) {
      return nearestAllPlayerSorted[0];
    }
    else if( teamId == OUR_TEAM_ID ) {
      return nearestTeammatesSorted[0];
    }
    else if( teamId == THEIR_TEAM_ID ) {
      return nearestEnemiesSorted[0];
    }
    else {
      throwException( InvalidIdException( "invalid teamId" ) );
    }
  }
  catch( Array<int>::IndexOutOfBoundsException ) {
    throwException( NoPlayerWithThisIdException() );
  }
}

const ModelledPlayer& Situation::nearestPlayer( int teamId ) const
{
  return thePlayers[indexOfNearestPlayer( teamId )];
}

int Situation::stepsOfNearestPlayer( int teamId ) const
{
  try {
    return stepsToBallOfPlayers[indexOfNearestPlayer( teamId )];
  }
  catch( NoPlayerWithThisIdException ) {
    return ModelledPlayer::MAX_STEPS_TO_BALL;
  }
}

void Situation::estimateForAllStepsToBall() const
{
  int playerIndex;
  int playerId;

  INFO( Log::SITUATION, 2, "ENTER estimateForAllStepsToBall with " +
        String( playerCount() ) + " players.\n" );
  INFO( Log::SITUATION, 9, "Ball: " + theBall.toString() + "\n" );
  INFO( Log::SITUATION, 9, "started at time: " + 
        String( (long) globalTimer->time() ) + "\n" );
  stepsToBallOfPlayers.removeAll();
  nearestAllPlayerSorted.removeAll();
  nearestTeammatesSorted.removeAll();
  nearestEnemiesSorted.removeAll();
  for( playerIndex = 0; playerIndex < playerCount(); playerIndex++ ) {
    int steps = thePlayers[playerIndex].estimatedStepsToBall( theBall );
    stepsToBallOfPlayers.addObject( steps );
    INFO( Log::SITUATION, 6, "Player (" + String( playerIndex ) + "): " );
    INFO( Log::SITUATION, 9, thePlayers[playerIndex].toString() );
    INFO( Log::SITUATION, 6, "   estimatedStepsToBall: " +
          String( stepsToBallOfPlayers[playerIndex] ) + "\n" );

    playerId = thePlayers[playerIndex].id();
    
    sortIntoField( playerIndex, nearestAllPlayerSorted );
    
    if( thePlayers[playerIndex].team() == OUR_TEAM_ID ) {
      sortIntoField( playerIndex, nearestTeammatesSorted );
    }
    else if( thePlayers[playerIndex].team() == THEIR_TEAM_ID ) {
      sortIntoField( playerIndex, nearestEnemiesSorted );
    }
    else { // nichtzuordnungsf"ahige Spieler z"ahlen als Feinde
      sortIntoField( playerIndex, nearestEnemiesSorted );
    }
  }

  if( IS_ACTIVE( Log::SITUATION, 6 ) ) {
    cout << "Sorted Field of all Players :" << nearestAllPlayerSorted;
    cout << "Sorted Field of Teammates :" << nearestTeammatesSorted;
    cout << "Sorted Field of Enemies :" << nearestEnemiesSorted;
  }
  
  areStepsToBallEstimated = YES;
  INFO( Log::SITUATION, 2, "LEAVE estimateForAllStepsToBall\n" );
  INFO( Log::SITUATION, 9, "ended at time: " + 
        String( (long) globalTimer->time() ) + "\n" );
}

void Situation::sortIntoField( int playerIndex, Array<int>& sortedField ) const
{
  int i, j;
  int count = sortedField.count();
  for( i = 0; i < count &&
         stepsToBallOfPlayers[playerIndex] >
         stepsToBallOfPlayers[sortedField[i]]; i++ ) {
  }
  // richtige Stelle gefunden (i)
  // Platz schaffen durch kopieren
  sortedField.addObject( UNDEF_INT );
  for( j = count; i < j; j-- ) {
    sortedField[j] = sortedField[j-1];
  } 
  // neuen Wert einf"ugen
  sortedField[i] = playerIndex;
}

int Situation::indexOfNearestPlayerToPos( const Vector& otherPos, 
    const int teamId ) const
{
  int i;
  double dist;
  double mindist = DBL_MAX;
  int index = UNDEF_INT;
  if ( teamId != OUR_TEAM_ID && teamId != THEIR_TEAM_ID && 
       teamId != BOTH_TEAMS_ID ) {
    throwException( InvalidIdException("Invalid TeamId.") );
  }
  for ( i = 0; i < playerCount(); i++ ) {
  
    // Alle Spieler ueberspringen, die nicht zu einem erfragen Team gehoeren
    int playerTeamId = thePlayers[i].team();
    if ( playerTeamId == UNDEF_TEAM_ID ) {
      playerTeamId = THEIR_TEAM_ID;
    }
    if ( playerTeamId != teamId && teamId != BOTH_TEAMS_ID ) {
      continue;
    }

    dist = ( thePlayers[i].pos() - otherPos ).length();
    if ( dist < mindist ) {
      mindist = dist;
      index = i;
    }
  }
  if ( index == UNDEF_INT ) {
    throwException( NoPlayerWithThisIdException() );
  }
  return index;
}


double Situation::distOfNearestPlayerToPos( const Vector& otherPos, 
    const int teamId ) const
{
  return ( thePlayers[indexOfNearestPlayerToPos( otherPos )].pos() - 
      otherPos ).length();
}
     
   


String Situation::toString() const
{
  String output;
  output = "( Situation at simStep " + String( simStep() ) + ":\n" +
    "ball: " + ball().toString() + ",\n" +
    String( playerCount() ) + " players:";
  for( int i = 0; i < playerCount(); i++) {
    output = output + "\n   " + playerWithIndex(i).toString();
  }
  if( playerCount() > 0 ) {
    output = output + "\n relative ball: Pos: " + 
      (ball().pos() - playerWithIndex(0).pos()).toString() + " Speed: " +
      (ball().speed() - playerWithIndex(0).speed()).toString();
  }
  output = output + " )\n";

  return output;
}

double Situation::offsideLineX() const
{
  if ( theOffsideLineX == UNDEF_DOUBLE ) {
    theOffsideLineX = computeOffsideLineX();  
  }
  return theOffsideLineX;
}
    
double Situation::computeOffsideLineX() const
{
  INFO( Log::SITUATION, 3, "ENTER Situation::computeOffsideLineX\n");    
  double lastPlayerX = 0.0;
  double secondLastPlayerX = 0.0;
  double newOffsideLineX;
  double static const offsideSafetydist = 2.0;
  theOffsideLineXReliability.set( Reliability::PERFECT );
  // x-Koordinaten des letzten und vorletzten gegnerischen Spielers
  // bestimmen.
  for( int i = 0; i < playerCount(); i++) {
    const ModelledPlayer& player = playerWithIndex( i );
    if ( player.team() != OUR_TEAM_ID ) {
      INFO( Log::SITUATION, 8, "Checking player " + 
          String( i ) + "( " + String( player.id() ) + 
          " ) x: " + String( player.pos().x() ) + "\n" );
      if ( player.pos().x() - offsideSafetydist > lastPlayerX ) {
        INFO( Log::SITUATION, 8, "Is new last player\n" );
        secondLastPlayerX = lastPlayerX;
        lastPlayerX = player.pos().x() - offsideSafetydist;
        theOffsideLineXReliability.set( max( offsideLineXReliability().get(),
            player.posReliability().get() ) );
      }
      else if ( player.pos().x() - offsideSafetydist > secondLastPlayerX ) {
        INFO( Log::SITUATION, 8, "Is new second last player\n" );
        secondLastPlayerX = player.pos().x() - offsideSafetydist;
        theOffsideLineXReliability.set( max( offsideLineXReliability().get(),
            player.posReliability().get() ) );
      }
    }
    else {
      INFO( Log::SITUATION, 8, "Skipping player (unknown or team mate)\n" );
    }
  }
  
  if ( ball().posReliability().get() < Reliability::RANDOM &&
       ball().pos().x() - offsideSafetydist > secondLastPlayerX ) {
    newOffsideLineX = ball().pos().x() - offsideSafetydist;
    theOffsideLineXReliability.set( ball().posReliability().get() );
  }
  else {
    newOffsideLineX = secondLastPlayerX;
  }
  INFO( Log::SITUATION, 5, "offsideLineX: " + String( newOffsideLineX ) +
      "\n");  
  INFO( Log::SITUATION, 3, "LEAVE Situation::computeOffsideLineX\n");
  return newOffsideLineX;
}  

bool Situation::isOffsideWithIndex( int index ) const
{
  if ( playerWithIndex( index ).team() == THEIR_TEAM_ID ) {
    throwException( InvalidIdException(
        "Can only check offside condition for team members." ) );
  }
  return ( playerWithIndex( index ).pos().x() > offsideLineX() );
}

bool Situation::isOffsideWithId( int playerId ) const
{
  int index = indexOfPlayerWithId( playerId );
  return ( isOffsideWithIndex( index ) );
}
    
StorableObject* Situation::copy() const {
  return new Situation( *this );
}

const Array<int>& Situation::getOurTeamSorted() const {
  if( !areStepsToBallEstimated ) {
    estimateForAllStepsToBall();
  }
  return nearestTeammatesSorted;
}

const Array<int>& Situation::getTheirTeamSorted() const {
  if( !areStepsToBallEstimated ) {
    estimateForAllStepsToBall();
  }
  return nearestEnemiesSorted;
}

const Array<int>& Situation::getStepstoBall() const{
  return stepsToBallOfPlayers;
}

const int Situation::playerStepsToBall(int index) const{
  return stepsToBallOfPlayers[index];
}





