
 
#include "Timer.h"
#include "Log.h"
#include "WorldModel.h"
#include "ModelledBall.h"
#include "Skills.h"
#include <sys/time.h>
#include <signal.h>

/*
 * Aktueller Stand
 *
 *   $RCSfile: Timer.cc,v $
 *   $Revision: 1.29 $
 *   $Date: 1998/08/14 13:21:26 $
 *   $Author: myritz $ 
 *
 */

Timer *globalTimer;

Timer::Timer( Sensors *sensors, Effectors *effectors,
	      Deliberator *deliberator ) {
  theSensors = sensors;
  theEffectors = effectors;
  theDeliberator = deliberator;
  theTimeCounter = 0;
  theTimeSinceLastSentMainAction = 0;
  theTimeSinceLastSentSubAction = 0;
  haveToSee = YES;
  haveToSendSubAction = NO;
  haveToSendMainAction = NO;
  haveToDecideNew = NO;
  theLastSentMainActionSoccerServerTime = 0;
  // erstmal auf neue Informationen warten, bevor nachgedacht wird
  theEffectors->setPlan( globalSkills->doNothing() );
  isLocked = YES;
  // alle 12.5 msecs Timer aufrufen
  // theInternTimeStep = 12500;
  theInternTimeStep = 12500;
  theSentMainStep = globalConf->simulatorStep * 1000 / theInternTimeStep;
  theSentSubStep = theSentMainStep / 4;
    
}

void Timer::run() {

  sigset_t mask;
  struct sigaction vec;
  sigemptyset (&mask);
  sigaddset (&mask, SIGALRM);
  vec.sa_handler = ((void (*) (int))globalSchedule);
  vec.sa_mask = mask;
  // vec.sa_flags = 0;
  vec.sa_flags = SA_NODEFER;
  if (sigaction (SIGALRM, &vec, NULL) == -1) {
    throwException( TimerException( "Timer: run(): signal error" ) );
  }
  
  itimerval zeit;
  zeit.it_value.tv_sec = 0;
  zeit.it_value.tv_usec = theInternTimeStep;
  zeit.it_interval.tv_sec = 0;
  zeit.it_interval.tv_usec = theInternTimeStep;
  setitimer (ITIMER_REAL, &zeit, NULL);
  isLocked = NO;

  INFO( Log::TIMER, 2, "TIMER STARTED\n" );
}

void Timer::lock() {
  if( isLocked ) {
    throwException( 
        TimerException( "tried to lock() while lock() is still active" ) );
  }
  else {
    isLocked = YES;
  }
}

void Timer::unlock() {
  if( !isLocked ) {
    throwException( 
        TimerException( "tried to unlock() while lock() is not active" ) );
  }
  else {
    isLocked = NO;
  }
}

void Timer::stop() {
  
  itimerval zeit;
  isLocked = YES;

  // SIGALARM abschalten;
  zeit.it_value.tv_sec = 0;
  zeit.it_value.tv_usec = 0;
  zeit.it_interval.tv_sec = 0;
  zeit.it_interval.tv_usec = 0;
  setitimer (ITIMER_REAL, &zeit, NULL);
  
  // Signalbehandlung zuruecksetzen;
  sigset_t mask;
  struct sigaction vec;
  sigemptyset (&mask);
  vec.sa_handler = SIG_DFL;
  vec.sa_mask = mask;
  vec.sa_flags = 0;
  sigaction (SIGALRM, &vec, NULL);

  INFO( Log::TIMER, 2, "TIMER STOPPED\n" );
}
  
unsigned long Timer::time() const {
  return (long) ( ( (double) theTimeCounter * theInternTimeStep ) / 1000 );
}

long Timer::timeSinceLastSentSubAction() const {
  return (long) ( ( (double) theTimeSinceLastSentSubAction *
                    theInternTimeStep ) / 1000 );
}

long Timer::timeSinceLastSentMainAction() const {
  return (long) ( ( (double) theTimeSinceLastSentMainAction *
                    theInternTimeStep ) / 1000 );
}


void Timer::schedule() {
  theTimeCounter++;
  theTimeSinceLastSentMainAction++;
  theTimeSinceLastSentSubAction++;

  haveToSee = YES;
  if( theTimeSinceLastSentMainAction >= theSentMainStep ) {
    // In den Spielzustaenden in dennen die Zeit im SoccerServer nicht
    // weiterlaeuft muss selber fuer die Zeiten zum Absenden der MainActions
    // gesorgt werden.
    int gameState = globalWorldModel->gameState();
    if( gameState == WorldModel::BEFORE_KICK_OFF ||
        gameState == WorldModel::TIME_OVER ||
        gameState == WorldModel::GOAL ||
        gameState == WorldModel::OFFSIDE ) { 
      haveToSendMainAction = YES;
    }
    else if( theTimeSinceLastSentSubAction >= theSentSubStep ) {
      haveToSendSubAction = YES;
    }   
  }
  else if( theTimeSinceLastSentSubAction >= theSentSubStep ) {
    haveToSendSubAction = YES;
  }
 
}

void Timer::mainLoop() 
{
  lock();
  try {
    INFO( Log::TIMER, 4, "   new mainLoop at time: " +
          String( (long) time() ) + "\n" );

    if( haveToSee ) {
      INFO( Log::TIMER, 10, "   Sensors started at time: " + 
            String( (long) time() )+ "\n" );
      theSensors->sense();
      theSoccerServerSimStep = theSensors->simStepOfSoccerServer();
      haveToSee = NO;
      INFO( Log::TIMER, 10, "   Sensors ended at time: " + 
            String( (long) time() )+ "\n" );
    }

    INFO( Log::TIMER, 7, "   SoccerServer simStep: " + 
          String( theSoccerServerSimStep ) + "\n" );
    INFO( Log::TIMER, 15, "   LastSentMainActionSoccerServerTime: " +
          String( theLastSentMainActionSoccerServerTime ) + "\n" );
    INFO( Log::TIMER, 15, "   TimeSinceLastSentSubAction: " +
          String( timeSinceLastSentSubAction() ) + "\n" );
    INFO( Log::TIMER, 15, "   TimeSinceLastMainAction: " +
          String( timeSinceLastSentMainAction() ) + "\n" );


    if( theSensors->isNewSeeInfo() ) {
      INFO( Log::TIMER, 10, "   WorldModel started at time: " + 
            String( (long) time() ) + "\n" );
      globalWorldModel->integrateSensorInfo();
      theSensors->initSensors();
      haveToDecideNew = YES;
      INFO( Log::TIMER, 10, "   WorldModel ended at time: " + 
            String( (long) time() ) + "\n" );
    }
    else {
      if( theSensors->isNewRefereeInfo() ) {
        INFO( Log::TIMER, 10, "   WorldModel Referee started at time: " + 
              String( (long) time() ) + "\n" );
        globalWorldModel->integrateHearInfo();
        theSensors->initSensors();
        haveToDecideNew = YES;
        theEffectors->clearQueue();
        INFO( Log::TIMER, 10, "   WorldModel Referee ended at time: " + 
              String( (long) time() ) + "\n" ); 
      }
    }

    if( haveToDecideNew ) {
      INFO( Log::TIMER, 10, "   Deliberator started at time: " + 
            String( (long) time() ) + "\n" );
      theDeliberator->startDeliberation();
      haveToDecideNew = NO;
      INFO( Log::TIMER, 10, "   Deliberator ended at time: " + 
            String( (long) time() ) + "\n" );
    }

    if( theSoccerServerSimStep > theLastSentMainActionSoccerServerTime &&
        theTimeSinceLastSentSubAction >= theSentSubStep ) {
      INFO( Log::TIMER, 5, "   new SoccerServer simStep: " + 
            String( theSoccerServerSimStep ) + "\n" );
      haveToSendMainAction = YES;
    } 
    try {
      if( haveToSendMainAction ) {
        INFO( Log::TIMER, 10, "   Effectors Main started at time: " + 
              String( (long) time() )+ "\n" );
        haveToSendMainAction = NO;
        haveToSendSubAction = NO;
        theLastSentMainActionSoccerServerTime = theSoccerServerSimStep;
        theTimeSinceLastSentMainAction = 0;
        theTimeSinceLastSentSubAction = 0;
        theEffectors->performMainAction( theSensors->lastBodyInfos() );
        INFO( Log::TIMER, 10, "   Effectors Main ended at time: " + 
              String( (long) time() )+ "\n" );
      }
      else if( haveToSendSubAction ) {
        INFO( Log::TIMER, 10, "   Effectors Sub started at time: " + 
              String( (long) time() )+ "\n" );
        haveToSendSubAction = NO;
        theTimeSinceLastSentSubAction = 0;
        theEffectors->performSubAction();
        INFO( Log::TIMER, 10, "   Effectors Sub ended at time: " + 
              String( (long) time() )+ "\n" );
      }
    }
    catch( Effectors::NothingToSendException ) {
      INFO( Log::TIMER, 10, "   Effectors added Action started at time: " +
            String( (long) time() )+ "\n" );
      theEffectors->performAction( Action( SENSE_BODY_ACTION, 
          theSoccerServerSimStep ) );
      INFO( Log::TIMER, 10, "   Effectors added Action ended at time: " +
            String( (long) time() )+ "\n" );
    }
    if( !haveToDecideNew ) {
      if( !theEffectors->isActionLeft() ) {
        globalWorldModel->setDecisionReason( 
            WorldModel::DECISION_REASON_PLAN_ENDED );
        haveToDecideNew = YES;
      }
    }
  }
  catch ( ModelledBall::KickImpossibleException e ) {
    cerr << "Caught KickImpossibleException " << e.toString() << endl;
  }
  
  unlock();
}

Timer::~Timer() {
  stop();
}


void globalSchedule( int signalNumber ) {
  globalTimer->schedule();
}


















